{+-----------------------------------------------------------------------------+
 | Class:       TmwCustomEdit
 | Created:     11.98
 | Last change: 10.1.99
 | Author:      Martin Waldenburg
 | Description: study on how to create a custom edit control without using
 |              a Windows edit control.
 | Version:     0.74 internal beta (see VERSION.RTF for version history)
 | Copyright (c) 1998 Martin Waldenburg
 | All rights reserved.
 |
 | Thanks to : Woo Young Bum, Angus Johnson, Michael Trier, James Jacobson,
 |             Thomas Kurz, Primoz Gabrijelcic, Michael Beck, Andy Jeffries,
 |             Edward Kreis, Brad Stowers, Willo van der Merwe
 |
 | LICENCE CONDITIONS
 |
 | USE OF THE ENCLOSED SOFTWARE
 | INDICATES YOUR ASSENT TO THE
 | FOLLOWING LICENCE CONDITIONS.
 |
 |
 |
 | These Licence Conditions are exlusively
 | governed by the Law and Rules of the
 | Federal Republic of Germany.
 |
 | Redistribution and use in source and binary form, with or without
 | modification, are permitted provided that the following conditions
 | are met:
 |
 | 1. Redistributions of source code must retain the above copyright
 |    notice, this list of conditions and the following disclaimer.
 |    If the source is modified, the complete original and unmodified
 |    source code has to distributed with the modified version.
 |
 | 2. Redistributions in binary form must reproduce the above
 |    copyright notice, these licence conditions and the disclaimer
 |    found at the end of this licence agreement in the documentation
 |    and/or other materials provided with the distribution.
 |
 | 3. Software using this code must contain a visible line of credit.
 |
 | 4. If my code is used in a "for profit" product, you have to donate
 |    to a registered charity in an amount that you feel is fair.
 |    You may use it in as many of your products as you like.
 |    Proof of this donation must be provided to the author of
 |    this software.
 |
 | 5. If you for some reasons don't want to give public credit to the
 |    author, you have to donate three times the price of your software
 |    product, or any other product including this component in any way,
 |    but no more than $500 US and not less than $200 US, or the
 |    equivalent thereof in other currency, to a registered charity.
 |    You have to do this for every of your products, which uses this
 |    code separately.
 |    Proof of this donations must be provided to the author of
 |    this software.
 |
 |
 | DISCLAIMER:
 |
 | THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS'.
 |
 | ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 | PARTICULAR PURPOSE ARE DISCLAIMED.
 |
 | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 | (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 | OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 | INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 | WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 | NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 |
 |  Martin.Waldenburg@T-Online.de
 |
 | Known problems:
 |   - dragging cannot be canceled with the chord clicking - can't fix!
 |
 +----------------------------------------------------------------------------+}

{$I MWEDIT.INC}                                                                 //mt 12/16/1998 added
//mt 12/16/1998 removed {$ObjExportAll On} since it's defined in the include
 
{$IFDEF MWE_FPC}
  {$mode objfpc}
  {$H+}
{$ENDIF}

unit mwCustomEdit;

interface

uses
// Nasty BCB linker error when doing it this way:                               //mt 12/10/1998
//   Windows, WinProcs, Messages, Classes, Controls, Printers,
//   Forms, Graphics, SysUtils, StdCtrls, Clipbrd, ExtCtrls, mwPasSyn;
  {$IFDEF MWE_FPC}
     LCLLinux,
  {$ELSE}
     Windows,
  {$ENDIF}
  Messages, SysUtils, Classes, Controls, Graphics,
  mwKeyCmds,                                                                    //bds 12/23/1998
  ExtCtrls, Forms, StdCtrls, Clipbrd, mwHighlighter;


const
   DIGIT            = ['0'..'9'];
// ALPHA            = ['A'..'Z', 'a'..'z'];                                     //mt 12/10/1998
// break these up because we exceed the 4 byte limit when combined.
   ALPHA_UC         = ['A'..'Z'];
   ALPHA_LC         = ['a'..'z'];

// Max number of book/gutter marks returned from GetMarksForLine - that really
// should be enough.
   maxMarks = 16;                                                               //gp 1/10/1999

type
  PIntArray = ^TIntArray;                                                       //bds 12/24/1998
  TIntArray = array[0..0] of integer;                                           //bds 12/24/1998

  TIndexEvent = procedure(Index: Integer) of object;
  TPaintEvent = procedure(Sender: TObject; ACanvas: TCanvas)                    //jdj 12/12/1998
                  of object;

  {begin}                                                                       //bds 12/20/1998
//  TEditorCommand = word;   moved to mwKeyCmds.pas                             //bds 12/23/1998
  TProcessCommandEvent = procedure(var Command: TmwEditorCommand;
                     var AChar: char; Data: pointer) of object;                 //bds 12/23/1998

  // Replaced by TMark                                                          //CdeB 12/16/1998
  (*
  PSetBookMarkRec = ^TSetBookMarkRec;  { Used in ecSetMarker command }
  TSetBookMarkRec = record
    MarkerNum: integer;
    MarkerPos: TPoint;
  end;
  *)
  {end}                                                                         //bds 12/20/1998

  TCaretType = (ctVerticalLine, ctHorizontalLine, ctHalfBlock, ctBlock);        //bds 12/17/1998

  {begin}                                                                       //tskurz 12/10/1998
  // added mw as a prefix otherwise crNone conflicts with the                   //mt 12/16/1998
  // crNone cursor constant.
  TChangeReason = (mwcrInsert, mwcrDragDropInsert, mwcrDelete,
                   mwcrDeleteAfterCursor,                                       //bds 12/20/1998
                   mwcrDragDropDelete, mwcrLineBreak, mwcrNone);

  TChangePtr = ^TChange;
  TChange = record
              ChangeStr: PChar;
              ChangeReason: TChangeReason;
              ChangeStartPos,
              ChangeEndPos: TPoint;
            end;
            
  //had to move it up here or the header creator on BCB bombs                   //mt 12/16/1998
  {+----------------------------------------------+
   |               -Bookmarks-
   |There are 10 available numbered 0 to 9
   |The X and Y values are 1 based as are the
   |  CaretX and CaretY properties
   |Shift+Ctrl+TheBookMarkNumber will Set/Clear
   |Ctrl+TheBookMarkNumber will Goto
   |The bitmaps are in mwCustomEdit.res and
   |  are named MARK_ + TheBookMarkNumber
  +----------------------------------------------+}
  // Replaced by TMark                                                          //CdeB 12/16/1998
  {
  TBookMarkRec = record                                                         //jdj 12/10/1998
    IsSet: Boolean;
    X: Integer;
    Y: Integer;
    Glyph: TBitmap;
  end;
  }

  TmwCustomEdit = class;                                                        //tskurz 12/15/1998

  {begin}                                                                       //CdeB 12/16/1998, //gp 1/9/1999
  TMark = class
  protected
    fLine, fColumn, fImage: Integer;
    fEdit: TmwCustomEdit;
    fVisible: boolean;                                                          
    fInternalImage: boolean;                                                    
    fBookmarkNum: integer;
    function GetEdit: TmwCustomEdit; virtual;
    procedure SetColumn(const Value: Integer); virtual;
    procedure SetImage(const Value: Integer); virtual;
    procedure SetLine(const Value: Integer); virtual;
    procedure SetVisible(const Value: boolean);
    procedure SetInternalImage(const Value: boolean);
    function GetIsBookmark: boolean;
    procedure SetIsBookmark(const Value: boolean);
  public
    constructor Create(owner: TmwCustomEdit);                                                            
    property Line: integer read fLine write SetLine;
    property Column: integer read fColumn write SetColumn;
    property ImageIndex: integer read fImage write SetImage;
    property BookmarkNumber: integer read fBookmarkNum write fBookmarkNum;
    property Visible: boolean read fVisible write SetVisible;
    property InternalImage: boolean read fInternalImage write SetInternalImage;
    property IsBookmark: boolean read GetIsBookmark write SetIsBookmark;
  end;

  TPlaceMarkEvent = procedure(var mark: TMark) of object;                       //gp 1/9/1999 - changed (again) and renamed

  TMarks = array [1..maxMarks] of TMark;

  { A list of mark objects. Each object cause a litle picture to be drawn in the
    gutter. }
  TMarkList = class(TList)
  protected
    fEdit: TmwCustomEdit;                                                       //gp 1/9/1999
    fOnChange: TNotifyEvent;
    function Get(Index: Integer): TMark;
    procedure Put(Index: Integer; Item: TMark);
    Procedure DoChange;
  public
    constructor Create(owner: TmwCustomEdit);                                   //gp 1/9/1999
    function Add(Item: TMark): Integer;
    function First: TMark;
    function Last: TMark;
    procedure Insert(Index: Integer; Item: TMark);
    function Remove(Item: TMark): Integer;
    procedure Delete(Index: Integer);
    procedure ClearLine(line: integer);
    procedure Place(mark: TMark);
    procedure GetMarksForLine(line: integer; var marks: TMarks);                //gp 1/10/1999
    {$IFDEF MWE_FPC}
    property Items[Index: Integer]: TMark read Get write Put; 
    {$ELSE}
    property Items[Index: Integer]: TMark read Get write Put; default;
    {$ENDIF}
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;
  {end}                                                                         //CdeB 12/16/1998

  TUndoList = class
  private
    fList: TList;
    fCanUndo: Integer;
    fMaxUndo: Integer;
    fOwner: TmwCustomEdit;                                                      //tskurz 12/15/1998
    function GetCanUndo: Integer;
    procedure SetMaxUndo(const Value: Integer);
  protected
    procedure RemoveChange(index: Integer);
  public
    constructor Create(AOwner: TmwCustomEdit);                                  //tskurz 12/15/1998
    destructor Destroy; override;
    procedure AddChange(ChangeReason: TChangeReason; ChangeStartPos,
               ChangeEndPos: TPoint; ChangeStr: PChar);
    function GetChange(var ChangeStartPos, ChangeEndPos: TPoint;
               var ChangeStr: PChar): TChangeReason;
    function GetChangeReason: TChangeReason;                                    //tskurz 12/16/1998
    procedure ClearList;                                                        //tskurz 1/4/1999
    property CanUndo: Integer read GetCanUndo;
    property MaxUndo: Integer read FMaxUndo write SetMaxUndo;
  end;
  {end}                                                                         //tskurz 12/10/1998


  TmwEditList = class(TStringList) 
  private
    FOnAdded: TNotifyEvent;
    FOnDeleted: TIndexEvent;
    FOnInserted: TIndexEvent;
    FOnLoaded: TNotifyEvent;
    fOnPutted: TIndexEvent;
  protected
    procedure Put(Index: Integer; const S: string); override;
  public
    function Add(const S: string): Integer; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure SetTextStr(const Value: string); override;
    property OnAdded: TNotifyEvent read FOnAdded write FOnAdded;
    property OnDeleted: TIndexEvent read FOnDeleted write FOnDeleted;
    property OnInserted: TIndexEvent read FOnInserted write FOnInserted;
    property OnLoaded: TNotifyEvent read FOnLoaded write FOnLoaded;
    property OnPutted: TIndexEvent read FOnPutted write FOnPutted;
  published
  end;

  TmwSelectedColor = class(TPersistent)
  private
    FBG: TColor;
    FFG: TColor;
    FOwner: TmwCustomEdit;
    procedure SetBG(Value: TColor);
    procedure SetFG(Value: TColor);
  public
    constructor Create(AOwner: TmwCustomEdit);
  published
    property Background: TColor read FBG write SetBG default clHighLight;       //gp 12/18/1998
    property Foreground: TColor read FFG write SetFG default clHighLightText;   //gp 12/18/1998
  end;

  TmwBookMarkOpt = class(TPersistent)                                           //jdj 12/12/1998
  private
    fEnableKeys: Boolean;
    fGlyphsVisible: Boolean;
    fLeftMargin: Integer;
    fOwner: TmwCustomEdit;
    fXoffset: integer;
    procedure SetGlyphsVisible(Value: Boolean);
    procedure SetLeftMargin(Value: Integer);
    procedure SetXOffset(const Value: integer);
  public
    constructor Create(AOwner: TmwCustomEdit);
  published
    property EnableKeys: Boolean read fEnableKeys write fEnableKeys default True;
    property GlyphsVisible: Boolean read fGlyphsVisible write SetGlyphsVisible default True;
    property LeftMargin: Integer read fLeftMargin write SetLeftMargin default 2;
    property Xoffset: integer read fXoffset write SetXOffset default 12;
  end;

  PGutterOffsets = ^TGutterOffsets;
  TGutterOffsets = array [0..0] of integer;

  TmwCustomEdit = class(TCustomControl)
  private
    fBlockBegin: TPoint;
    fBlockEnd: TPoint;
    fBottomLine: Integer;
    fCaretX: Integer;
    fCaretY: Integer;
    fCaretYPix: Integer;
    fCaretVisible: Boolean;
    fCharsInWindow: Integer;
    // Don't assign to directly, always use CharsWidth property                 //bds 12/24/1998
    fCharWidth: Integer;
    fDblClicked: Boolean;
    fEditRect: TRect;
    fEditTimer: TTimer;
    fElapsed: Boolean;
    fFontDummy: TFont;
    fGutterColor: TColor;
    fGutterRect: TRect;
    fGutterWidth: Integer;
    fInEditRect: Boolean;
    fInserting: Boolean;
//    fLineRect: TRect;                                                         //bds 12/27/1998
    fLines: TStrings;
    // NEVER assign to directly, always use SetLinesInWindow procedure!         //gp 1/10/1999
    fLinesInWindow: Integer;
    fLeftChar: Integer;
    // Don't assign to directly, always use MaxLeftChar property                //bds 12/24/1998
    fMaxLeftChar: Integer;
    fPaintLock: Integer;
    fReadOnly: Boolean;
    fRightChar: Integer;
    fRightEdge: Integer;
    fRightEdgeOffset: Integer;
    FScrollBars: TScrollStyle;
    fTextHeight: Integer;
    fTextOffset: Integer;
    fTopLine: Integer;
    fHighLighter: TmwCustomHighLighter;
    fSelectedColor: TmwSelectedColor;                                           //jdj 12/09/1998
    fUndoList : TUndoList;                                                      //tskurz 12/10/1998
    fRedoList : TUndoList;                                                      //tskurz 1/4/1999
    fBookMarks: array[0..9] of TMark;                                           //jdj 12/10/1998 //CdeB 12/16/1998 - modified to use marks
    fDragBlockBegin: TPoint;                                                    //ked 12/10/1998
    fDragBlockEnd: TPoint;                                                      //ked 12/10/1998
    fMouseDownX: integer;                                                       //ked 12/10/1998
    fMouseDownY: integer;                                                       //ked 12/10/1998
    fWaitForDragging: boolean;                                                  //gp 12/12/1998
    fBookMarkOpt: TmwBookMarkOpt;                                               //jdj 12/12/1998
    fOnPaint: TPaintEvent;                                                      //jdj 12/12/1998
    fOnChange: TNotifyEvent;                                                    //tskurz 12/15/1998
    fBorderStyle: TBorderStyle;                                                 //bds 12/16/1998
    fHideSelection: boolean;                                                    //bds 12/16/1998
    fScrollPastEOL: boolean;                                                    //bds 12/16/1998
    fTextBM: TBitMap;                                                           //WvdM 12/18/1998                                                           
    fOverwriteCaret: TCaretType;                                                //bds 12/17/1998
    fInsertCaret: TCaretType;                                                   //bds 12/17/1998
    fCaretOffset: TPoint;                                                       //bds 12/17/1998
    fOnProcessCommand: TProcessCommandEvent;                                    //bds 12/20/1998
    fOnProcessUserCommand: TProcessCommandEvent;                                //bds 12/20/1998
    fKeyStrokes: TmwKeyStrokes;                                                 //bds 12/23/1998
    fModified: Boolean;                                                         //CdeB 12/16/1998
    fETODist: PIntArray;                                                        //bds 12/24/1998
    fMarkList: TMarkList;                                                       //CdeB 12/16/1998
    fBookmarkImages: TImageList;                                                //CdeB 12/16/1998, //gp 1/2/1999 - renamed
    fBookMarksGlyphs: array [0..9] of TBitmap;                                  //CdeB 12/16/1998
    fOnPlaceMark: TPlaceMarkEvent;                                              //CdeB 12/16/1998, //gp 1/9/1999 - renamed
    fGutterOffsets: PGutterOffsets;                                             //gp 1/10/1999
    fExtraLineSpacing: integer;                                                 //gp 1/10/1999
    function CaretXPix: Integer;
    function CaretYPix: Integer;
    procedure ComputeCaret(X, Y: Integer);
    function GetBlockBegin: TPoint;
    function GetBlockEnd: TPoint;
    function GetCaretX: Integer;
    function GetCaretY: Integer;
    function GetFont: TFont;
    function GetLeftChar: Integer;
    function GetLineCount: Integer;
    function GetLineText: String;
    function GetSelAvail: Boolean;
    function GetText: String;
    function GetTopLine: Integer;
    procedure FontChanged(Sender: TObject);
    function LeftSpaces(Line: String): Integer;
    procedure LinesChanged(Sender: TObject);
    procedure PaintControl;
    procedure SetBlockBegin(Value: TPoint);
    procedure SetBlockEnd(Value: TPoint);
    procedure SetWordBlock(Value: TPoint);
    procedure SetCaretX(Value: Integer);
    procedure SetCaretY(Value: Integer);
    procedure SetGutterWidth(Value: Integer);
    procedure SetFont(const Value: TFont);
    procedure SetLeftChar(Value: Integer);
    procedure SetLines(Value: TStrings);
    procedure SetLineText(Value: String);
    procedure SetScrollBars(const Value: TScrollStyle);
    procedure SetText(const Value: String);
    procedure SetTopLine(Value: Integer);
    procedure TimerFired(Sender: TObject);
    procedure UpdateCaret;
    procedure UpdateScrollBars;
    function GetSelText: String;
    procedure SetSelText(const Value: String);
    procedure ScanFrom(Index: Integer);
    procedure PaintHighLighted(TextCanvas: TCanvas);                            //bds 12/27/1998
    function GetCanUndo: Boolean;                                               //tskurz 12/10/1998
    function GetCanRedo: Boolean;                                               //tskurz 1/4/1999
    procedure SetGutterColor(Value: TColor);                                    //jdj 12/12/1998
    procedure SetRightEdge(Value: Integer);                                     //jdj 12/12/1998
    procedure SetMaxUndo(const Value: Integer);
    function GetMaxUndo: Integer;                                               //tskurz 12/14/1998
    procedure SetHighlighter(const Value: TmwCustomHighLighter);                //gp 12/16/1998
    procedure SetBorderStyle(Value: TBorderStyle);                              //bds 12/16/1998
    procedure SetHideSelection(const Value: boolean);                           //bds 12/16/1998
    procedure SetInsertMode(const Value: boolean);                              //bds 12/16/1998
    procedure SetScrollPastEOL(const Value: boolean);                           //bds 12/16/1998
    procedure SetInsertCaret(const Value: TCaretType);                          //bds 12/17/1998
    procedure SetOverwriteCaret(const Value: TCaretType);                       //bds 12/17/1998
    function GetCaretXY: TPoint;                                                //bds 12/17/1998
    procedure SetCaretXY(const Value: TPoint);                                  //bds 12/17/1998
    procedure SetKeystrokes(const Value: TmwKeyStrokes);                        //bds 12/23/1998
    procedure SetMaxLeftChar(const Value: integer);                             //bds 12/24/1998
    procedure SetCharWidth(const Value: integer);                               //bds 12/24/1998
    procedure SetLinesInWindow(const Value: integer);                           //gp 1/10/1999
    procedure SetExtraLineSpacing(const Value: integer);
    procedure SetSelTextExternal(const Value: String);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure DecPaintLock;
    procedure HideCaret;
    procedure IncPaintLock;
    procedure ListAdded(Sender: TObject);
    procedure ListDeleted(Index: Integer);
    procedure ListInserted(Index: Integer);
    procedure ListLoaded(Sender: TObject);
    procedure ListPutted(Index: Integer);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DblClick; override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure PaintGutter(TextCanvas: TCanvas); virtual;                        //bds 12/27/1998
    procedure PaintText(TextCanvas: TCanvas); virtual;                          //bds 12/27/1998
    procedure Paint; override;
    procedure SetName(const Value: TComponentName); override;
    procedure ShowCaret;
    procedure Loaded; override;                                                 //jdj 12/12/1998
    function CoorToIndex(CPos: TPoint): integer;                                //ked 12/10/1998
    function IndexToCoor(ind: integer): TPoint;                                 //ked 12/10/1998
    procedure DragOver(Source: TObject; X, Y: Integer;                          //ked 12/10/1998
                       State: TDragState; var Accept: Boolean); override;
    procedure InitializeCaret;                                                  //bds 12/16/1998
    procedure MarkListChange(Sender: TObject);                                  //CdeB 12/16/1998
    procedure SetBookmarkImages(const Value: TImageList);                       //CdeB 12/16/1998
    procedure WMEraseBkgnd(var Message: TMessage); message WM_ERASEBKGND;
    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMHScroll(var Message: TWMScroll); message WM_HSCROLL;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
    procedure ProcessCommand(var Command: TmwEditorCommand; var AChar: char;
                             Data: pointer); virtual;                           //bds 12/23/1998
    // If the translations requires Data, memory will be allocated for it via a
    // GetMem call.  The client must call FreeMem on Data if it is not NIL.
    function TranslateKeyCode(Code: word; Shift: TShiftState;
                          var Data: pointer): TmwEditorCommand;                 //bds 12/23/1998
    property PaintLock: Integer read fPaintLock;
    property SelAvail: Boolean read GetSelAvail;
    property CharWidth: integer read fCharWidth write SetCharWidth;             //bds 12/24/1998
    property MaxLeftChar: integer read fMaxLeftChar write SetMaxLeftChar;       //bds 12/24/1998
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure WndProc(var msg: TMessage); override;
    procedure DragDrop(Source: TObject; X, Y: Integer);                         //ked 12/10/1998
                override;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure PasteFromClipboard;
    procedure SelectAll;
    procedure ClearAll;                                                         //bds 12/20/1998
    procedure Undo;                                                             //tskurz 12/10/1998
    procedure Redo;                                                             //tskurz 1/4/1999
    procedure ClearBookMark(BookMark: Integer);                                 //jdj 12/10/1998, //CdeB 12/16/1998 - removed, //gp 12/31/1998 - reinstated
    procedure GotoBookMark(BookMark: Integer);                                  //jdj 12/10/1998, //CdeB 12/16/1998 - removed, //gp 12/31/1998 - reinstated
    procedure SetBookMark(BookMark: Integer; X: Integer; Y: Integer);           //jdj 12/10/1998, //CdeB 12/16/1998 - removed, //gp 12/31/1998 - reinstated
    function GetBookMark(BookMark: integer; var X, Y: integer): boolean;        //gp 1/10/1999
    function IsBookmark(BookMark: integer): boolean;                            //gp 1/10/1999
    procedure Notification(AComponent: TComponent;                              //gp 12/16/1998
                Operation: TOperation); override;
    procedure EnsureCursorPosVisible;                                           //bds 12/17/1998
    procedure CommandProcessor(Command: TmwEditorCommand; AChar: char;
                               Data: pointer); virtual;                         //bds 12/23/1998
    function NextWordPos: TPoint; virtual;
    function LastWordPos: TPoint; virtual;
    procedure SetDefaultKeystrokes; virtual;                                    //bds 12/23/1998
    procedure BeginUpdate;                                                      //jdj 12/27/1998
    procedure EndUpdate;                                                        //jdj 12/27/1998
    function GetSelEnd: integer;
    function GetSelStart: integer;
    procedure SetSelEnd(const Value: integer);
    procedure SetSelStart(const Value: integer);
    procedure ClearUndo;                                                        //gp 1/10/1999
    property Elapsed: Boolean read fElapsed;
    property BlockBegin: TPoint read GetBlockBegin write SetBlockBegin;
    property BlockEnd: TPoint read GetBlockEnd write SetBlockEnd;
    property CaretX: Integer read GetCaretX write SetCaretX;
    property CaretY: Integer read GetCaretY write SetCaretY;
    property CaretXY: TPoint read GetCaretXY write SetCaretXY;                  //bds 12/17/1998
    property LineText: String read GetLineText write SetLineText;               //gp 12/30/1998 - moved from published to public
    property CharsInWindow: Integer read fCharsInWindow;
    property LinesInWindow: Integer read fLinesInWindow;
    property SelText: String read GetSelText write SetSelTextExternal;          //tskurz 1/7/1999 - new Set
    property Text: String read GetText write SetText;
    property Modified: Boolean read fModified write fModified;                  //CdeB 12/16/1998
    property Marks: TMarkList read fMarkList;                                   //CdeB 12/16/1998
  published
    property Align;
    {begin}                                                                     //bds 12/16/1998
    {$IFDEF MWE_COMPILER_4_UP}
    property Anchors;
    property Constraints;
    {$ENDIF}
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle
               default bsSingle;
    {end}                                                                       //bds 12/16/1998
    property BookMarkOptions: TmwBookMarkOpt                                    //jdj 12/12/1998
               read fBookMarkOpt write fBookMarkOpt;
    property CanUndo: Boolean read GetCanUndo;
    property CanRedo: boolean read GetCanRedo;                                  //tskurz 1/4/1999
    property Color;
    property Ctl3D;
    property Enabled;
    property ExtraLineSpacing: integer                                          //gp 1/10/1999
               read fExtraLineSpacing write SetExtraLineSpacing default 0;
    property Font: TFont read GetFont write SetFont;
    property GutterColor: TColor read fGutterColor write SetGutterColor;        //jdj 12/12/1998 added Set
    property GutterWidth: Integer read fGutterWidth write SetGutterWidth;
    property Height;
    property HideSelection: boolean read fHideSelection write SetHideSelection  //bds 12/16/1998
               default false;
    property HighLighter: TmwCustomHighLighter                                  //gp 12/16/1998 - added Set
               read fHighLighter write SetHighlighter;
    property BookmarkImages: TImageList                                         //CdeB 12/16/1998, //gp 1/2/1999 - renamed
               read fBookmarkImages write SetBookmarkImages;
    property InsertCaret: TCaretType read FInsertCaret write SetInsertCaret     //bds 12/17/1998
               default ctVerticalLine;
    property InsertMode: boolean read fInserting write SetInsertMode            //bds 12/16/1998
               default true;
    property Keystrokes: TmwKeyStrokes                                          //bds 12/23/1998
               read FKeystrokes write SetKeystrokes;
    property LeftChar: Integer read GetLeftChar write SetLeftChar;
    property Lines: TStrings read fLines write SetLines;
    property LineCount: Integer read GetLineCount;
    property MaxUndo: Integer read GetMaxUndo write SetMaxUndo;                 //tskurz 12/14/1998
    property Name;
    property OverwriteCaret: TCaretType read FOverwriteCaret                    //bds 12/17/1998
               write SetOverwriteCaret default ctBlock;
    property ParentColor;
    property ParentCtl3D;                                                       //bds 12/16/1998
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly: Boolean read fReadOnly write fReadOnly;
    property RightEdge: Integer read fRightEdge write SetRightEdge default 80;  //jdj 12/12/1998 added Set-Default
    property ShowHint;
    property ScrollBars: TScrollStyle
               read FScrollBars write SetScrollBars default ssBoth;
    property ScrollPastEOL: boolean read FScrollPastEOL write SetScrollPastEOL  //bds 12/16/1998
               default true;
    property SelectedColor: TmwSelectedColor                                    //jdj 12/09/1998
               read FSelectedColor write FSelectedColor;
    property TabOrder;
    property TabStop default True;                                              //jdj 12/12/1998 Added Default
    property Tag;
    property TopLine: Integer read GetTopLine write SetTopLine;
    property Visible;
    property Width;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;             //tskurz 12/15/1998
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;                                                       //levin 1/6/1999
    property OnMouseMove;                                                       //levin 1/6/1999
    property OnMouseUp;                                                         //levin 1/6/1999
    property OnPaint: TPaintEvent read fOnPaint write fOnPaint;                 //jdj 12/12/1998
    property OnPlaceBookmark: TPlaceMarkEvent                                   //CdeB 12/16/1998
               read FOnPlaceMark write FOnPlaceMark;
    property OnProcessCommand: TProcessCommandEvent                             //bds 12/20/1998
               read FOnProcessCommand write FOnProcessCommand;
    property OnProcessUserCommand: TProcessCommandEvent                         //bds 12/20/1998
               read FOnProcessUserCommand write FOnProcessUserCommand;
    property OnStartDrag;
  end;

procedure Register;

implementation

{$IFNDEF MWE_FPC} 
  {$R mwCustomEdit.RES}                                                     //jdj 12/10/1998
{$ENDIF}

procedure Register;
begin
{$IFNDEF MWE_FPC}  
  RegisterComponents('mw', [TmwCustomEdit]);
{$ENDIF}
end;

function max(x, y: integer): integer;
begin
  if x > y then
    result := x else
    result := y;
end;

function min(x, y: integer): integer;
begin
  if x < y then
    result := x else
    result := y;
end;

function maxPoint(P1, P2: TPoint): TPoint;                                      //tskurz 12/14/1998
begin
  if P1.y > P2.y then
  begin
    result := P1;
    exit;
  end else
    if P1.y < P2.y then
    begin
      result := P2;
      exit;
    end;
  if P1.x > P2.x then
    result := P1 else
    result := P2;
end;

function minPoint(P1, P2: TPoint): TPoint;                                      //tskurz 12/14/1998
begin
  if P1.y < P2.y then
  begin
    result := P1;
    exit;
  end else
    if P1.y > P2.y then
    begin
      result := P2;
      exit;
    end;
  if P1.x < P2.x then
    result := P1 else
    result := P2;
end;

{ TmwEditList }

function TmwEditList.Add(const S: string): Integer;
begin
  BeginUpdate;
  Result := inherited Add(S);
  if Assigned(FOnAdded) then FOnAdded(Self);
  EndUpdate;
end;

procedure TmwEditList.Delete(Index: Integer);
begin
  BeginUpdate;
  inherited Delete(Index);
  if Assigned(FOnDeleted) then fOnDeleted(Index);
  EndUpdate;
end;

procedure TmwEditList.Insert(Index: Integer; const S: string);
begin
  BeginUpdate;
  inherited Insert(Index, S);
  if Assigned(FOnInserted) then fOnInserted(Index);
  EndUpdate;
end;

procedure TmwEditList.Put(Index: Integer; const S: string);
begin
  BeginUpdate;
  inherited Put(Index, S);
  if Assigned(FOnPutted) then fOnPutted(Index);
  EndUpdate;
end;

procedure TmwEditList.SetTextStr(const Value: string);
begin
  BeginUpdate;
  inherited SetTextStr(Value);
  if Assigned(FOnLoaded) then FOnLoaded(Self);
  EndUpdate;
end;


{ TmwSelectedColor }

constructor TmwSelectedColor.Create(AOwner: TmwCustomEdit);
begin
  inherited Create;
  FOwner := AOwner;
  FBG := clHighLight;
  FFG := clHighLightText;
end;

procedure TmwSelectedColor.SetBG(Value: TColor);
begin
  if (FBG <> Value) then
    begin
      FBG := Value;
      if Assigned(FOwner) then
        FOwner.Invalidate;
    end;
end;

procedure TmwSelectedColor.SetFG(Value: TColor);
begin
  if (FFG <> Value) then
    begin
      FFG := Value;
      if Assigned(FOwner) then
        FOwner.Invalidate;
    end;
end;

{ TmwCustomEdit }

procedure TmwCustomEdit.ComputeCaret(X, Y: Integer);
begin
  {begin}                                                                       //gp 12/22/1998
// changed into CaretXY so that CaretY is recalculated first
//CaretY := Y div fTextHeight + TopLine;
//CaretX := (X + LeftChar * fCharWidth - FGutterWidth - 2) div fCharWidth;
  CaretXY := Point((X + LeftChar * fCharWidth - FGutterWidth - 2) div fCharWidth,
                   Y div fTextHeight + TopLine);
  {end}                                                                         //gp 12/22/1998
end;

procedure TmwCustomEdit.CopyToClipboard;
begin
  if SelAvail then ClipBoard.SetTextBuf(PChar(SelText));
end;

procedure TmwCustomEdit.CutToClipboard;
begin
  if SelAvail then
  begin
    ClipBoard.SetTextBuf(PChar(SelText));
    FUndoList.AddChange(mwcrDelete,fBlockBegin,fBlockEnd,PChar(SelText));       //tskurz 12/10/1998
    SetSelText('');
  end;
end;

constructor TmwCustomEdit.Create(AOwner: TComponent);
var
  i: Integer;                                                                   //jdj 12/10/1998
begin
  fLines := TmwEditList.Create;
  fFontDummy := TFont.Create;
  fUndoList := TUndoList.Create(self);                                          //tskurz 12/15/1998
  fRedoList := TUndoList.Create(self);                                          //tskurz 1/4/1999
  inherited Create(AOwner);
  fTextBM := TBitmap.Create;                                                    //WvdM 12/18/1998
  {$IFDEF MWE_COMPILER_4_UP}                                                    //gp 11/09/1998, //mt 12/16/1998
    DoubleBuffered := false;                                                    //gp 12/30/1998 - disabled as we are doing buffering on our own
  {$ENDIF}
  if not (csDesigning in ComponentState) then fLines.Add('');                   //jdj 12/08/1998
  {$IFDEF MWE_FPC}
  TmwEditList(fLines).OnAdded := @ListAdded; 
  TmwEditList(fLines).OnDeleted := @ListDeleted; 
  TmwEditList(fLines).OnInserted := @ListInserted; 
  TmwEditList(fLines).OnLoaded := @ListLoaded; 
  TmwEditList(fLines).OnPutted := @ListPutted; 
  {$ELSE}
  TmwEditList(fLines).OnAdded := ListAdded;
  TmwEditList(fLines).OnDeleted := ListDeleted;
  TmwEditList(fLines).OnInserted := ListInserted;
  TmwEditList(fLines).OnLoaded := ListLoaded;
  TmwEditList(fLines).OnPutted := ListPutted;
  {$ENDIF}
  fSelectedColor := TmwSelectedColor.Create(Self);                              //jdj 12/09/1998
  fBookMarkOpt := TmwBookMarkOpt.Create(Self);                                  //jdj 12/12/1998
  fEditTimer := TTimer.Create(Self);
  fEditTimer.Interval := 100;
  {$IFDEF MWE_FPC}
     // Fixed: Causes a sigsev fault  
     //feditTimer.OnTimer := @TimerFired; 
  {$ELSE}
     feditTimer.OnTimer := TimerFired;
  {$ENDIF}
  fCharsInWindow := 10;
//fCharWidth := 8;                                                              //gp 12/24/1998 - calculated in FontChanged
//fLinesInWindow := 10;                                                         //gp 12/24/1998 - calculated in FontChanged
//fTextHeight := 16;                                                            //gp 12/24/1998 - calculated in FontChanged
// fRightEdge has to be set before FontChanged is called for the first time
  fRightEdge := 80;                                                             //gp 1/2/1999
  fGutterWidth := 30;
  fTextOffset := fGutterWidth + 2;
  ControlStyle := ControlStyle + [csOpaque, csSetCaption];
  Height := 150;
  Width := 200;
  Cursor := crIBeam;
  Color := clWindow;
  fCaretX := 0;
  fCaretY := 0;
  fCaretOffset := Point(0,0);                                                   //bds 12/17/1998
  fGutterColor := clBtnFace;
  fFontDummy.Name := 'Courier New';
  fFontDummy.Size := 10;
  Font.Assign(fFontDummy);
  {$IFDEF MWE_FPC}
     Font.OnChange := @FontChanged;
  {$ELSE}
     Font.OnChange := FontChanged; 
  {$ENDIF}
  FontChanged(nil);
  ParentFont := False;
  ParentColor := False;
  TabStop := True;                                                              //jdj 12/12/1998
  {$IFDEF MWE_FPC}
     TStringList(Lines).OnChange := @LinesChanged;
  {$ELSE}
     TStringList(Lines).OnChange := LinesChanged; 
  {$ENDIF}
  with fGutterRect do begin
    Left := 0;
    Top := 0;
    Right := fGutterWidth;                                                      //gp 12/12/1998
    Bottom := 150;
  end;
  fInserting := True;
  fLeftChar := 0;
  // Just to make the pointer valid and must come before MaxLeftChar assign     //bds 12/24/1998
  // Not really. In fact, this crashes by occassion. We should assign to nil    //MB  01/22/2000
//  fETODist := AllocMem(1);                                                      //bds 12/24/1998
  fETODist:= nil;
  MaxLeftChar := 1024;                                                          //bds 12/24/1998
  fPaintLock := 0;
  fReadOnly := False;
  fScrollBars := ssBoth;
  fTopLine := 0;
  fBlockBegin := (Point(0, 0));
  fBlockEnd := (Point(0, 0));

  for i := 0 to 9 do begin                                                      //jdj 12/10/1998
    fBookMarksGlyphs[i] := TBitmap.Create;                                      //CdeB 12/16/1998
    {$IFDEF MWE_FPC}
      { TODO: Write replacement code for the LoadFromResourceName call below. Needs
        to load the MARK bitmap. Need to extract the bitmap from the resource in
        Windows since there are no resource files in Linux. }
    {$ELSE}
    fBookMarksGlyphs[i].LoadFromResourceName(HInstance, 'MARK_' + IntToStr(i)); //CdeB 12/16/1998
    {$ENDIF}
  end;

  fBorderStyle := bsSingle;                                                     //bds 12/16/1998
  fHideSelection := false;                                                      //bds 12/16/1998
  fScrollPastEOL := true;                                                       //bds 12/16/1998
  fInsertCaret := ctVerticalLine;                                               //bds 12/17/1998
  fOverwriteCaret := ctBlock;                                                   //bds 12/17/1998
  fKeystrokes := TmwKeyStrokes.Create(Self);                                    //bds 12/23/1998
  fMarkList:= TMarkList.Create(self);                                           //CdeB 12/16/1998, //gp 1/9/1999
  {$IFDEF MWE_FPC}
  fMarkList.OnChange:= @MarkListChange;                                          //CdeB 12/16/1998 
  {$ELSE}
  fMarkList.OnChange:= MarkListChange;                                          //CdeB 12/16/1998
  {$ENDIF}
  SetDefaultKeystrokes;                                                         //bds 12/23/1998
end;

procedure TmwCustomEdit.CreateParams(var Params: TCreateParams);
const
  ScrollBar: array[TScrollStyle] of DWORD = (0, WS_HSCROLL, WS_VSCROLL,
    WS_HSCROLL or WS_VSCROLL);
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);                  //bds 12/16/1998
begin
  inherited CreateParams(Params);
  with Params do
  begin
    {begin}                                                                     //bds 12/16/1998
    Style := Style or ScrollBar[FScrollBars] or BorderStyles[fBorderStyle]
                   or WS_CLIPCHILDREN;
    {$IFDEF MWE_FPC}
    if NewStyleControls and Ctl3D and (fBorderStyle = TBorderStyle(bsSingle)) then 
    {$ELSE}
    if NewStyleControls and Ctl3D and (fBorderStyle = bsSingle) then 
    {$ENDIF}
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    {end}                                                                       //bds 12/16/1998
  end;
end;

procedure TmwCustomEdit.DecPaintLock;
begin
  dec(fPaintLock);
  {begin}                                                                       //bds 12/17/1998
//if PaintLock = 0 then Paint;
  if (PaintLock = 0) and (Parent <> NIL) then                                   //bds 12/25/1998
  begin
    UpdateScrollBars;
    Paint;
  end;
  {end}                                                                         //bds 12/17/1998
end;

destructor TmwCustomEdit.Destroy;
var
  i: Integer;                                                                   //jdj 12/10/1998
begin
  fKeyStrokes.Free;                                                             //bds 12/23/1998
  fFontDummy.Free;
  Lines.Free;
  fSelectedColor.Free;                                                          //jdj 12/09/1998
  fBookMarkOpt.Free;                                                            //jdj 12/12/1998
  fBookmarkImages := nil;                                                       //CdeB 12/16/1998, //gp 1/2/1999
  fMarkList.Free;                                                               //CdeB 12/16/1998
  fEditTimer.Free;
  fUndoList.Free;                                                               //tskurz 12/10/1998
  fRedoList.Free;                                                               //tskurz 1/4/1999
  for i := 0 to 9 do                                                            //jdj 12/10/1998
//  fBookMarks[i].Glyph.Free;                                                   //CdeB 12/16/1998 - removed
    fBookMarksGlyphs[i].Free;                                                   //CdeB 12/16/1998
  fTextBM.Free;                                                                 //WvdM 12/18/1998
  FreeMem(fETODist);                                                            //bds 12/24/1998
  if assigned(fGutterOffsets) then FreeMem(fGutterOffsets);                     //gp 1/10/1999

  inherited Destroy;
end;

function TmwCustomEdit.GetBlockBegin: TPoint;
begin
  Result := fBlockBegin;
  if fBlockEnd.Y < fBlockBegin.Y then Result := fBlockEnd else
    if fBlockEnd.Y = fBlockBegin.Y then
      if fBlockEnd.X < fBlockBegin.X then Result := fBlockEnd;
end;

function TmwCustomEdit.GetBlockEnd: TPoint;
begin
  Result := fBlockEnd;
  if fBlockEnd.Y < fBlockBegin.Y then Result := fBlockBegin else
    if fBlockEnd.Y = fBlockBegin.Y then
      if fBlockEnd.X < fBlockBegin.X then Result := fBlockBegin;
end;

function TmwCustomEdit.GetCaretX: Integer;
begin
  Result := fCaretX + 1;
end;

function TmwCustomEdit.CaretXPix: Integer;
begin
  Result := fCaretX * fCharWidth + fTextOffset;
end;

function TmwCustomEdit.GetCaretY: Integer;
begin
  Result := fCaretY + 1;
end;

function TmwCustomEdit.CaretYPix: Integer;
begin
  fCaretYPix := fCaretY * fTextHeight - fTopLine * fTextHeight + 1;
  Result := fCaretYPix;
end;

procedure TmwCustomEdit.FontChanged(Sender: TObject);
{begin}                                                                         //bds 12/24/1998
var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;
  TMFont: HFONT;
  TMLogFont: TLogFont;
begin
  GetObject(Font.Handle, SizeOf(TLogFont), @TMLogFont);

  // We need to get the font with bold and italics set so we have enough room
  // for the widest character it will draw.  This is because fixed pitch fonts
  // always have the same character width, but ONLY when the same attributes
  // are being used.  That is, a bold character can be wider than a non-bold
  // character in a fixed pitch font.
  TMLogFont.lfWeight := FW_BOLD;
  TMLOgFont.lfItalic := 1;
  TMFont := CreateFontIndirect(TMLogFont);
  try
    DC := GetDC(0);
    try
      Save := SelectObject(DC, TMFont);
      GetTextMetrics(DC, Metrics);
      SelectObject(DC, Save);
    finally
      ReleaseDC(0, DC);
    end;
  finally
    DeleteObject(TMFont);
  end;
  with Metrics do
  begin
    // Note:  Through trial-and-error I found that tmAveCharWidth should be used
    // instead of tmMaxCharWidth as you would think.  I'm basing this behavior
    // on the Delphi IDE editor behavior.  If tmMaxCharWidth is used, we end up
    // with chars being much farther apart than the same font in the IDE.
    CharWidth := tmAveCharWidth;

    fTextHeight := tmHeight + tmExternalLeading + fExtraLineSpacing;            //gp 1/10/1999
    {begin}                                                                     //gp 12/24/1998
    if assigned(Parent) then begin
      SetLinesInWindow((ClientHeight) div fTextHeight);                         //gp 1/10/1999 - changed to procedure call
      fCharsInWindow := (ClientWidth - fGutterWidth - 2) div fCharWidth;
    end;
    {end}                                                                       //gp 12/24/1998
    fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;
    if PaintLock = 0 then Invalidate;
  end;
end;

function TmwCustomEdit.GetFont: TFont;
begin
  Result := inherited Font;
end;

function TmwCustomEdit.GetLineCount: Integer;
begin
  Result := Lines.Count;
end;

function TmwCustomEdit.GetLeftChar: Integer;
begin
  Result := fLeftChar + 1;
end;

function TmwCustomEdit.GetLineText: String;
begin
  Result := '';
  if (Lines.Count > 0) and (fCaretY < Lines.Count) then
    Result := Lines[fCaretY];
end;

function TmwCustomEdit.GetSelAvail: Boolean;
begin
  Result := False;
  if (fBlockBegin.X <> fBlockEnd.X) or
    (fBlockBegin.Y <> fBlockEnd.Y) then
    Result := True;
end;

function TmwCustomEdit.GetSelText: String;
var
  BB, BE: TPoint;
  First, Last: Integer;
  Helper: TStringList;
begin
  Result := '';
  if SelAvail then
  begin
    Helper := TStringList.Create;
    BB := BlockBegin;
    BE := BlockEnd;
    if BB.Y = BE.Y then
    begin
      Result := (Copy(Lines[BB.Y - 1], BB.X, BE.X - BB.X));
    end else
    begin
      First := BB.Y - 1;
      Last := BE.Y - 1;
      Helper.Add(Copy(Lines[First], BB.X, Length(Lines[First])));
      inc(First);
      while First < Last do
      begin
        Helper.Add(Lines[First]);
        inc(First);
      end;
      Result := Helper.Text + Copy(Lines[First], 1, BE.X - 1);
    end;
    Helper.Free;
  end;
end;

function TmwCustomEdit.GetText: String;
begin
  Result := Lines.Text;
end;

function TmwCustomEdit.GetTopLine: Integer;
begin
  Result := fTopLine + 1;
end;

procedure TmwCustomEdit.HideCaret;
begin
  if fCaretVisible then
  begin
    {$IFDEF MWE_FPC}
    if LCLLinux.HideCaret(Handle) then fCaretVisible := False;
    {$ELSE}
    if Windows.HideCaret(Handle) then fCaretVisible := False;
    {$ENDIF}
  end;
end;

procedure TmwCustomEdit.IncPaintLock;
begin
  inc(fPaintLock);
end;

procedure TmwCustomEdit.KeyDown(var Key: Word; Shift: TShiftState);
var
{begin}                                                                         //bds 12/20/1998
//  moveBkm: boolean;
  Data: pointer;
  C: char;
  Cmd: TmwEditorCommand;                                                        //bds 12/23/1998
{end}                                                                           //bds 12/20/1998
begin
  {$IFDEF DebugKeyEvents} OutputDebugString('>KeyDown'); try {$ENDIF}

  {$IFDEF MWE_FPC}
  inherited KeyDown(Key, Shift);
  {$ELSE}
  inherited;
  {$ENDIF}

  if Assigned(OnKeyDown) then
    OnKeyDown(Self, Key, Shift);

  {begin}                                                                       //bds 12/20/1998
  Data := NIL;
  C := #0;
  try
    Cmd := TranslateKeyCode(Key, Shift, Data);
    if Cmd <> ecNone then
    begin
      Key := 0; // eat it.
      CommandProcessor(Cmd, C, Data);
    end;
  finally
    if Data <> NIL then
      FreeMem(Data);
  end;
(*
  Case (ssCtrl in Shift) of
    True:
      begin
        Case Key of
          Ord('C'): CopyToClipboard;
          Ord('V'):
            if not ReadOnly then PasteFromClipboard;
          Ord('X'):
            if not ReadOnly then CutToClipboard;
          Ord('c'): CopyToClipboard;
          Ord('v'):
            if not ReadOnly then PasteFromClipboard;
          Ord('x'):
            if not ReadOnly then CutToClipboard;
          else if ((ssShift in Shift) and (Key in [48..57])
            and fBookMarkOpt.fEnableKeys) then                                  //jdj 12/12/1998
          begin
            if fBookMarks[Ord(Key) - 48].IsSet then begin
              moveBkm := (fBookMarks[Ord(Key) - 48].Y <> CaretY);               //gp 12/11/1998
              ClearBookMark(Ord(Key) - 48);
              if moveBkm then                                                   //gp 12/11/1998
                SetBookMark(Ord(Key) - 48, CaretX, CaretY);
            end
            else SetBookMark(Ord(Key) - 48, CaretX, CaretY);
            Exit;
          end
          else if (Key in [48..57])
            and fBookMarkOpt.fEnableKeys then                                   //jdj 12/12/1998
            begin                                                               //jdj 12/10/1998
              GotoBookMark(Key - 48);
              Exit;
            end;
        end;
      end;
  end;
  case Key of
    VK_UP:     ProcessVK_UP(Shift);
    VK_DOWN:   ProcessVK_DOWN(Shift);
    VK_NEXT:   ProcessVK_NEXT(Shift);
    VK_PRIOR:  ProcessVK_PRIOR(Shift);
    VK_END:    ProcessVK_END(Shift);
    VK_HOME:   ProcessVK_HOME(Shift);
    VK_LEFT:   ProcessVK_LEFT(Shift);
    VK_RIGHT:  ProcessVK_RIGHT(Shift);
    VK_INSERT: ProcessVK_INSERT(Shift);
    VK_DELETE: ProcessVK_DELETE(Shift);
    VK_BACK:   ProcessVK_BACK(Shift);
    VK_RETURN: ProcessVK_RETURN(Shift);
  end;
*)
{end}                                                                           //bds 12/20/1998

  {$IFDEF DebugKeyEvents} finally OutputDebugString('<KeyDown'); end; {$ENDIF}
end;

procedure TmwCustomEdit.Loaded;                                                 //jdj 12/12/1998
begin
  inherited Loaded;
  fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;
  UpdateScrollBars;
end;

procedure TmwCustomEdit.KeyPress(var Key: Char);
// Changes stop caret from going out of view when character width is less       // jdj 12/08/1998
// than 25
{begin}                                                                         //bds 12/20/1998
{
var
  Len: Integer;
  Temp, Helper: string;
  StartOfBlock : TPoint;                                                        //tskurz 12/10/1998
  oldScroll: boolean;                                                           //gp 12/16/1998
}
{end}                                                                           //bds 12/20/1998
begin
  {$IFDEF DebugKeyEvents} OutputDebugString('>KeyPress'); try {$ENDIF}
  try
    if Assigned(OnKeyPress) then
      OnKeyPress(Self, Key);

{begin}                                                                         //bds 12/20/1998
  CommandProcessor(ecChar, Key, NIL);
(*
    if Key < #32 then Exit;
    if ReadOnly then Exit;
    if SelAvail then
      begin
        FUndoList.AddChange(mwcrDelete,fBlockBegin,fBlockEnd,PChar(SelText));   //tskurz 12/10/1998
        StartOfBlock := fBlockBegin;                                            //tskurz 12/10/1998
        SetSelText(Key);
        FUndoList.AddChange(mwcrInsert,StartOfBlock,fBlockEnd,PChar(SelText));  //tskurz 12/10/1998
        Exit;
      end;
    Temp := LineText;
    Len := Length(Temp);
    if Len < CaretX then
      begin
        Helper := StringOfChar(' ', CaretX - Len);
        Temp := Temp + Helper;
        Helper := '';
      end;
    case fInserting of
      True:
        begin
          oldScroll := fScrollPastEOL;                                          //gp 12/16/1998
          fScrollPastEOL := true;                                               //gp 12/16/1998
          try                                                                   //gp 12/16/1998
            IncPaintLock;
            StartOfBlock := Point(CaretX,CaretY);                               //tskurz 12/10/1998
            Insert(Key, Temp, CaretX);
            CaretX := CaretX + 1;
            LineText := TrimRight(Temp); ;
            FUndoList.AddChange(mwcrInsert,StartOfBlock,Point(CaretX,CaretY),nil);
                                                                                //tskurz 12/10/1998
            Temp := '';
            if CaretX >= LeftChar + fCharsInWindow then                         //jdj 12/08/1998
              LeftChar := LeftChar + min(25, fCharsInWindow - 1);
            DecPaintLock;
          finally fScrollPastEOL := oldScroll; end;                             //gp 12/16/1998
        end;
      False:
        begin
          oldScroll := fScrollPastEOL;                                          //gp 12/16/1998
          fScrollPastEOL := true;                                               //gp 12/16/1998
          try                                                                   //gp 12/16/1998
            IncPaintLock;
            StartOfBlock := Point(CaretX,CaretY);                               //tskurz 12/10/1998
            Helper := Temp[CaretX];                                             //tskurz 12/10/1998
            Temp[CaretX] := Key;
            CaretX := CaretX + 1;
            LineText := TrimRight(Temp);
            FUndoList.AddChange(mwcrInsert,StartOfBlock,Point(CaretX,CaretY),PChar(Helper));
                                                                                //tskurz 12/10/1998
            Temp := '';
            if CaretX >= LeftChar + fCharsInWindow then                         //jdj 12/08/1998
              LeftChar := LeftChar + min(25, fCharsInWindow - 1);
            DecPaintLock;
          finally fScrollPastEOL := oldScroll; end;                             //gp 12/16/1998
        end;
    end;
*)
{end}                                                                           //bds 12/20/1998
  finally
  end;
  {$IFDEF DebugKeyEvents} finally OutputDebugString('<KeyPress'); end; {$ENDIF}
end;

function TmwCustomEdit.LeftSpaces(Line: String): Integer;
var
  Len: Integer;
begin
  Result := 0;
  Len := Length(Line);
  while Result < Len do
  begin
    if Line[Result + 1] > #32 then break;
    inc(Result);
  end;
end;

procedure TmwCustomEdit.LinesChanged(Sender: TObject);
begin
  if HandleAllocated then
  begin
    UpdateScrollBars;
    SetBlockBegin(CaretXY);                                                     //bds 12/20/1998
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.MouseDown(Button: TMouseButton; Shift: TShiftState;     //tk 12/09/1998
  X, Y: Integer);
var
  bWasSel: boolean;                                                             //ked 12/10/1998
  BlStart, BlEnd, Dwn: integer;
begin
  {$IFDEF DebugMouseEvents} OutputDebugString('>MouseDown'); try {$ENDIF}
  if (Button = mbRight) and (Shift = [ssRight]) and Assigned(PopupMenu)         //levin 1/6/1999
    then exit;
  if ssDouble in Shift then Exit;                                               //gp 12/27/1998 - ignore stray MouseDown on double click
  {begin}                                                                       //ked 12.10.1998
  if SelAvail then begin
    //remember selection state, as it will be cleared later
    bWasSel := true;
    fDragBlockBegin := BlockBegin;
    fDragBlockEnd := BlockEnd;
    fMouseDownX := X;
    fMouseDownY := Y;
  end else
    bWasSel := false;
  {end}                                                                         //ked 12/10/1998

  inherited MouseDown(Button, Shift, X, Y);
  MouseCapture := True;
  ComputeCaret(X, Y);
  //if mousedown occured in selected block then begin drag operation
  BlStart := CoorToIndex(fDragBlockBegin);                                      //ked 12/10/1998
  BlEnd := CoorToIndex(fDragBlockEnd);
  Dwn := CoorToIndex(CaretXY);                                                  //bds 12/20/1998
  fWaitForDragging := false;                                                    //gp 12/12/1998
  if bWasSel and (Dwn >= BlStart) and (Dwn < BlEnd)
    then fWaitForDragging := true                                               //gp 12/12/1998
    else begin
      if not fDblClicked then begin
        if ssShift in Shift then
          SetBlockEnd(CaretXY)                                                  //bds 12/20/1998
        else
          SetBlockBegin(CaretXY);                                               //bds 12/20/1998
      end;
    end;
//  fDblClicked := False;                                                       //gp 12/12/1998
  SetFocus;
  {$IFDEF DebugMouseEvents} finally OutputDebugString('<MouseDown'); end; {$ENDIF}
end;

procedure TmwCustomEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
//var
//threshold: integer;                                                           //gp 12/16/1998, //gp 12/24/1998 - removed
begin
  {$IFDEF DebugMouseMove} OutputDebugString('>MouseMove'); try {$ENDIF}
  inherited MouseMove(Shift, x, y);
  {begin}
  if MouseCapture and fWaitForDragging and not ReadOnly then begin              //bds 12/20/1998
    {begin}                                                                     //gp 12/16/1998
//  {$IFDEF MWE_COMPILER_4_UP}
//    threshold := Mouse.DragThreshold;
//  {$ELSE}
//    threshold := 5; // default in D4
//  {$ENDIF}
//  if (Abs(fMouseDownX - X) >= threshold) or
//     (Abs(fMouseDownY - Y) >= threshold) then begin
    {end}                                                                       //gp 12/16/1998
    {begin}                                                                     //gp 12/24/1998 - rewritten
    if (Abs(fMouseDownX - X) >= GetSystemMetrics(SM_CXDRAG)) or
       (Abs(fMouseDownY - Y) >= GetSystemMetrics(SM_CYDRAG)) then begin
    {end}                                                                       //gp 12/24/1998
      fWaitForDragging := false;
      {$IFDEF DebugDragEvents} OutputDebugString('BeginDrag'); {$ENDIF}
      BeginDrag(false);                                                         //ked 12/10/1998
    end;
  end
  {end}                                                                         //gp 12/12/1998
  else begin
    Case X > fGutterWidth of
      True: Cursor := crIBeam;
      False: Cursor := crArrow;
    end;
    if (ssLeft in Shift) and MouseCapture then                                  //gp 12/12/1998
    begin
      if fInEditRect then
        ComputeCaret(X, Y);
      SetBlockEnd(CaretXY);                                                     //bds 12/20/1998
      Application.ProcessMessages;
      while (X < fGutterWidth) and (MouseCapture) and not fInEditRect do
      begin
        if fElapsed then
        begin
          fElapsed := False;
          LeftChar := LeftChar - 1;
          CaretX := LeftChar;
          SetBlockEnd(CaretXY);                                                 //bds 12/20/1998
        end;
        Application.ProcessMessages;
      end;

      while (X > ClientWidth) and MouseCapture and not fInEditRect do
      begin
        if fElapsed then
        begin
          fElapsed := False;
          LeftChar := LeftChar + 1;
          CaretX := fLeftChar + fCharsInWindow;
          SetBlockEnd(CaretXY);                                                 //bds 12/20/1998
        end;
        Application.ProcessMessages;
      end;

      while (Y < 0) and MouseCapture and not fInEditRect do
      begin
        if Y < -10 then
        begin
          TopLine := TopLine - fLinesInWindow;
          CaretY := TopLine;
          SetBlockEnd(CaretXY);                                                 //bds 12/20/1998
          Application.ProcessMessages;
        end else
        begin
          TopLine := TopLine - 1;
          CaretY := TopLine;
          SetBlockEnd(CaretXY);                                                 //bds 12/20/1998
          Application.ProcessMessages;
        end;
      end;

      while (Y > ClientHeight) and MouseCapture and not fInEditRect do
      begin
        if Y > (ClientHeight + 10) then
        begin
          TopLine := TopLine + fLinesInWindow;
          CaretY := fTopLine + fLinesInWindow;
          SetBlockEnd(CaretXY);                                                 //bds 12/20/1998
          Application.ProcessMessages;
        end else
        begin
          TopLine := TopLine + 1;
          CaretY := fTopLine + fLinesInWindow;
          SetBlockEnd(CaretXY);                                                 //bds 12/20/1998
          Application.ProcessMessages;
        end;
      end;
    end;
  end;
  {$IFDEF DebugMouseMove} finally OutputDebugString('<MouseMove'); end; {$ENDIF}
end;

procedure TmwCustomEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  {$IFDEF DebugMouseEvents} OutputDebugString('>MouseUp'); try {$ENDIF}
  inherited MouseUp(Button, Shift, X, Y);
  if (Button = mbRight) and (Shift = [ssRight]) and Assigned(PopupMenu)         //levin 1/6/1999
    then exit;
  MouseCapture := False;
  {begin}                                                                       //gp 12/12/1998
  if fWaitForDragging and (not fDblClicked) then begin
    ComputeCaret(X,Y);
    SetBlockBegin(CaretXY);                                                     //bds 12/20/1998
    SetBlockEnd(CaretXY);                                                       //bds 12/20/1998
    fWaitForDragging := false;
  end;
  fDblClicked := False;
  {end}                                                                         //gp 12/12/1998
  {$IFDEF DebugMouseEvents} finally OutputDebugString('<MouseUp'); end; {$ENDIF}
end;

procedure TmwCustomEdit.Paint;
begin
  if PaintLock <> 0 then exit;
  PaintControl;
end;

procedure TmwCustomEdit.PaintControl;
//var                                                                           
//  TextBM: TBitMap;                                                            //WvdM 12/18/1998

  procedure PaintRightEdge;
  begin
    with fTextBM.Canvas do                                                      //WvdM 12/18/1998
    begin
      Pen.Color := clSilver;
      pen.Width := 1;
      MoveTo(fRightEdgeOffset, 0);
      LineTo(fRightEdgeOffset, ClientHeight);
    end;
  end;

begin
  {begin}                                                                       //WvdM 12/18/1998
//  TextBM := TBitmap.Create;
//  try
    with fTextBM do begin
      Canvas.Font.Assign(Font);
      // Canvas.Brush.Color := Color;                                           //mt 1/3/1999 same line twice
      Height := ClientHeight;
      Width := ClientWidth;
      Canvas.Brush.Color := Color;
      Canvas.FillRect(ClientRect);
      if Lines.Count > 0 then
        if Assigned(fHighLighter) then
          PaintHighLighted(fTextBM.Canvas) else                                 //bds 12/25/1998
          PaintText(fTextBM.Canvas);                                            //bds 12/25/1998
    end;
    PaintRightEdge;
    PaintGutter(fTextBM.Canvas);                                                //bds 12/25/1998
    if Assigned(fOnPaint) then                                                  //jdj 12/12/1998
      fOnPaint(Self,fTextBM.Canvas);                                          
    HideCaret;
    Canvas.CopyRect(ClientRect, fTextBM.Canvas, ClientRect);
    UpdateCaret;
//  finally
//    TextBM.Free;                                                              
//  end;
  {end}                                                                         //WvdM 12/18/1998
end;

procedure TmwCustomEdit.PaintGutter(TextCanvas: TCanvas);
var
  i: Integer;
  pass: integer;                                                                //gp 1/10/1999
begin
  with TextCanvas do
    begin
      Brush.Color := fGutterColor;
      FillRect(fGutterRect);
      Pen.Color := clBtnHighlight;
      pen.Width := 1;
      MoveTo(fGutterWidth - 2, 0);
      LineTo(fGutterWidth - 2, ClientHeight);
      Pen.Color := clBtnShadow;
      MoveTo(fGutterWidth - 1, 0);
      LineTo(fGutterWidth - 1, ClientHeight);
      Pen.Color := Color{clWindow};                                             //bds 12/30/1998
      pen.Width := 2;
      MoveTo(fGutterWidth + 1, 0);
      LineTo(fGutterWidth + 1, ClientHeight);
      Brush.Color := Color;
      (*                                                                        //CdeB 12/16/1998 - removed
//      if (fGutterWidth > fBookMarks[0].Glyph.Width + 1) then
      if (fGutterWidth > fBookMarks[0].Glyph.Width +
        fBookMarkOpt.LeftMargin + 1) then                                       //jdj 12/12/1998
        for i := 0 to 9 do                                                      //jdj 12/10/1998
          begin
            if fBookMarks[i].IsSet and
              (fBookMarks[i].Y >= TopLine) and
              (fBookMarks[i].Y <= (TopLine + fLinesInWindow)) and
              (fBookMarks[i].Y <= fLines.Count) and                             //jdj 12/11/1998
              (fBookMarkOpt.GlyphsVisible) then                                 //jdj 12/12/1998
                                                                                //jdj 12/12/1998
//            Draw(2, (fBookMarks[i].Y - 1) * fTextHeight - fTopLine * fTextHeight + 1, fBookMarks[i].Glyph)
              Draw(fBookMarkOpt.LeftMargin, (fBookMarks[i].Y - 1) * fTextHeight - fTopLine * fTextHeight + 1, fBookMarks[i].Glyph)
            else if (fBookMarks[i].Y > fLines.Count) then                       //jdj 12/11/1998
              fBookMarks[i].IsSet := False;
          end;
      *)                                                                        //CdeB 12/16/1998
      if BookMarkOptions.GlyphsVisible then begin                               //gp 12/30/1998
        FillChar(fGutterOffsets^,fLinesInWindow*SizeOf(integer),0);
        {begin}                                                                 //CdeB 12/16/1998, //gp 1/10/1999
        for pass := 1 to 2 do begin // first paint bookmarks, then marks
          for i := 0 to Marks.count-1 do begin
            {$IFDEF MWE_FPC}
            if (Marks.Items[i].Line >= TopLine) and (Marks.Items[i].Line < TopLine+LinesInWindow)
                and Marks.Items[i].Visible and
                (((pass = 1) and Marks.Items[i].IsBookmark) or ((pass = 2 ) and (not Marks.Items[i].IsBookmark))) then begin
              if Assigned(fBookmarkImages) and
                 (not Marks.Items[i].InternalImage) then begin
                if Marks.Items[i].ImageIndex <= fBookmarkImages.Count then begin
                  {$IFOPT R+}{$DEFINE RestoreR}{$R-}{$ELSE}{$UNDEF RestoreR}{$ENDIF}
                  { TODO: Check the Draw function because it takes one more parameter than
                    the Delphi equivilent Draw function }
                  fBookmarkImages.Draw(TextCanvas, fBookMarkOpt.LeftMargin+fGutterOffsets^[Marks.Items[i].Line-TopLine],
                    (Marks.Items[i].Line-TopLine)*fTextHeight, Marks.Items[i].ImageIndex, true);
                  Inc(fGutterOffsets^[Marks.Items[i].Line-TopLine],fBookMarkOpt.XOffset);
                  {$IFDEF RestoreR}{$UNDEF RestoreR}{$R+}{$ENDIF}
                end;
              end
              else begin
                if Marks.Items[i].ImageIndex in [0..9] then begin // sanity check
                  {$IFOPT R+}{$DEFINE RestoreR}{$R-}{$ELSE}{$UNDEF RestoreR}{$ENDIF}
                  TextCanvas.Draw(fBookMarkOpt.LeftMargin+fGutterOffsets^[Marks.Items[i].Line-TopLine],
                    (Marks.Items[i].Line-TopLine)*fTextHeight, fBookMarksGlyphs[Marks.Items[i].ImageIndex]);
                  Inc(fGutterOffsets^[Marks.Items[i].Line-TopLine],fBookMarkOpt.XOffset);
                  {$IFDEF RestoreR}{$UNDEF RestoreR}{$R+}{$ENDIF}
                end;
              end;
            end; // if
            {$ELSE}
            if (Marks[i].Line >= TopLine) and (Marks[i].Line < TopLine+LinesInWindow)
                and Marks[i].Visible and
                (((pass = 1) and Marks[i].IsBookmark) or ((pass = 2 ) and (not Marks[i].IsBookmark))) then begin
              if Assigned(fBookmarkImages) and
                 (not Marks[i].InternalImage) then begin
                if Marks[i].ImageIndex <= fBookmarkImages.Count then begin
                  {$IFOPT R+}{$DEFINE RestoreR}{$R-}{$ELSE}{$UNDEF RestoreR}{$ENDIF}
                  fBookmarkImages.Draw(TextCanvas, fBookMarkOpt.LeftMargin+fGutterOffsets^[Marks[i].Line-TopLine],
                    (Marks[i].Line-TopLine)*fTextHeight, Marks[i].ImageIndex);
                  Inc(fGutterOffsets^[Marks[i].Line-TopLine],fBookMarkOpt.XOffset);
                  {$IFDEF RestoreR}{$UNDEF RestoreR}{$R+}{$ENDIF}
                end;
              end
              else begin
                if Marks[i].ImageIndex in [0..9] then begin // sanity check
                  {$IFOPT R+}{$DEFINE RestoreR}{$R-}{$ELSE}{$UNDEF RestoreR}{$ENDIF}
                  TextCanvas.Draw(fBookMarkOpt.LeftMargin+fGutterOffsets^[Marks[i].Line-TopLine],
                    (Marks[i].Line-TopLine)*fTextHeight, fBookMarksGlyphs[Marks[i].ImageIndex]);
                  Inc(fGutterOffsets^[Marks[i].Line-TopLine],fBookMarkOpt.XOffset);
                  {$IFDEF RestoreR}{$UNDEF RestoreR}{$R+}{$ENDIF}
                end;
              end;
            end; // if
            {$ENDIF}
          end; // for i
        end; // for pass
        {end}                                                                   //CdeB 12/16/1998, //gp 1/10/1999
      end;
    end;
end;

// jdj 12/08/1998:
// Changes selected color to use FSelectedColor
// WvdM 12/18/1998: Speed optimization - rewrote most of this - caused lots of
//                  problems.
// mw 12/20/1998: Rewrote again to fix problems, caused by speed optimisation.
// mw 12/21/1998: only Curier New is a true fixed Pitch font.
// bds 12/27/1998:
//  Major rewrite to optimize for speed and simplicity, and added comments so
//  it will be easier to figure out what's going on in here.  The majority of
//  the increase in speed comes from using ExtTextOut to draw the text using an
//  array of distances between characters instead of using TextOut to draw each
//  character one at a time.  The array of distances is needed because we may
//  have bold, italics (or both) mixed with normal text.  Most fonts have
//  different character widths for the different styles, so we have to find one
//  width that will fit regardless of the style and use that.  That's what the
//  FETODist array (managed elsewhere) is for.
//  I could have made the code smaller (less lines), but I think it would
//  overly complicate it.  Since this is an open project with many people
//  working on the code, I thought it best to not do this so that others who may
//  have to work on it can figure out what it is up to more quickly.  In other
//  words, I've sacrificed a millisecond or two in performance in exchange for
//  more readable/maintainable code.
//  Note:  It would have been faster and simpler if I could have set the
//  background mode to transparent (SetBkMode) for the selected areas to prevent
//  the text from overpainting selection backgrounds, but it won't work.
//  TCanvas resets the background mode when you change the brush color.
//  Oh well, it's not the first time the VCL has stepped on my API toes.  :)
{begin}                                                                         //bds 12/27/1998
procedure TmwCustomEdit.PaintHighLighted(TextCanvas: TCanvas);
var
  YOffset, LineIndex, TokenPos: Integer;
  PaintOffset, StartIndex, StopIndex, TempIndex: integer;
  BB, BE: TPoint;
  BG, FG: TColor;
  Token: String;
  EntireLineSelected: boolean;
begin
////////////////////////////////////////////////////////////////////////////////
// We could speed things up some by calling GetClipBox and painting only the
//    rows that are in that rect.
// Some other optimization tricks:
//   * Highighter returned a PChar into the data instead of making a copy of the
//     token and returning it.
////////////////////////////////////////////////////////////////////////////////

  // Tell the highlighter component what canvas it will be using.  The
  // highlighter will use this to set the canvas' Brush.Color, Font.Color and
  // Font.Style properties whenever a token is retrieved.  That means we only
  // have to fool with setting colors when we draw the selection highlight stuff
  fHighLighter.SetCanvas(TextCanvas);

  // Offset the blocks to be zero-based.
  BB.X := BlockBegin.X - 1;
  BB.Y := BlockBegin.Y - 1;
  BE.X := BlockEnd.X - 1;
  BE.Y := BlockEnd.Y - 1;

  // The vertical position we will paint at.  Start at top and increment by
  // fTextHeight for each iteration of the loop (see bottom of for loop).
  YOffset := 0;

  // Tell highlighter component where we are starting from
  fHighLighter.SetRange(Lines.Objects[fTopLine]);

  // Loop from the top line in the window to the last line in the window.  Make
  // sure we don't go past the last line in the case where it is not at the very
  // bottom of the window.
  for LineIndex := fTopLine to Min(Lines.Count-1, fTopLine + fLinesInWindow) do
  begin
    // Tell highlighter what line we are working on.  This will get the first
    // token set up (set the correct colors and font style).
    fHighLighter.SetLine(Lines[LineIndex]);

    // See if the line is entirely within the selection block.  We check this
    // a couple of times so we store it in var for shorthand.
    EntireLineSelected := ((LineIndex > BB.Y) and (LineIndex < BE.Y)) or
       ((LineIndex = BB.Y) and (LineIndex < BE.Y) and (BB.X <= LeftChar));

//  TokenPos := 0;                                                              //gp 12/30/1998 - removed (Delphi claims that value is never used)
    Token := '';

    // If there is no selection area, or it is hidden, or the selection does
    // not include the current line at all, or the entire line is selected,
    // painting is relatively simple.
    if ((not SelAvail) or (HideSelection and not Self.Focused)) or
       ((LineIndex < BB.Y) or (LineIndex > BE.Y)) or EntireLineSelected then
    begin
      // Paint the selection highlight.  Not all fonts will completely fill in
      // entire rect, so we have to do it to make sure.
      if EntireLineSelected then
      begin
        // Paint the selection highlight for the rest of the line.
        TextCanvas.Brush.Color := FSelectedColor.Background;
        TextCanvas.FillRect(Rect(0, YOffset, ClientWidth, YOffset+fTextHeight));
      end;
      // Repeat until we get to the end of the line.
      while not fHighLighter.GetEol do
      begin
        // Get the current token (string) to be painted.
        Token := fHighLighter.GetToken;
        // Get it's index in the line string
        TokenPos := fHighLighter.GetTokenPos;
        // If this token begins outside of the viewing area, there's no reason
        // bother with any more of this line at all.
        if TokenPos > LeftChar + CharsInWindow + 1 then
          break;
        // If this token comes entirely before the viewing area, we don't need
        // to do anything with it.  Just get the next token and loop around.
        if TokenPos + Length(Token) < LeftChar then
        begin
          fHighlighter.Next;
          continue;
        end;
        if EntireLineSelected then
        begin
          // GetToken updates the canvas brush (background) and font color.  We
          // don't want our selection highlight background that we painted above
          // to get painted over with a new color, so we set the canvas' colors
          // back to the proper selection colors.
          TextCanvas.Brush.Color := FSelectedColor.Background;
          TextCanvas.Font.Color := FSelectedColor.Foreground;
        end;
        // Paint the token string
        ExtTextOut(TextCanvas.Handle, fTextOffset + TokenPos * fCharWidth,
           YOffset, 0, NIL, PChar(Token), Length(Token), @FETODist[0]);
        // Tell highlighter to find the next token. When it does, it will set
        // the correct colors and font style for that token.
        fHighLighter.Next;
      end;
    end else begin
      // We have both selected and non-selected areas on the line.  Painting is
      // more complicated because we have to draw in the selection highlight.
      // There are 4 different situations that we need to deal with:
      // 1) Selection contained entirely on the line,
      // 2) Selection starts on the line and ends on another line,
      // 3) Selection started on another line and ends on this one.

      if (LineIndex = BB.Y) and (BB.Y = BE.Y) then
      begin
        // 1) Selection contained entirely on the line.

        // Repeat until we get to the end of the line.
        while not fHighLighter.GetEol do
        begin
          // Get the current token (string) to be painted.
          Token := fHighLighter.GetToken;
          // Get it's index in the line string
          TokenPos := fHighLighter.GetTokenPos;
          // If this token begins outside of the viewing area, there's no reason
          // bother with any more of this line at all.
          if TokenPos > LeftChar + CharsInWindow + 1 then
            break;
          // If this token comes entirely before the viewing area, we don't need
          // to do anything with it.  Just get the next token and loop around.
          if TokenPos + Length(Token) < LeftChar then
          begin
            fHighlighter.Next;
            continue;
          end;
          // Figure out where we begin painting horizontally
          PaintOffset := fTextOffset + TokenPos * fCharWidth;
          // Start with the beginning of the token
          StartIndex := TokenPos;
          // Stop at the end of the token
          StopIndex := TokenPos + Length(Token);
          if StopIndex > StartIndex then // do we have anything?
          begin
            // Paint up to selection start.  That length will be the StopIndex
            // or the start of the selection mark, whichever is less.
            TempIndex := Min(StopIndex, BB.X) - StartIndex;
            if TempIndex > 0 then // Have anything to paint?
            begin
              // Paint the pre-selection text
              ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
                 PChar(Token), TempIndex, @FETODist[0]);
              // Update the horizontal offset where text is drawn to past the
              // text we just painted.
              Inc(PaintOffset, TempIndex * CharWidth);
              // Update the start index into the line text to skip past the
              // text we just painted.
              Inc(StartIndex, TempIndex);
            end;
          end;

          // Paint the selection text.  That length will be the StopIndex or the
          // end of the selection mark, whichever is less.
          TempIndex := Min(StopIndex, BE.X) - StartIndex;
          if TempIndex > 0 then // Have anything to paint?
          begin
            // Save the proper colors for this token because we'll need to reset
            // them if the selection ends before the token itself.
            BG := TextCanvas.Brush.Color;
            FG := TextCanvas.Font.Color;
            // Set the selection highlight colors
            TextCanvas.Brush.Color := FSelectedColor.Background;
            TextCanvas.Font.Color := FSelectedColor.Foreground;
            // Not all fonts will fill the rectangle completely, so we have to.
            TextCanvas.FillRect(Rect(PaintOffset, YOffset,
               PaintOffset + TempIndex * CharWidth, YOffset + fTextHeight));
            // Paint the selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Token[StartIndex-TokenPos+1], TempIndex, @FETODist[0]);
            // Restore the token's proper colors
            TextCanvas.Brush.Color := BG;
            TextCanvas.Font.Color := FG;
            // Update the horizontal offset where text is drawn to past the
            // text we just painted.
            Inc(PaintOffset, TempIndex * CharWidth);
            // Update the start index into the line text to skip past the
            // text we just painted.
            Inc(StartIndex, TempIndex);
          end;

          // Paint the post-selection text.  The length is whatever is left over
          TempIndex := StopIndex - StartIndex;
          if TempIndex > 0 then // Have anything to paint?
            // Paint post-selection text.
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Token[StartIndex-TokenPos+1], TempIndex, @FETODist[0]);

          // Tell highlighter to find the next token. When it does, it will set
          // the correct colors and font style for that token.
          fHighLighter.Next;
        end;
      end else if (LineIndex = BB.Y) and (BB.Y <> BE.Y) then
      begin
        // 2) Selection starts on the line and ends on another line.

        // Initialize in case we never get into the loop because we use it
        // afterwards, too.
        PaintOffset := fTextOffset + fLeftChar * fCharWidth;
        // Repeat until we get to the end of the line.
        while not fHighLighter.GetEol do
        begin
          // Get the current token (string) to be painted.
          Token := fHighLighter.GetToken;
          // Get it's index in the line string
          TokenPos := fHighLighter.GetTokenPos;
          // If this token begins outside of the viewing area, there's no reason
          // bother with any more of this line at all.
          if TokenPos > LeftChar + CharsInWindow + 1 then
            break;
          // If this token comes entirely before the viewing area, we don't need
          // to do anything with it.  Just get the next token and loop around.
          if TokenPos + Length(Token) < LeftChar then
          begin
            fHighlighter.Next;
            continue;
          end;
          // Figure out where we begin painting horizontally
          PaintOffset := fTextOffset + TokenPos * fCharWidth;
          // Start with the beginning of the token
          StartIndex := TokenPos;
          // Stop at the end of the token
          StopIndex := TokenPos + Length(Token);
          if StopIndex > StartIndex then // do we have anything?
          begin
            // Paint up to selection start.  That length will be the StopIndex
            // or the start of the selection mark, whichever is less.
            TempIndex := Min(StopIndex, BB.X) - StartIndex;
            if TempIndex > 0 then // Have anything to paint?
            begin
              // Paint pre-selection text
              ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
                 PChar(Token), TempIndex, @FETODist[0]);
              // Update the horizontal offset where text is drawn to past the
              // text we just painted.
              Inc(PaintOffset, TempIndex * CharWidth);
              // Update the start index into the line text to skip past the
              // text we just painted.
              Inc(StartIndex, TempIndex);
            end;
          end;

          // Paint the selection text.  We know the selection ends on a
          // different line, so whatever is left is selected.
          TempIndex := StopIndex - StartIndex;
          if TempIndex > 0 then // Have anything to paint?
          begin
            // Set the selection highlight colors. We don't have to save the
            // current token colors here because we know the rest of this line
            // is in the selection area and they won't be needed.
            TextCanvas.Brush.Color := FSelectedColor.Background;
            TextCanvas.Font.Color := FSelectedColor.Foreground;
            // Not all fonts will fill the rectangle completely, so we have to.
            TextCanvas.FillRect(Rect(PaintOffset, YOffset,
               PaintOffset + TempIndex * CharWidth, YOffset + fTextHeight));
            // Paint the selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Token[StartIndex-TokenPos+1], TempIndex, @FETODist[0]);
            // Update the horizontal offset where text is drawn to past the
            // text we just painted.
            Inc(PaintOffset, TempIndex * CharWidth);
          end;

          // Tell highlighter to find the next token. When it does, it will set
          // the correct colors and font style for that token.
          fHighLighter.Next;
        end;
        // Paint the selection highlight out to the end of the line.
        TextCanvas.Brush.Color := FSelectedColor.Background;
        TextCanvas.FillRect(Rect(PaintOffset, YOffset, ClientWidth,
           YOffset + fTextHeight));
      end else if (LineIndex = BE.Y) and (BB.Y <> BE.Y) then
      begin
        // 3) Selection started on another line and ends on this one.

        // Repeat until we get to the end of the line.
        while not fHighLighter.GetEol do
        begin
          // Get the current token (string) to be painted.
          Token := fHighLighter.GetToken;
          // Get it's index in the line string
          TokenPos := fHighLighter.GetTokenPos;
          // If this token begins outside of the viewing area, there's no reason
          // bother with any more of this line at all.
          if TokenPos > LeftChar + CharsInWindow + 1 then
            break;
          // If this token comes entirely before the viewing area, we don't need
          // to do anything with it.  Just get the next token and loop around.
          if TokenPos + Length(Token) < LeftChar then
          begin
            fHighlighter.Next;
            continue;
          end;
          // Figure out where we begin painting horizontally
          PaintOffset := fTextOffset + TokenPos * fCharWidth;
          // Start with the beginning of the token
          StartIndex := TokenPos;
          // Stop at the end of the token
          StopIndex := TokenPos + Length(Token);
          if StopIndex > StartIndex then // do we have anything?
          begin
            // Paint selection text.  The length is the StopIndex or the end of
            // the selection block, whichever is less.
            TempIndex := Min(StopIndex, BE.X) - StartIndex;
            if TempIndex > 0 then // Have anything to paint?
            begin
              // Save these because they were set up by the fHighter.SetLine
              // and/or fHighlighter.Next call.  We have to restore them when
              // we get out of the selected area.
              BG := TextCanvas.Brush.Color;
              FG := TextCanvas.Font.Color;
              // Set the selection highlight colors
              TextCanvas.Brush.Color := FSelectedColor.Background;
              TextCanvas.Font.Color := FSelectedColor.Foreground;
              // Not all fonts will fill the rectangle completely, so we have to
              TextCanvas.FillRect(Rect(PaintOffset, YOffset,
                 PaintOffset + TempIndex * CharWidth, YOffset + fTextHeight));
              // Paint the selection text
              ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0,
                 NIL, PChar(Token), TempIndex, @FETODist[0]);
              // Update the horizontal offset where text is drawn to past the
              // text we just painted.
              Inc(PaintOffset, TempIndex * CharWidth);
              // Update the start index into the line text to skip past the
              // text we just painted.
              Inc(StartIndex, TempIndex);
              // Restore the token's proper colors
              TextCanvas.Brush.Color := BG;
              TextCanvas.Font.Color := FG;
            end;
          end;

          // Paint the post-selection text.  The length is whatever is left over
          TempIndex := StopIndex - StartIndex;
          if TempIndex > 0 then // Have anything to paint?
            // Paint post-selection text.
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Token[StartIndex-TokenPos+1], TempIndex, @FETODist[0]);

          // Tell highlighter to find the next token. When it does, it will set
          // the correct colors and font style for that token.
          fHighLighter.Next;
        end;
      end;
    end;

    // This line has been painted.  Increment the vertical offset and loop back
    // for the next line of text.
    Inc(YOffset, fTextHeight);
  end;
  // Tell the highlighter we are done.
  fHighLighter.SetCanvas(NIL);
end;
{end}                                                                           //bds 12/27/1998


// jdj 12/08/1998:
//  Changes selected color to use FSelectedColor
// bds 12/27/1998:
//  Major rewrite to optimize for speed and simplicity, and added comments so
//  it will be easier to figure out what's going on in here.  The majority of
//  the increase in speed comes from using ExtTextOut to draw the text using an
//  array of distances between characters instead of using TextOut to draw each
//  character one at a time.  The array of distances is needed because we may
//  have bold, italics (or both) mixed with normal text.  Most fonts have
//  different character widths for the different styles, so we have to find one
//  width that will fit regardless of the style and use that.  That's what the
//  FETODist array (managed elsewhere) is for.
//  I could have made the code smaller (less lines), but I think it would
//  overly complicate it.  Since this is an open project with many people
//  working on the code, I thought it best to not do this so that others who may
//  have to work on it can figure out what it is up to more quickly.  In other
//  words, I've sacrificed a millisecond or two in performance in exchange for
//  more readable/maintainable code.
{begin}                                                                         //bds 12/28/1998
procedure TmwCustomEdit.PaintText(TextCanvas: TCanvas);
var
  YOffset, LineIndex: Integer;
  PaintOffset, LineLen, TempIndex, StartIndex, StopIndex: integer;
  BB, BE: TPoint;
begin
////////////////////////////////////////////////////////////////////////////////
// Could speed up by getting clip rect and doing only line/chars inside it.
////////////////////////////////////////////////////////////////////////////////

  if HideSelection and not Self.Focused then begin
    // Selection should not be painted, so lie about the begin/end points.
    BB := Point(-1, -1);
    BE := Point(-1, -1);
  end else begin
    // Offset the begin/end points so they are zero-based
    BB := Point(BlockBegin.X - 1,BlockBegin.Y - 1);
    BE := Point(BlockEnd.X - 1, BlockEnd.Y - 1);
  end;

  // The vertical position we will paint at.  Start at top and increment by
  // fTextHeight for each iteration of the loop (see bottom of for loop).
  YOffset := 0;
  // Loop from the top line in the window to the last line in the window.  Make
  // sure we don't go past the last line in the case where it is not at the very
  // bottom of the window.
  for LineIndex := fTopLine to Min(fTopLine + fLinesInWindow, Lines.Count-1) do
  begin
    // We use the length of the line in several places, so only make one call
    // to the length function.
    LineLen := Length(Lines[LineIndex]);
    // Make sure the line text isn't scrolled off screen horizontally
    if (LeftChar > LineLen) then
    begin
      // Entire text of the line has been scrolled off horizontally, but we may
      // still need to draw the selection highlight if the entire row is
      // withing the selected block.
      // We handle this seperately because we don't want to do any indexing into
      // the Lines[LineIndex] string since the index would be past the end of
      // the string.  That would give an access violation.
      // Pseudo code: if current line is after selection start line or before
      //   selection end line, or if current line is the selection start line
      //   and the selection is multiple lines then fill the line rectangle
      //   with selection color.
      if ((LineIndex > BB.Y) and (LineIndex < BE.Y)) or
         ((LineIndex = BB.Y) and (BB.Y <> BE.Y)) then
      begin
        // Entire line is selected, but there is no text (scrolled offscreen).
        // Just paint the highlight.
        TextCanvas.Brush.Color := FSelectedColor.Background;
        TextCanvas.FillRect(Rect(fGutterWidth, YOffset, ClientWidth,
           YOffset + fTextHeight));
      end;
    end else begin
      // We have text to draw on this line.
      // PaintOffset is where we draw the text horizontally.  Initially, it
      // skips the gutter and accounts for horizontal scrolling (LeftChar).
      PaintOffset := fTextOffset + fLeftChar * fCharWidth;

      // Always start with the leftmost visible character
      StartIndex := LeftChar;
      // Stop at the end of the string, or the rightmost visible character,
      // whichever is less.
      StopIndex := LeftChar + Min(CharsInWindow + 1, LineLen-LeftChar + 1) - 1;

      // Figure out the selection status of line.  There are 4 different
      // situations that we need to deal with:
      // 1) Entire line is selected, or has no selection at all,
      // 2) Selection contained entirely on the line,
      // 3) Selection starts on the line and ends on another line,
      // 4) Selection started on another line and ends on this one.
      if ((LineIndex > BB.Y) and (LineIndex < BE.Y)) or
         ((LineIndex < BB.Y) or (LineIndex > BE.Y)) or
         (not SelAvail) then
      begin
        // 1) Entire line is selected, or has no selection at all.
        // We just need to draw the highlight mark, or not, and then paint all
        // the line text at one time.
        if ((LineIndex > BB.Y) and (LineIndex < BE.Y)) then
        begin
          // Entire line is selected, draw selection highlight
          TextCanvas.Brush.Color := FSelectedColor.Background;
          TextCanvas.Font.Color := FSelectedColor.Foreground;
          TextCanvas.FillRect(Rect(fGutterWidth, YOffset, ClientWidth,
             YOffset + fTextHeight));
        end else begin
          // Line is before begining of selection block line, or after it.  Just
          // set the normal colors.
          TextCanvas.Brush.Color := Color;
          TextCanvas.Font.Color := Font.Color;
        end;
        // Draw the text starting with leftmost character in the window to the
        // end of the string, or to the rightmost visible character, whichever
        // is less.  Don't bother if there isn't anything to draw.
        if StopIndex >= StartIndex then
          ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
             @Lines[LineIndex][StartIndex], StopIndex-StartIndex+1, @FETODist[0]);
      end else if (LineIndex = BB.Y) and (LineIndex = BE.Y) then
      begin
        // 2) Selection contained entirely on the line.
        // We have three sections of text to draw: the text before the selection
        // start in normal colors, the selected text in highlight colors, and
        // the text after the end of the selection in normal colors again.

        // Draw the three sections of text
        if StopIndex >= StartIndex then // do we have anything?
        begin
          // Paint up to selection start.  That length will be the StopIndex or
          // the start of the selection mark, whichever is less.
          TempIndex := Min(StopIndex, BB.X) - StartIndex + 1;
          if TempIndex > 0 then  // Have anything to draw?
          begin
            TextCanvas.Brush.Color := Color;
            TextCanvas.Font.Color := Font.Color;
            // Draw the pre-selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Lines[LineIndex][StartIndex], TempIndex, @FETODist[0]);
            // Update the horizontal offset where text is drawn to past the
            // text we just painted.
            Inc(PaintOffset, TempIndex * CharWidth);
            // Update the start index into the line text to skip past the
            // text we just painted.
            Inc(StartIndex, TempIndex);
          end;

          // Paint selection text.  The length is the StopIndex or the end of
          // the selection block, whichever is less.
          TempIndex := Min(StopIndex, BE.X) - StartIndex + 1;
          if TempIndex > 0 then // Have anything to draw?
          begin
            TextCanvas.Brush.Color := FSelectedColor.Background;
            TextCanvas.Font.Color := FSelectedColor.Foreground;
            // Not all fonts will fill the rectangle completely, so we have to.
            TextCanvas.FillRect(Rect(PaintOffset, YOffset,
               PaintOffset + TempIndex * CharWidth, YOffset + fTextHeight));
            // Draw the selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Lines[LineIndex][StartIndex], TempIndex, @FETODist[0]);
            // Update the horizontal offset where text is drawn to past the
            // text we just painted.
            Inc(PaintOffset, TempIndex * CharWidth);
            // Update the start index into the line text to skip past the
            // text we just painted.
            Inc(StartIndex, TempIndex);
          end;

          // Paint after end of selection.  The length is whatever is left over
          // after the last two sections.
          TempIndex := StopIndex - StartIndex + 1;
          if TempIndex > 0 then // Have anything to draw?
          begin
            TextCanvas.Brush.Color := Color;
            TextCanvas.Font.Color := Font.Color;
            // Draw the post-selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Lines[LineIndex][StartIndex], TempIndex, @FETODist[0]);
          end;
        end;
      end else if (LineIndex = BB.Y) and (BB.Y <> BE.Y) then
      begin
        // 3) Selection starts on the line and ends on another line, i.e.
        // highlight area fills to visible end of line not just end of text.
        // We have three sections to draw: the text before the selection
        // start in normal colors, the selected text in highlight colors, and
        // the selection area without text to end of visible area in highlight
        // colors again.

        // Draw the three sections
        if StopIndex >= StartIndex then // do we have anything?
        begin
          // Paint up to selection start.  That length will be the StopIndex or
          // the start of the selection mark, whichever is less. 
          TempIndex := Min(StopIndex, BB.X) - StartIndex + 1;
          if TempIndex > 0 then // Have anything to draw?
          begin
            TextCanvas.Brush.Color := Color;
            TextCanvas.Font.Color := Font.Color;
            // Draw the pre-selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Lines[LineIndex][StartIndex], TempIndex, @FETODist[0]);
            // Update the horizontal offset where text is drawn to past the
            // text we just painted.
            Inc(PaintOffset, TempIndex * CharWidth);
            // Update the start index into the line text to skip past the
            // text we just painted.
            Inc(StartIndex, TempIndex);
          end;

          // Paint selection text.  The length is whatever is left over after
          // pre-selection text
          TempIndex := StopIndex - StartIndex + 1;
          if TempIndex > 0 then // Have anything to draw?
          begin
            TextCanvas.Brush.Color := FSelectedColor.Background;
            TextCanvas.Font.Color := FSelectedColor.Foreground;
            // Not all fonts will fill the rectangle completely, so we have to.
            TextCanvas.FillRect(Rect(PaintOffset, YOffset,
               PaintOffset + TempIndex * CharWidth, YOffset + fTextHeight));
            // Draw the selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Lines[LineIndex][StartIndex], TempIndex, @FETODist[0]);
            // Update the horizontal offset where text is drawn to past the
            // text we just painted.
            Inc(PaintOffset, TempIndex * CharWidth);
          end;
        end;

        // Fill the remaining visible area with the highlight colors
        TextCanvas.Brush.Color := FSelectedColor.Background;
        TextCanvas.FillRect(Rect(PaintOffset, YOffset, ClientWidth,
           YOffset + fTextHeight));
      end else if (LineIndex = BE.Y) and (BB.Y <> BE.Y) then
      begin
        // 4) Selection started on another line and ends on this one, i.e.
        // selection fills from beginning of line). We have two sections to
        // draw: the selected text at the beginning in highlight colors, and
        // the text following the selection area in normal colors.

        // Draw the two sections
        if StopIndex >= StartIndex then // do we have anything?
        begin
          // Paint selection text.  The length is the StopIndex or the end of
          // the selection block, whichever is less.
          TempIndex := Min(StopIndex, BE.X) - StartIndex + 1;
          if TempIndex > 0 then // Have anything to paint?
          begin
            TextCanvas.Brush.Color := FSelectedColor.Background;
            TextCanvas.Font.Color := FSelectedColor.Foreground;
            // Not all fonts will fill the rectangle completely, so we have to.
            TextCanvas.FillRect(Rect(PaintOffset, YOffset,
               PaintOffset + TempIndex * CharWidth, YOffset + fTextHeight));
            // Draw the selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Lines[LineIndex][StartIndex], TempIndex, @FETODist[0]);
            // Update the horizontal offset where text is drawn to past the
            // text we just painted.
            Inc(PaintOffset, TempIndex * CharWidth);
            // Update the start index into the line text to skip past the
            // text we just painted.
            Inc(StartIndex, TempIndex);
          end;

          // Paint after end of selection.  The length is whatever is left over
          // after the selection.
          TempIndex := StopIndex - StartIndex + 1;
          if TempIndex > 0 then // Have anything to paint?
          begin
            TextCanvas.Brush.Color := Color;
            TextCanvas.Font.Color := Font.Color;
            // Draw post-selection text
            ExtTextOut(TextCanvas.Handle, PaintOffset, YOffset, 0, NIL,
               @Lines[LineIndex][StartIndex], TempIndex, @FETODist[0]);
          end;
        end;
      end;
    end;

    // This line has been painted.  Increment the vertical offset and loop back
    // for the next line of text.
    Inc(YOffset, fTextHeight);
  end; { for LineIndex }
end;
{end}                                                                           //bds 12/28/1998

procedure TmwCustomEdit.PasteFromClipboard;
var
   StartOfBlock: TPoint;                                                        //tskurz 12/10/1998
begin
  if Clipboard.HasFormat(CF_TEXT) then
  begin
    if SelAvail then
      FUndoList.AddChange(mwcrDelete,fBlockBegin,fBlockEnd,PChar(GetSelText));  //tskurz 1/7/1999
    StartOfBlock := fBlockBegin;                                                //tskurz 12/10/1998
    SetSelText(Clipboard.AsText);                                               //tskurz 1/7/1999
    FUndoList.AddChange(mwcrInsert,StartOfBlock,fBlockBegin,PChar(GetSelText)); //tskurz 12/10/1998
  end;
end;

procedure TmwCustomEdit.SelectAll;
begin
  SetBlockBegin(Point(1, 1));
  if Lines.Count > 0 then
    SetBlockEnd(Point(Length(Lines[Lines.Count - 1]) + 1, Lines.Count{ - 1}));  //bds 12/20/1998
  CaretXY := Point(Length(Lines[Lines.Count-1])+1, Lines.Count);                //bds 12/20/1998
end;

procedure TmwCustomEdit.SetBlockBegin(Value: TPoint);
var
  NeedPaint: Boolean;
begin
  if Value.X < 1 then Value.X := 1 else
    if Value.X > fMaxLeftChar then Value.X := fMaxLeftChar;
  if Value.Y < 1 then Value.Y := 1 else
    if Value.Y > Lines.Count then Value.Y := Lines.Count;
  if SelAvail then NeedPaint := True else NeedPaint := False;
  fBlockBegin := Value;
  fBlockEnd := Value;
  if NeedPaint then Paint;
end;

procedure TmwCustomEdit.SetBlockEnd(Value: TPoint);
begin
  if Value.X < 1 then Value.X := 1 else
    if Value.X > fMaxLeftChar then Value.X := fMaxLeftChar;
  if Value.Y < 1 then Value.Y := 1 else
    if Value.Y > Lines.Count then Value.Y := Lines.Count;
  fBlockEnd := Value;
  if PaintLock = 0 then Paint;
end;

// jdj 12/09/1998:
// Smooths caret movement
procedure TmwCustomEdit.SetCaretX(Value: Integer);
var
  Len: integer;
begin
  if Value < 1 then Value := 1;
  {begin}                                                                       //bds 12/16/1998
//if Value > (fMaxLeftChar - fCharsInWindow) then Value := fMaxLeftChar - fCharsInWindow;
  if fScrollPastEOL then
  begin
    if Value > (fMaxLeftChar - fCharsInWindow) then
      Value := fMaxLeftChar - fCharsInWindow;
  end else begin
    {begin}                                                                     //gp 12/16/1998
    Len := Length(LineText) + 1;
    if Value > Len then
      Value := Length(LineText)+1;
    if Value >= (LeftChar + fCharsInWindow) then
      LeftChar := ((Value + 1) - fCharsInWindow)
    else if Value < LeftChar then LeftChar := Value;
    {end}                                                                       //gp 12/16/1998
  end;
  {end}                                                                         //bds 12/16/1998
  fCaretX := Value - 1;
  EnsureCursorPosVisible;                                                       //bds 12/17/1998
  if PaintLock = 0 then                                                         //jdj 12/08/1998
    UpdateCaret;
end;

procedure TmwCustomEdit.SetCaretY(Value: Integer);
var
  Len: integer;
begin
  if Value < 1 then Value := 1;
  if Value > Lines.Count then Value := Lines.Count;
  fCaretY := Value - 1;
  {begin}                                                                       //bds 12/16/1998
//if PaintLock = 0 then                                                         //jdj 12/11/1998
//  UpdateCaret;
  if not FScrollPastEOL then begin
    {begin}                                                                     //gp 12/16/1998
    Len := Length(LineText) + 1;
    if CaretX > Len then begin
      CaretX := Len;
      if CaretX < LeftChar then LeftChar := CaretX;                             //gp 12/16/1998
      exit;
    end;
    {end}                                                                       //gp 12/16/1998
  end;
  EnsureCursorPosVisible;                                                       //bds 12/17/1998
  if PaintLock = 0 then                                                         //jdj 12/11/1998
    UpdateCaret;
  {end}                                                                         //bds 12/16/1998
end;

procedure TmwCustomEdit.SetFont(const Value: TFont);
var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;
  AveCW, MaxCW: Integer;
  {$IFDEF MWE_FPC}
  TempFont: TFont;
  {$ENDIF}
begin
  DC := GetDC(0);
  Save := SelectObject(DC, Value.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, Save);
  ReleaseDC(0, DC);
  with Metrics do
  begin
    AveCW := tmAveCharWidth;
    MaxCW := tmMaxCharWidth;
  end;
  case AveCW = MaxCW of
	  {$IFDEF MWE_FPC}
	  // MWE: sometimes a direct assign works, sometimes not.
	  //      strange compiler bug. This indirect assignment works mostly
    True: 
    begin
    	TempFont := Value;
    	inherited Font := TempFont;
    end;
	  {$ELSE}
    True: inherited Font := Value;
	  {$ENDIF}
    False:
      begin
        with fFontDummy do
        begin
          Color := Value.Color;
          Pitch := fpFixed;
          Size := Value.Size;
          Style := Value.Style;
        end;
        inherited Font := fFontDummy;
      end;
  end;
end;

procedure TmwCustomEdit.SetGutterWidth(Value: Integer);
var
  OldValue: Integer;
begin
  OldValue := GutterWidth;
  if OldValue <> Value then
    begin
      if Value < 0 then
        Value := 0;
      fGutterWidth := Value;
      fGutterRect.Right := Value;
      fCharsInWindow := (ClientWidth - fGutterWidth - 2) div fCharWidth;
      fRightChar := fLeftChar + fCharsInWindow;
      fEditRect := ClientRect;
      fEditRect.Left := fGutterWidth;
      fTextOffset := fGutterWidth + 2;                                          //jdj 12/12/1998
      fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;                //mw 12/20/1998
      UpdateScrollBars;                                                         //jdj 12/12/1998
      if PaintLock = 0 then
        Paint;
    end;
end;

procedure TmwCustomEdit.SetLeftChar(Value: Integer);
var
  OldValue: Integer;
begin
  OldValue := LeftChar;
  if OldValue <> Value then
  begin
    if Value < 1 then Value := 1;
    if Value > fMaxLeftChar - fCharsInWindow then Value := fMaxLeftChar - fCharsInWindow;
    fLeftChar := Value - 1;
    fTextOffset := fGutterWidth + 2 - fLeftChar * fCharWidth;
    fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;
    UpdateScrollBars;
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.SetLines(Value: TStrings);
begin
  if HandleAllocated then
    Lines.Assign(Value);
end;

procedure TmwCustomEdit.SetLineText(Value: String);
begin
  if (Lines.Count > 0) and
    (fCaretY <= Lines.Count - 1) then Lines[fCaretY] := Value;
end;

// jdj 12/09/1998:
// Stops putting name in control when loading in IDE and Text is empty
procedure TmwCustomEdit.SetName(const Value: TComponentName);
var
  OldName: string;
begin
  OldName := Name;
  inherited SetName(Value);
//  if csDesigning in ComponentState then
  if (csDesigning in ComponentState) and                                        //jdj 12/09/1998
  (not (csLoading in ComponentState)) then
    if OldName = TrimRight(Text) then
      Text := Value;
end;

procedure TmwCustomEdit.SetScrollBars(const Value: TScrollStyle);
begin
  if (FScrollBars <> Value) then
  begin
    FScrollBars := Value;
    RecreateWnd;
    UpdateScrollBars;
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.SetSelText(Const Value: String);
var
  BB, BE: TPoint;
  TempString: String;
  Temp: TStringList;
  TempY: Integer;

  procedure DeleteSelection;
  var
    I, Count, Start: Integer;
    TempString2: String;
  begin
    IncPaintLock;
    CaretX := BB.X;
    if BB.Y = BE.Y then
    begin
      TempString := Lines[BB.Y - 1];
      Delete(TempString, BB.X, BE.X - BB.X);
      Lines[BB.Y - 1] := TrimRight(TempString);
      TempString := '';
    end else
    begin
      CaretY := BB.Y;
      TempString := Lines[BB.Y - 1];
      Delete(TempString, BB.X, Length(TempString));
      TempString2 := Lines[BE.Y - 1];
      Delete(TempString2, 1, BE.X - 1);
      TempString := TempString + TempString2;
      Start := BB.Y - 1;
      Count := BE.Y - BB.Y;
      {begin}                                                                   //CdeB 12/16/1998
      for i:= 0 to Marks.count-1 do
        {$IFDEF MWE_FPC}
        if Marks.Items[i].Line >= BE.Y then
          Marks.Items[i].Line := Marks.Items[i].Line - count
        else if Marks.Items[i].Line > BB.Y then
          Marks.Items[i].Line:= BB.Y;
        {$ELSE}
        if Marks[i].Line >= BE.Y then
          Marks[i].Line := Marks[i].Line - count
        else if Marks[i].Line > BB.Y then
          Marks[i].Line:= BB.Y;
        {$ENDIF}
      {end}                                                                     //CdeB 12/16/1998
      for I := 1 to Count do Lines.Delete(Start);
      Lines[Start] := TrimRight(TempString);
    end;
    DecPaintLock;
  end;

  procedure InsertText;
  var
    I, Len: Integer;
    TempString2, Helper: String;
    j: Integer;                                                                 //CdeB 12/16/1998
  begin
    if Value = '' then exit;
    {begin}                                                                     //CdeB 12/16/1998
    // Convert #10 into #10#13
    SetLength(Helper, Length(Value)*2);
    j:= 0;
    for i:= 1 to length(Value) do
      if ((i=1) and (Value[i]=#10)) or ((i>1) and (Value[i]=#10) and (Value[i-1]<>#13)) then
      Begin
        inc(j);
        Helper[j]:= #13;
        inc(j);
        Helper[j]:= Value[i];
      end else Begin
        inc(j);
        Helper[j]:= Value[i];
      end;
    SetLength(Helper, j);
    Temp.Text := Helper + #13#10;
    {end}                                                                       //CdeB 12/16/1998
    IncPaintLock;
    Temp.Text := Value + #13#10;
    if Temp.Count = 1 then
    begin
      TempString := LineText;
      Len := fCaretX - Length(TempString);
      if Len > 0 then
      begin
        Helper := StringOfChar(' ', Len);
        TempString := TempString + Helper;
      end;
      {$IFDEF MWE_FPC} 
      System.Insert(Temp[0], TempString, CaretX);
      {$ELSE}
      Insert(Temp[0], TempString, CaretX);
      {$ENDIF}
      LineText := TempString;
      CaretX := CaretX + Length(Temp[0]);
      TempString := '';
    end else
    begin
      TempString := LineText;
      Len := CaretX - Length(TempString);
      if Len > 0 then
      begin
        Helper := StringOfChar(' ', Len);
        TempString := TempString + Helper;
      end;
      TempString := Copy(TempString, 1, fCaretX) + Temp[0];
      TempString2 := Copy(LineText, CaretX, Length(LineText));
      Lines[fCaretY] := TempString;
      TempString := '';
      TempY := CaretY;
      for I := 1 to Temp.Count - 2 do
      begin
        Lines.Insert(TempY, Temp[I]);
        inc(TempY);
      end;
      {begin}                                                                   //CdeB 12/16/1998
      for i:= 0 to Marks.count-1 do
        {$IFDEF MWE_FPC}
        if Marks.Items[i].Line >= CaretY then
          Marks.Items[i].Line := Marks.Items[i].Line + Temp.Count - 1;
        {$ELSE}
        if Marks[i].Line >= CaretY then
          Marks[i].Line := Marks[i].Line + Temp.Count - 1;
        {$ENDIF}
      {end}                                                                     //CdeB 12/16/1998
      Lines.Insert(TempY, Temp[Temp.Count - 1] + TempString2);
      CaretX := Length(Temp[Temp.Count - 1]) + 1;
      inc(TempY);
      CaretY := TempY;
      TempString2 := '';
    end;
    DecPaintLock;
  end;
begin
  IncPaintLock;
  Lines.BeginUpdate;
  Temp := TStringList.Create;
  try
    BB := BlockBegin;
    BE := BlockEnd;
    if SelAvail then
    begin
      DeleteSelection;
      InsertText;
    end else InsertText;
  finally
    Temp.Free;
  end;
  Lines.EndUpdate;
  DecPaintLock;
end;

procedure TmwCustomEdit.SetText(const Value: String);
begin
  Lines.Text := Value;
  if PaintLock = 0 then Invalidate;
end;

procedure TmwCustomEdit.SetTopLine(Value: Integer);
var
  OldValue: Integer;
begin
  OldValue := TopLine;
  if OldValue <> Value then
  begin
    if Value > Lines.Count then Value := Lines.Count;
    if Value < 1 then Value := 1;
    fTopLine := Value - 1;
    UpdateScrollBars;
    if PaintLock = 0 then Paint;
  end;
end;

procedure TmwCustomEdit.ShowCaret;
begin
  {$IFDEF MWE_FPC}
  if LCLLinux.ShowCaret(Handle) then fCaretVisible := True;
  {$ELSE}
  if Windows.ShowCaret(Handle) then fCaretVisible := True;
  {$ENDIF}
end;

procedure TmwCustomEdit.TimerFired(Sender: TObject);
begin
  fElapsed := not fElapsed;
end;

procedure TmwCustomEdit.UpdateCaret;
var
  CX, CY: Integer;
begin
  if not Focused then Exit;                                                     //jdj 12/12/1998
  {begin}                                                                       //bds 12/17/1998
//CX := CaretXPix;
//CY := CaretYPix;
  CX := CaretXPix + FCaretOffset.X;
  CY := CaretYPix + FCaretOffset.Y;
  {end}                                                                         //bds 12/17/1998
  SetCaretPos(CX, CY);
  ShowCaret;
  if (CX < GutterWidth) or (CX > ClientWidth) or
    (CY < 0) or (CY > ClientHeight - fTextHeight) then
    HideCaret;
end;

procedure TmwCustomEdit.UpdateScrollBars;
var
  ScrollInfo: TScrollInfo;
begin
  if PaintLock <> 0 then exit;
  if fScrollBars in [ssBoth, ssVertical] then
  begin
    ScrollInfo.cbSize := SizeOf(ScrollInfo);
    ScrollInfo.fMask := SIF_ALL;
    ScrollInfo.nMin := 1;
    ScrollInfo.nMax := Lines.Count;
    ScrollInfo.nPage := fLinesInWindow;
    ScrollInfo.nPos := fTopLine;
    ScrollInfo.nTrackPos := 0;
    SetScrollInfo(Handle, SB_VERT, ScrollInfo, True);
    ShowScrollBar(Handle, SB_VERT, True);
  end;
  if fScrollBars in [ssBoth, ssHorizontal] then
  begin
    ScrollInfo.fMask := SIF_ALL;
    ScrollInfo.nMin := 1;
    {begin}                                                                     //bds 12/17/1998
    //This looks very wrong.....
    //ScrollInfo.nMax := 1024 - fLinesInWindow;
    //I'm not sure this is right, but it has to be closer... 
    ScrollInfo.nMax := fMaxLeftChar - fCharsInWindow;
    {end}                                                                       //bds 12/17/1998
    ScrollInfo.nPage := fCharsInWindow;
    ScrollInfo.nPos := fLeftChar;
    ScrollInfo.nTrackPos := 0;
    SetScrollInfo(Handle, SB_HORZ, ScrollInfo, True);
    ShowScrollBar(Handle, SB_HORZ, True);
  end;
end;

procedure TmwCustomEdit.WMEraseBkgnd(var Message: TMessage);
begin
  Message.Result := 1;
end;

procedure TmwCustomEdit.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  Msg.Result := DLGC_WANTARROWS;
  Msg.Result := Msg.Result or DLGC_WANTCHARS;
end;

procedure TmwCustomEdit.WMHScroll(var Message: TWMScroll);
var
  Delta: Integer;
begin
  Delta := 0;
  case Message.ScrollCode of
    SB_BOTTOM: Delta := fMaxLeftChar; //Scrolls to the lower right.
    SB_ENDSCROLL: Delta := 0; //Ends scroll.
    SB_LINELEFT: Delta := -1; //Scrolls left by one unit.
    SB_LINERIGHT: Delta := 1; //Scrolls right by one unit.
    SB_PAGELEFT: Delta := -fCharsInWindow; //Scrolls left by the width of the window.
    SB_PAGERIGHT: Delta := fCharsInWindow; //Scrolls right by the width of the window.
    SB_THUMBPOSITION: Delta := Message.Pos - LeftChar; //Scrolls to the absolute position. The current position is specified by the nPos parameter.
    SB_THUMBTRACK: Delta := Message.Pos - LeftChar; //Drags scroll box to the specified position. The current position is specified by the nPos parameter.
    SB_TOP: Delta := -fLeftChar; //Scrolls to the upper left.
  end;
  LeftChar := LeftChar + Delta;
end;

procedure TmwCustomEdit.WMKillFocus(var Message: TWMKillFocus);
begin
  {$IFDEF MWE_FPC}
    inherited WMKillFocus(Message);
  {$ELSE}
    inherited;
  {$ENDIF}
  HideCaret;
  {$IFDEF MWE_FPC}
    DestroyCaret;
  {$ELSE}
    Windows.DestroyCaret;
  {$ENDIF}
  if FHideSelection and SelAvail then Invalidate;                               //bds 12/16/1998
  //free memory bitmap
  fTextBM.Width := 1;                                                           //gp 12/30/1998, //gp 1/9/1999 - changed from 0 to 1 to prevent various problems
  fTextBM.Height := 1;                                                          //gp 12/30/1998, //gp 1/9/1999 - changed from 0 to 1 to prevent various problems
end;

procedure TmwCustomEdit.WMNCHitTest(var Message: TWMNCHitTest);
var
  aPoint: TPoint;
begin
  {$IFDEF MWE_FPC}
    {TODO: Create a WMNCHitTest method in the parent object }
  {$ELSE}
    inherited;
  {$ENDIF}
  aPoint := ScreenToClient(Point(Message.XPos, Message.YPos));
  if PtInRect(fEditRect, aPoint) then fInEditRect := True else
    fInEditRect := False;
end;

procedure TmwCustomEdit.WMSetFocus(var Message: TWMSetFocus);
begin
  {$IFDEF MWE_FPC}
    inherited WMSetFocus(Message);
  {$ELSE}
    inherited;
  {$ENDIF}
  {begin}                                                                       //bds 12/16/1998
//  CreateCaret(Handle, 0, 2, fTextHeight - 2);
//  UpDateCaret;
  InitializeCaret;
  if FHideSelection and SelAvail then
    Invalidate;
  {end}                                                                         //bds 12/16/1998
end;

procedure TmwCustomEdit.WMSize(var Message: TWMSize);
begin
  {$IFDEF MWE_FPC}
    Assert(False, Format('trace:[TmwCustomEdit.WMSize] %s --> Width: %d, Height: %d CLWidth:%d, CLHeight: %d', [ClassNAme, Message.Width, Message.Height, ClientWidth, ClientHeight]));
    Assert(False, Format('trace:[TmwCustomEdit.WMSize] fGutterWidth: %d, fCharWidth: %d, fTextHeight: %d', [fGutterWidth, fCharWidth, fTextHeight]));
    inherited WMSize(Message);
  {$ELSE}
    inherited;
  {$ENDIF}
  fCharsInWindow := (ClientWidth - fGutterWidth - 2) div fCharWidth;
  SetLinesInWindow((ClientHeight) div fTextHeight);                             //gp 1/10/1999 - changed to procedure call
  fBottomLine := fTopLine + fLinesInWindow;
  fRightChar := fLeftChar + fCharsInWindow;
  fGutterRect.Bottom := ClientHeight;
  fEditRect := ClientRect;
  fEditRect.Left := fGutterWidth;
{                                                                               //bds 12/27/1998
  with fLineRect do
  begin
    Left := fGutterWidth;
    Right := ClientWidth;
  end;
}                                                                               //bds 12/27/1998
end;

procedure TmwCustomEdit.WMVScroll(var Message: TWMScroll);
var
  Delta: Integer;
begin
  Delta := 0;
  case Message.ScrollCode of
    SB_BOTTOM: Delta := Lines.Count - TopLine; //Scrolls to the lower right.
    SB_ENDSCROLL: Delta := 0; //Ends scroll.
    SB_LINEDOWN: Delta := 1; //Scrolls one line down.
    SB_LINEUP: Delta := -1; //Scrolls one line up.
    SB_PAGEDOWN: Delta := fLinesInWindow; //Scrolls one page down.
    SB_PAGEUP: Delta := -fLinesInWindow; //Scrolls one page up.
    SB_THUMBPOSITION: Delta := Message.Pos - TopLine; //Scrolls to the absolute position. The current position is specified by the nPos parameter.
    SB_THUMBTRACK: Delta := Message.Pos - TopLine; //Drags scroll box to the specified position. The current position is specified by the nPos parameter.
    SB_TOP: Delta := -TopLine; // Scrolls to the upper left.
  end;
  TopLine := TopLine + Delta;
end;

procedure TmwCustomEdit.ScanFrom(Index: Integer);
var
  I: Integer;
begin
  if Index >= Lines.Count - 1 then exit;
  I := Index;
  fHighLighter.SetLine(Lines[I]);

  inc(I);
  while not fHighLighter.GetEol do fHighLighter.Next;
  {$IFDEF MWE_FPC}
  while fHighLighter.GetRange <> Pointer(fLines.Objects[I]) do
  {$ELSE}
  while fHighLighter.GetRange <> fLines.Objects[I] do
  {$ENDIF}
  begin
    {$IFDEF MWE_FPC}
    Lines.Objects[I] := TObject(fHighLighter.GetRange);
    {$ELSE}
    Lines.Objects[I] := fHighLighter.GetRange;
    {$ENDIF}
    fHighLighter.SetLine(Lines[I]);
    while not fHighLighter.GetEol do fHighLighter.Next;
    inc(I);
    if I = Lines.Count then break;
  end;
end;

procedure TmwCustomEdit.ListAdded(Sender: TObject);
var
  LastIndex: Integer;
begin
  if Assigned(fHighLighter) then
  begin
    if Lines.Count > 1 then
    begin
      LastIndex := Lines.Count - 1;
      fHighLighter.SetRange(Lines.Objects[LastIndex - 1]);
      fHighLighter.SetLine(Lines[LastIndex - 1]);
      while not fHighLighter.GetEol do fHighLighter.Next;
      {$IFDEF MWE_FPC}
      Lines.Objects[LastIndex] := TObject(fHighLighter.GetRange);
      {$ELSE}
      Lines.Objects[LastIndex] := fHighLighter.GetRange;
      {$ENDIF}
    end else
    begin
      fHighLighter.ReSetRange;
      {$IFDEF MWE_FPC}
      Lines.Objects[0] := TObject(fHighLighter.GetRange);
      {$ELSE}
      Lines.Objects[0] := fHighLighter.GetRange;
      {$ENDIF}
    end;
  end;
end;

procedure TmwCustomEdit.ListDeleted(Index: Integer);
var
  OldPaintLock: Integer;                                                        //jdj 12/11/1998
begin
  if Assigned(fHighLighter) then
    begin
      if Lines.Count > 1 then
        begin
          if Index > 0 then
            begin
              fHighLighter.SetRange(Lines.Objects[Index - 1]);
              ScanFrom(Index - 1);
            end
          else
            begin
              fHighLighter.ReSetRange;
              {$IFDEF MWE_FPC}
              Lines.Objects[0] := TObject(fHighLighter.GetRange);
              {$ELSE}
              Lines.Objects[0] := fHighLighter.GetRange;
              {$ENDIF}
              ScanFrom(0);
            end;
        end
      else if Lines.Count = 1 then
        begin
          fHighLighter.ReSetRange;
          {$IFDEF MWE_FPC}
          Lines.Objects[0] := TObject(fHighLighter.GetRange);
          {$ELSE}
          Lines.Objects[0] := fHighLighter.GetRange;
          {$ENDIF}
        end;
    end;
  {begin}                                                                       //jdj 12/11/1998
  OldPaintLock := fPaintLock;
  fPaintLock := 0;
  UpdateScrollBars;
  fPaintLock := OldPaintLock;
  {end}                                                                         //jdj 12/11/1998
end;

procedure TmwCustomEdit.ListInserted(Index: Integer);
var
  OldPaintLock: Integer;                                                        //jdj 12/11/1998
begin
  if Assigned(fHighLighter) then
    begin
      if Lines.Count > 1 then
        begin
          if Index > 0 then
            begin
              fHighLighter.SetRange(Lines.Objects[Index - 1]);
              ScanFrom(Index - 1);
            end
          else
            begin
              fHighLighter.ReSetRange;
              {$IFDEF MWE_FPC}
              Lines.Objects[0] := TObject(fHighLighter.GetRange);
              {$ELSE}
              Lines.Objects[0] := fHighLighter.GetRange;
              {$ENDIF}
              ScanFrom(0);
            end;
        end
      else if Lines.Count = 1 then
        begin
          fHighLighter.ReSetRange;
          {$IFDEF MWE_FPC}
          Lines.Objects[0] := TObject(fHighLighter.GetRange);
          {$ELSE}
          Lines.Objects[0] := fHighLighter.GetRange;
          {$ENDIF}
        end;
    end;
  {begin}                                                                       //jdj 12/11/1998
  OldPaintLock := fPaintLock;
  fPaintLock := 0;
  UpdateScrollBars;
  fPaintLock := OldPaintLock;
  {end}                                                                         //jdj 12/11/1998
end;

procedure TmwCustomEdit.ListLoaded(Sender: TObject);
var
  i: Integer;
begin
  {begin}                                                                       //jdj 12/11/1998
//for i := 0 to 9 do                                                            //CdeB 12/16/1998 - removed
//  fBookMarks[i].IsSet := False;                                               //CdeB 12/16/1998 - removed
  CaretXY := Point(1,1);                                                        //gp 12/30/1998
  LeftChar := 1;
  TopLine := 1;
  for i := 0 to fUndoList.FList.Count - 1 do
    fUndoList.RemoveChange(i);
  {end}                                                                         //jdj 12/11/1998
  // commented out - I don't know why this is here                              //gp 12/13/1998
  //  if Assigned(fHighLighter) then
  //    begin
  //
  //    end;
end;

// mw 12/17/1998 - rewritten
procedure TmwCustomEdit.ListPutted(Index: Integer);
begin
  if Assigned(fHighLighter) then
  begin
    fHighLighter.SetRange(Lines.Objects[Index]);
    ScanFrom(Index);
  end;
end;

procedure TmwCustomEdit.SetWordBlock(Value: TPoint);
var
  Runner: TPoint;
  TempString: String;
  IdChars: TIdentChars;                                                         //mt 12/22/1998
begin                                                                           //tskurz 12/14/1998 - lots of changes
  if Value.X < 1 then Value.X := 1 else
    if Value.X > fMaxLeftChar then Value.X := fMaxLeftChar;
  if Value.Y < 1 then Value.Y := 1 else
    if Value.Y > Lines.Count then Value.Y := Lines.Count;
  TempString := Lines[Value.Y - 1];
  if TempString = '' then exit;

  // Click on right side of text
  if Length(TempString) < Value.X then Value.X := Length(TempString);
  Runner := Value;

  {begin}                                                                       //mt 12/23/1998
  if fHighlighter <> NIL then
    IdChars := fHighlighter.IdentChars
  else
    IDchars := [#33..#255];
  {end}                                                                         //mt 12/23/1998

//  if not (TempString[Runner.X] in ALPHA+DIGIT) then break;                    //mt 12/10/1998
// Add both ALPHA parts because we had to break them up.
// replaced ALPHA_UC+ALPHA_LC+DIGIT constants with IdChars throughout           //mt 12/23/1998
  if not (TempString[Runner.X] in IdChars) then
  begin
    // no word under cursor and next char right is not start of a word
    if (Runner.X > 1) and (not (TempString[Runner.X] in IdChars)) then          //mt 12/23/1998
    begin
      // find end of word on the left side
      while Runner.X > 0 do
      begin
        if (TempString[Runner.X] in IdChars) then break;                        //mt 12/23/1998
        Dec(Runner.X);
      end;
    end;
    // no word on the left side, so look to the right side
    if not (TempString[Runner.X] in IdChars) then                               //mt 12/23/1998
    begin
      Runner := Value;
      while Runner.X < fMaxLeftChar do
      begin
        if (TempString[Runner.X] in IdChars) then break;                        //mt 12/23/1998
        Inc(Runner.X);
      end;
      if Runner.X > fMaxLeftChar then
        exit;
    end;
    Value := Runner;
  end;

  while Runner.X > 0 do
  begin
//  if not (TempString[Runner.X] in ALPHA+DIGIT) then break;                    //mt 12/10/1998
// Add both ALPHA parts because we had to break them up.
    if not (TempString[Runner.X] in IdChars) then break;                        //mt 12/23/1998
    Dec(Runner.X);
  end;
  Inc(Runner.X);
  if Runner.X < 1 then Runner.X := 1;
  fBlockBegin := Runner;

  Runner := Value;
  while Runner.X < fMaxLeftChar do
  begin
//  if not (TempString[Runner.X] in ALPHA+DIGIT) then break;                    //mt 12/10/1998
// Add both ALPHA parts because we had to break them up.
    if not (TempString[Runner.X] in IdChars) then break;                        //mt 12/23/1998
    Inc(Runner.X);
  end;
  if Runner.X > fMaxLeftChar then Runner.X := fMaxLeftChar;
  fBlockEnd := Runner;
// set caret to the end of selected block
  CaretXY := Runner;                                                            //gp 12/27/1998

  if PaintLock = 0 then Paint;
end;

procedure TmwCustomEdit.DblClick;
begin
  {$IFDEF DebugMouseEvents} OutputDebugString('>DblClick'); try {$ENDIF}
   SetWordBlock(CaretXY);                                                       //bds 12/20/1998
  {$IFDEF MWE_FPC}
   inherited DblClick;
  {$ELSE}
   inherited;
  {$ENDIF}
   fDblClicked := True;
  {$IFDEF DebugMouseEvents} finally OutputDebugString('<DblClick'); end; {$ENDIF}
end;

function TmwCustomEdit.GetCanUndo: Boolean;
begin
   result := (fUndoList.CanUndo > 0);
end;

function TmwCustomEdit.GetCanRedo: Boolean;
begin
   result := (fRedoList.CanUndo > 0);
end;

{begin}                                                                         //tskurz 1/4/1999
procedure TmwCustomEdit.Redo;
var
  ChangeStr : PChar;
  ChangeStartPos,
  ChangeEndPos: TPoint;
  ChangeReason : TChangeReason;
begin
  if not CanRedo then exit;
  ChangeReason := FRedoList.GetChange(ChangeStartPos,ChangeEndPos,ChangeStr);
  case ChangeReason of
    mwcrInsert,
    mwcrDragDropInsert :
      begin
        IncPaintLock;
        CaretXY := ChangeStartPos;
        SetBlockBegin(ChangeStartPos);
        DecPaintLock;
        SetSelText(ChangeStr);                                                  //tskurz 1/7/1999
        FUndoList.AddChange(ChangeReason,ChangeStartPos,
           fBlockEnd,PChar(GetSelText));                                        //tskurz 1/7/1999
        StrDispose(ChangeStr);
      end;
{begin}                                                                         //bds 12/20/1998
    mwcrDeleteAfterCursor:
      begin
        IncPaintLock;
        CaretXY := ChangeStartPos;
        SetBlockBegin(ChangeStartPos);
        SetBlockEnd(ChangeEndPos);
        DecPaintLock;
        FUndoList.AddChange(ChangeReason,fBlockBegin,fBlockBegin,               //tskurz 1/7/1999
          PChar(GetSelText));
        SetSelText(ChangeStr);                                                  //tskurz 1/7/1999
        StrDispose(ChangeStr);
        CaretXY := ChangeStartPos;
      end;
{end}
    mwcrDelete,
    mwcrDragDropDelete :
      begin
        IncPaintLock;
        SetBlockBegin(ChangeStartPos);
        SetBlockEnd(ChangeEndPos);
        DecPaintLock;
        FUndoList.AddChange(ChangeReason,fBlockBegin,fBlockEnd,                 //tskurz 1/7/1999
          PChar(GetSelText));
        SetSelText(ChangeStr);                                                  //tskurz 1/7/1999
        StrDispose(ChangeStr);
        CaretXY := ChangeStartPos;
        if FRedoList.GetChangeReason = mwcrDragDropInsert then Redo;
      end;
    mwcrLineBreak :
      begin
        CommandProcessor(ecLineBreak, #13, nil);
      end;
  else
  end;
end;
{end}                                                                           //tskurz 1/4/1999

procedure TmwCustomEdit.Undo;
var
  ChangeStr : PChar;
  ChangeStartPos,
  ChangeEndPos: TPoint;
  ChangeReason : TChangeReason;
  Temp: String;
begin
  if not CanUndo then exit;
  ChangeReason := FUndoList.GetChange(ChangeStartPos,ChangeEndPos,ChangeStr);
  case ChangeReason of
    mwcrInsert :
      begin
        IncPaintLock;
        SetBlockBegin(ChangeStartPos);
        SetBlockEnd(ChangeEndPos);
        DecPaintLock;
        FRedoList.AddChange(ChangeReason,fBlockBegin,fBlockEnd,                 //tskurz 1/7/1999
          PChar(GetSelText));
        SetSelText(ChangeStr);                                                  //tskurz 1/7/1999
        StrDispose(ChangeStr);
        CaretXY := ChangeStartPos;                                              //bds 12/20/1998
      end;
    mwcrDragDropInsert :                                                        //tskurz 12/16/1998
      begin
        IncPaintLock;
        SetBlockBegin(ChangeStartPos);
        SetBlockEnd(ChangeEndPos);
        DecPaintLock;
        FRedoList.AddChange(ChangeReason,fBlockBegin,fBlockEnd,                 //tskurz 1/7/1999
          PChar(GetSelText));
        SetSelText(ChangeStr);                                                  //tskurz 1/7/1999
        StrDispose(ChangeStr);
        CaretXY := ChangeStartPos;                                              //bds 12/20/1998
        if FUndoList.GetChangeReason = mwcrDragDropDelete then Undo;
      end;
{begin}                                                                         //bds 12/20/1998
    mwcrDeleteAfterCursor:
      begin
        IncPaintLock;
        CaretXY := ChangeStartPos;                                              //bds 12/20/1998
        SetBlockBegin(ChangeStartPos);
        DecPaintLock;
        SetSelText(ChangeStr);                                                  //tskurz 1/7/1999
        FRedoList.AddChange(ChangeReason,ChangeStartPos,fBlockEnd,              //tskurz 1/7/1999
          PChar(GetSelText));
        StrDispose(ChangeStr);
        CaretXY := ChangeEndPos;
      end;
{end}                                                                           //bds 12/20/1998
    mwcrDelete,
    mwcrDragDropDelete :  //tskurz 12/16/1998
      begin
        IncPaintLock;
        // If there's no selection, we have to set
        // the Caret's position manualy.
        CaretXY := ChangeStartPos;                                              //bds 12/20/1998
        SetBlockBegin(ChangeStartPos);
        DecPaintLock;
        SetSelText(ChangeStr);                                                  //tskurz 1/7/1999
        FRedoList.AddChange(ChangeReason,ChangeStartPos,fBlockEnd,              //tskurz 1/7/1999
          PChar(GetSelText));
        StrDispose(ChangeStr);
      end;
    mwcrLineBreak:
      begin
        IncPaintLock;
        // If there's no selection, we have to set
        // the Caret's position manualy.
        CaretXY := ChangeStartPos;                                              //gp 12/22/1998
        FRedoList.AddChange(ChangeReason,fBlockBegin,fBlockEnd,nil);            //tskurz 1/4/1999
        if CaretY > 0 then
        begin
//        Temp := Lines.Strings[fCaretY] + ChangeStr;                           //tskurz 12/21/1998
          Temp := Lines.Strings[fCaretY];                                       //tskurz 12/21/1998
          if (Length(Temp) < fCaretX) and (LeftSpaces(ChangeStr) = 0) then      //tskurz 1/4/1999
             Temp := Temp + StringOfChar(' ', fCaretX-Length(Temp));            //tskurz 1/4/1999
          Lines.Delete(ChangeEndPos.y);
        end;
        CaretXY := ChangeStartPos;                                              //bds 12/20/1998
        LineText := TrimRight(Temp+ChangeStr);                                  //gp 12/24/1998
        //CaretX := ChangeStartPos.x;
        DecPaintLock;
        StrDispose(ChangeStr);
      end;
  else
  end;
end;
{end}                                                                           //tskurz 12/10/1998

{begin}                                                                         //jdj 12/10/1998
                                                                                //CdeB 12/16/1998 - removed, //gp 12/31/1998 - reinstantianted
procedure TmwCustomEdit.ClearBookMark(BookMark: Integer);
begin
(*                                                                              //gp 12/31/1998 - removed
  if (BookMark in [0..9]) then
    begin
      fBookMarks[BookMark].IsSet := False;
      Invalidate;
    end;
*)
  {begin}                                                                       //CdeB 12/16/1998
  if (BookMark in [0..9]) and assigned(fBookMarks[BookMark]) then begin         //gp 12/31/1998
    FMarkList.Remove(fBookMarks[Bookmark]);
    fBookMarks[BookMark].Free;
    fBookMarks[BookMark] := nil;
  end
  {end}                                                                         //CdeB 12/16/1998
end;

procedure TmwCustomEdit.GotoBookMark(BookMark: Integer);
begin
(*                                                                              //gp 12/31/1998 - removed
  if (BookMark in [0..9]) and
    fBookMarks[BookMark].IsSet and
    (fBookMarks[BookMark].Y <= fLines.Count) then
    begin
      if (fBookMarks[BookMark].Y < TopLine) or
        (fBookMarks[BookMark].Y >= (TopLine + fLinesInWindow)) then
        TopLine := fBookMarks[BookMark].Y - (fLinesInWindow div 2);
//      CaretX := fBookMarks[BookMark].X;                                       //bds 12/20/1998
      if (fBookMarks[BookMark].X < LeftChar) or
        (fBookMarks[BookMark].X >= LeftChar + fCharsInWindow) then
        LeftChar := fBookMarks[BookMark].X - (fCharsInWindow div 2);
//      CaretY := fBookMarks[BookMark].Y;                                       //bds 12/20/1998
      CaretXY := Point(fBookMarks[BookMark].X, fBookmarks[BookMark].Y);         //bds 12/20/1998
    end;
*)                                                                              //gp 12/31/1998
  {begin}                                                                       //CdeB 12/16/1998
  if (BookMark in [0..9]) and assigned(fBookMarks[BookMark]) and                //gp 12/31/1998
     (fBookMarks[BookMark].Line <= fLines.Count) then begin
    CaretXY := Point(fBookMarks[BookMark].Column,fBookMarks[BookMark].Line);
    EnsureCursorPosVisible;
  end;
  {end}                                                                         //CdeB 12/16/1998
end;

procedure TmwCustomEdit.SetBookMark(BookMark: Integer; X: Integer; Y: Integer);
var
  i: Integer;
  mark: TMark;                                                                  //gp 1/9/1999
begin
(*                                                                              //gp 12/31/1998 - removed
  if (BookMark in [0..9]) and (Y <= fLines.Count) then
    begin
      for i := 0 to 9 do
        if (fBookMarks[i].Y = Y) then
          fBookMarks[i].IsSet := False;
      fBookMarks[BookMark].IsSet := True;
      fBookMarks[BookMark].X := X;
      fBookMarks[BookMark].Y := Y;
      Invalidate;
    end;
*)                                                                              //gp 12/31/1998
  {begin}                                                                       //CdeB 12/16/1998, //gp 1/9/1999 - changed a lot
  if (BookMark in [0..9]) and (Y <= fLines.Count) then begin
    mark := TMark.Create(self);
    with mark do begin
      Line := Y;
      Column := X;
      ImageIndex := Bookmark;
      BookmarkNumber := Bookmark;
      Visible := true;
      InternalImage := (BookmarkImages = nil);
    end;
    if Assigned(FOnPlaceMark) then FOnPlaceMark(mark);
    if (mark <> nil) and (BookMark in [0..9]) then begin                        
      for i := 0 to 9 do                                                        
        if assigned(fBookMarks[i]) and (fBookMarks[i].Line = Y) then               
          ClearBookmark(i);                                                     
      if assigned(fBookMarks[BookMark]) then ClearBookmark(BookMark);           
      fBookMarks[BookMark]:= mark;
      FMarkList.Add(fBookMarks[BookMark]);
    end;
  end;
  {end}                                                                         //CdeB 12/16/1998, //gp 1/9/1999
end;
{end}                                                                           //jdj 12/10/1998

procedure TmwCustomEdit.WndProc(var msg: TMessage);                             //gp 12/12/1998
// Prevent Alt-Backspace from beeping and wire it to the Undo at the same time.
begin
  if (msg.Msg = WM_SYSCHAR) AND (msg.wParam = VK_BACK) and
     ((msg.lParam AND $20000000) = $20000000) then begin
// Now just prevents beeping.  Translation to editor command handled elsewhere. //bds 12/20/1998
//    Undo;                                                                     //bds 12/20/1998
    msg.Msg := 0;
  end
  else 
  {$IFDEF MWE_FPC}
  inherited WndProc(msg);
  {$ELSE}
  inherited;
  {$ENDIF}
end;

{begin}                                                                         //ked 12/10/1998
function TmwCustomEdit.CoorToIndex(CPos: TPoint): integer;
begin
  result := CharsInWindow * CPos.Y + CPos.X;
end;

function TmwCustomEdit.IndexToCoor(ind: integer): TPoint;       
begin
  result.y := ind div CharsInWindow;
  result.x := ind - result.y * CharsInWindow;
end;

procedure TmwCustomEdit.DragOver(Source: TObject; X, Y: Integer;
                             State: TDragState; var Accept: Boolean);
begin
  {$IFDEF DebugDragEvents} try OutputDebugString(PChar(Format('>DragOver %p %d %d %d',[pointer(Source),x,y,Ord(State)]))); {$ENDIF}
  {$IFDEF MWE_FPC}
  inherited DragOver(Source, X, Y, State, Accept);
  {$ELSE}
  inherited;
  {$ENDIF}
  if Accept or(Source is TmwCustomEdit) then begin
    Accept := True;
    if ((GetKeyState(VK_CONTROL) shr 31)=0) then begin
      if DragCursor<>crDrag then
        DragCursor := crDrag;
    end else //Ctrl is pressed => change cursor to indicate copy instead of move
      if DragCursor<>crMultiDrag then
        DragCursor := crMultiDrag;
    if State=dsDragLeave then //restore prev caret position
      ComputeCaret(FMouseDownX, FMouseDownY)
    else //position caret under the mouse cursor
      ComputeCaret(X, Y);
  end;
  {$IFDEF DebugDragEvents} finally OutputDebugString('<DragOver'); end; {$ENDIF}
end;

procedure TmwCustomEdit.DragDrop(Source: TObject; X, Y: Integer);
var
  CurInd, DragBegInd, DragEndInd: Integer;
  sWK: string;
  dropLine, oldLines, correction: integer;                                      //gp 12/17/1998
  oldScroll: boolean;                                                           //gp 12/16/1998
begin
  {$IFDEF DebugDragEvents} try OutputDebugString(PChar(Format('>DragDrop %p %d %d',[pointer(Source),x,y]))); {$ENDIF}
  if ReadOnly then exit;                                                        //bds 12/20/1998
  IncPaintLock;
  try
  {$IFDEF MWE_FPC}
    inherited DragDrop(Source, X, Y);
  {$ELSE}
    inherited;
  {$ENDIF}
    if (Source=self) and (SelAvail) then begin
      ComputeCaret(X,Y);
      sWk := SelText; //text to be moved/copied
      CurInd := CoorToIndex(CaretXY);                                           //bds 12/20/1998
      DragBegInd := CoorToIndex(fDragBlockBegin);
      DragEndInd := CoorToIndex(fDragBlockEnd);
      //copy or move text if drop occured out of selection
      if (CurInd<DragBegInd) or (CurInd>DragEndInd) then begin
        if (GetKeyState(VK_CONTROL) shr 31)=0 then begin
          //if Ctrl is not down then text will be moved
          if (DragBegInd<CurInd) and (FDragBlockEnd.Y = CaretY) then
            //take into account deleted text if dropping occured on the same line
            CurInd := CurInd - (DragEndInd - DragBegInd);
          FUndoList.AddChange(mwcrDragDropDelete, fBlockBegin,
            fBlockEnd, PChar(SelText));                                         //tskurz 12/16/1998
          oldLines := Lines.Count;                                              //gp 12/16/1998
          dropLine := CaretY;                                                   //gp 12/17/1998
          SelText := '';
        {begin}                                                                 //gp 12/16/1998
          if dropLine > CaretY
            then correction := oldLines-Lines.Count
            else correction := 0;
        end
        else correction := 0;
        {end}                                                                   //gp 12/16/1998
        //move
        BlockBegin := IndexToCoor(CurInd);
        {begin}                                                                 //gp 12/16/1998
        BlockBegin := Point(BlockBegin.X,BlockBegin.Y-correction);           
        CurInd := CoorToIndex(BlockBegin);                       
        oldScroll := fScrollPastEOL;                             
        fScrollPastEOL := true;
        try                                                      
        {end}                                                                   //gp 12/16/1998
          CaretX := BlockBegin.x;
          CaretY := BlockBegin.Y;
          BlockEnd := BlockBegin;
          SelText := sWk;
        finally fScrollPastEOL := oldScroll; end;                               //gp 12/16/1998
        FUndoList.AddChange(mwcrDragDropInsert,IndexToCoor(CurInd),
          BlockEnd,PChar(SelText));                                             //tskurz 12/16/1998
      end;
    end;
  finally
    DecPaintLock;
  end;
  {$IFDEF DebugDragEvents} finally OutputDebugString('<DragDrop'); end; {$ENDIF}
end;
{end}                                                                           //ked 12/10/1998

procedure TmwCustomEdit.SetGutterColor(Value: TColor);                          //jdj 12/12/1998
begin
  if fGutterColor <> Value then
    begin
      FGutterColor := Value;
      Invalidate;
    end;
end;

procedure TmwCustomEdit.SetRightEdge(Value: Integer);                           //jdj 12/12/1998
begin
  if (fRightEdge <> Value) then
    begin
      fRightEdge := Value;
      fRightEdgeOffset := fTextOffset + fRightEdge * fCharWidth;
      Invalidate;
    end;
end;

{ TmwBookMarkOpt }

{begin}                                                                         //jdj 12/12/1998
constructor TmwBookMarkOpt.Create(AOwner: TmwCustomEdit);
begin
  inherited Create;
  FOwner := AOwner;
  fEnableKeys := True;
  fGlyphsVisible := True;
  fLeftMargin := 2;
  fXOffset := 12;
end;

procedure TmwBookMarkOpt.SetGlyphsVisible(Value: Boolean);
begin
  if (fGlyphsVisible <> Value) then
    begin
      fGlyphsVisible := Value;
      FOwner.Invalidate;
    end;
end;

procedure TmwBookMarkOpt.SetLeftMargin(Value: Integer); 
begin
  if (fLeftMargin <> Value) then
    begin
      fLeftMargin := Value;
      FOwner.Invalidate;
    end;
end;
{end}                                                                           //jdj 12/12/1998

procedure TmwBookMarkOpt.SetXOffset(const Value: integer);                      //gp 1/10/1999
begin
  fXoffset := Value;
  if assigned(fOwner) then fOwner.Invalidate;
end;

function TmwCustomEdit.GetMaxUndo: Integer;                                     //tskurz 12/14/1998
begin
   result := fUndoList.MaxUndo;
end;

procedure TmwCustomEdit.SetMaxUndo(const Value: Integer);                       //tskurz 12/14/1998
begin
  if Value > -1 then
  begin
     fUndoList.MaxUndo := Value;
     fRedoList.MaxUndo := Value;                                                //tskurz 1/4/1999
  end;
end;

procedure TmwCustomEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = fHighLighter) then
    begin
      fHighLighter := nil;
      Invalidate;                                                               //jdj 12/18/1998
    end;
  if (Operation = opRemove) and (AComponent = fBookmarkImages) then begin       //gp 1/2/1999
    fBookmarkImages := nil;
    Invalidate;
  end;
end;

procedure TmwCustomEdit.SetHighlighter(const Value: TmwCustomHighLighter);
begin
//  fHighLighter := Value;                                                      //jdj 12/18/1998
  if Value <> nil then
    begin
//    Value.mwEdit := Self;                                                     //jdj 12/18/1998
      Value.mwEditList.AddInstance(Self);                                       //jdj 12/18/1998
      Value.FreeNotification(Self);
    end
  else if Assigned(fHighLighter) then
    fHighLighter.mwEditList.RemoveInstance(Self);                               //jdj 12/18/1998
//  fHighLighter.mwEdit := nil;                                                 //jdj 12/18/1998
  fHighLighter := Value;                                                        //jdj 12/18/1998
  Invalidate;                                                                   //jdj 12/18/1998
end;

{begin}                                                                         //bds 12/16/1998
procedure TmwCustomEdit.SetBorderStyle(Value: TBorderStyle);
begin
  if fBorderStyle <> Value then
  begin
    fBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TmwCustomEdit.SetHideSelection(const Value: boolean);
begin
  if fHideSelection <> Value then
  begin
    FHideSelection := Value;
    Invalidate;
  end;
end;

procedure TmwCustomEdit.SetInsertMode(const Value: boolean);
begin
  if fInserting <> Value then
  begin
    fInserting := Value;
    if not (csDesigning in ComponentState) then
      // Reset the caret.
      InitializeCaret;
  end;
end;

procedure TmwCustomEdit.InitializeCaret;
{begin}                                                                         //bds 12/17/1998
var
  ct: TCaretType;
  cw, ch: integer;
{end}                                                                           //bds 12/17/1998
begin
  // CreateCaret automatically destroys the previous one, so we don't have to
  // worry about cleaning up the old one here with DestroyCaret.
  // Ideally, we will have properties that control what these two carets look like.
  {begin}                                                                       //bds 12/17/1998
//if InsertMode then
//  CreateCaret(Handle, 0, 2, fTextHeight - 2)
//else
//  CreateCaret(Handle, 0, fCharWidth, fTextHeight - 2);
  if InsertMode then
    ct := FInsertCaret
  else
    ct := FOverwriteCaret;
  case ct of
    ctHorizontalLine:
      begin
        cw := fCharWidth;
        ch := 2;
        FCaretOffset := Point(0,fTextHeight-2);
      end;
    ctHalfBlock:
      begin
        cw := fCharWidth;
        ch := (fTextHeight - 2) div 2;
        FCaretOffset := Point(0,ch);
      end;
    ctBlock:
      begin
        cw := fCharWidth;
        ch := fTextHeight - 2;
        FCaretOffset := Point(0,0);
      end;
  else // ctVerticalLine
    cw := 2;
    ch := fTextHeight - 2;
    FCaretOffset := Point(-1,0);                                                //bds 12/27/1998
  end;
  CreateCaret(Handle, 0, cw, ch);
  {end}                                                                         //bds 12/17/1998
  UpdateCaret;
end;

procedure TmwCustomEdit.SetScrollPastEOL(const Value: boolean);
begin
  if FScrollPastEOL <> Value then
  begin
    FScrollPastEOL := Value;
    if not FScrollPastEOL then
      CaretX := CaretX;  // Reset pos in case it's past EOL currently.
  end;
end;
{end}                                                                           //bds 12/16/1998

{begin}                                                                         //bds 12/17/1998
procedure TmwCustomEdit.SetInsertCaret(const Value: TCaretType);
begin
  if FInsertCaret <> Value then
  begin
    FInsertCaret := Value;
    InitializeCaret;
  end;
end;

procedure TmwCustomEdit.SetOverwriteCaret(const Value: TCaretType);
begin
  if FOverwriteCaret <> Value then
  begin
    FOverwriteCaret := Value;
    InitializeCaret;
  end;
end;

function TmwCustomEdit.GetCaretXY: TPoint;
begin
  Result := Point(CaretX, CaretY);
end;

procedure TmwCustomEdit.SetCaretXY(const Value: TPoint);
begin
  IncPaintLock;
  try
//    CaretX := Value.X;                                                        //bds 12/20/1998
    // Set Y first because ScrollPastEOL can effect the X pos
    CaretY := Value.Y;
    CaretX := Value.X;                                                          //bds 12/20/1998
  finally
    DecPaintLock;
  end;
end;

{begin}                                                                         //bds 12/24/1998
procedure TmwCustomEdit.SetMaxLeftChar(const Value: integer);
var
  x: integer;
//  p: ^PIntArray;
begin
  if fMaxLeftChar <> Value then
  begin
    fMaxLeftChar := Value;
    // Note:  this is incomplete.  I only implemented enough to get the FETODist
    // stuff working.  More needs to be done if this is going to work properly
    // p := @FETODist;
    ReallocMem(FETODist, fMaxLeftChar * SizeOf(Integer));
    // Must have range checking turned off here!
    {$IFDEF MWE_FPC}
    for x := 0 to fMaxLeftChar-1 do
      {$IFOPT R+} {$DEFINE MWE_RESET_RANGE_CHECK} {$R-} {$ENDIF}
      FETODist^[x] := fCharWidth;
      {$IFDEF MWE_RESET_RANGE_CHECK} {$R+} {$UNDEF MWE_RESET_RANGE_CHECK} {$ENDIF}
    {$ELSE}
    for x := 0 to fMaxLeftChar-1 do
      {$IFOPT R+} {$DEFINE MWE_RESET_RANGE_CHECK} {$R-} {$ENDIF}
      FETODist[x] := fCharWidth;
      {$IFDEF MWE_RESET_RANGE_CHECK} {$R+} {$UNDEF MWE_RESET_RANGE_CHECK} {$ENDIF}
    {$ENDIF}
    Invalidate;
  end;
end;

procedure TmwCustomEdit.SetCharWidth(const Value: integer);
var
  x: integer;
begin
  if fCharWidth <> Value then
  begin
    fCharWidth := Value;
    // Must have range checking turned off here!
    {$IFDEF MWE_FPC}
    for x := 0 to fMaxLeftChar-1 do
      {$IFOPT R+} {$DEFINE MWE_RESET_RANGE_CHECK} {$R-} {$ENDIF}
      FETODist^[x] := fCharWidth;
      {$IFDEF MWE_RESET_RANGE_CHECK} {$R+} {$UNDEF MWE_RESET_RANGE_CHECK} {$ENDIF}
    {$ELSE}
    for x := 0 to fMaxLeftChar-1 do
      {$IFOPT R+} {$DEFINE MWE_RESET_RANGE_CHECK} {$R-} {$ENDIF}
      FETODist[x] := fCharWidth;
      {$IFDEF MWE_RESET_RANGE_CHECK} {$R+} {$UNDEF MWE_RESET_RANGE_CHECK} {$ENDIF}
    {$ENDIF}
    Invalidate;
  end;
end;
{end}                                                                           //bds 12/24/1998

procedure TmwCustomEdit.EnsureCursorPosVisible;
begin
  IncPaintLock;
  try
    // Make sure X is visible
    if fCaretX < (LeftChar - 1) then
      LeftChar := fCaretX + 1
    else if fCaretX >= (fCharsInWindow + LeftChar) then                         //bds 12/24/1998
      LeftChar := fCaretX - fCharsInWindow + 2;

    // Make sure Y is visible
    if fCaretY < (TopLine - 1) then
      TopLine := fCaretY + 1
    else if fCaretY > (TopLine + (LinesInWindow - 2)) then                      //bds 12/20/1998
      TopLine := fCaretY - (LinesInWindow - 2);                                 //bds 12/20/1998
  finally
    DecPaintLock;
  end;
end;
{end}                                                                           //bds 12/17/1998


{begin}                                                                         //bds 12/23/1998
procedure TmwCustomEdit.SetKeystrokes(const Value: TmwKeyStrokes);
begin
  if Value = NIL then
    FKeystrokes.Clear
  else
    FKeystrokes.Assign(Value);
end;
{end}                                                                           //bds 12/23/1998


procedure TmwCustomEdit.SetDefaultKeystrokes;
begin
  {begin}                                                                       //bds 12/29/1998
  // The guts of this have been moved into mwKeyCmds.pas
  FKeystrokes.ResetDefaults;
  {end}                                                                         //bds 12/29/1998
end;


{begin}                                                                         //bds 12/20/1998
// If the translations requires Data, memory will be allocated for it via a
// GetMem call.  The client must call FreeMem on Data if it is not NIL.
function TmwCustomEdit.TranslateKeyCode(Code: word; Shift: TShiftState;
                          var Data: pointer): TmwEditorCommand;                 //bds 12/23/1998

{begin}                                                                         //bds 12/23/1998
var
  i: integer;
begin
  i := Keystrokes.FindKeycode(Code, Shift);
  if i >= 0 then
  {$IFDEF MWE_FPC}
    Result := Keystrokes.Items[i].Command
  {$ELSE}
    Result := Keystrokes[i].Command
  {$ENDIF}
  else
    Result := ecNone;
// removed old code
{end}                                                                           //bds 12/23/1998
end;

procedure TmwCustomEdit.CommandProcessor(Command: TmwEditorCommand; AChar: char;//bds 12/23/1998
   Data: pointer);
const
  ALPHANUMERIC = DIGIT + ALPHA_UC + ALPHA_LC;
var
  CX: Integer;
  Len: Integer;
  Temp: string;
  Temp2: String;
  Helper: string;
  SpaceCount1: Integer;
  SpaceCount2: Integer;
  BackCounter: Integer;
  StartOfBlock: TPoint;
  oldScroll: boolean;
  moveBkm: boolean;                                                            
  WP: TPoint;
  Caret: TPoint;                                                                //gp 12/27/1998
  i: integer;                                                                   //CdeB 12/16/1998
begin
  ProcessCommand(Command, AChar, Data);
  if (Command = ecNone) or (Command >= ecUserFirst) then exit;

  IncPaintLock;
  try
    case Command of
      ecLeft,
      ecSelLeft:
        begin
          CX := CaretX - 1;
          if (Command = ecSelLeft) then
          begin
            if not SelAvail then SetBlockBegin(Point(CX + 1, CaretY));
            SetBlockEnd(Point(CX, CaretY));
          end else
            SetBlockBegin(Point(CX, CaretY));
          if (not fScrollPastEOL) and (CX < 1) then
          begin
            if CaretY > 1 then
            begin
              CaretY := CaretY - 1;
              CaretX := Length(LineText) + 1;
            end;
          end else
            CaretX := CX;
        end;
      ecRight,
      ecSelRight:
        begin
          CX := CaretX + 1;
          if (Command = ecSelRight) then
          begin
            if not SelAvail then SetBlockBegin(Point(CX - 1, CaretY));
            SetBlockEnd(Point(CX, CaretY));
          end else
            SetBlockBegin(Point(CX, CaretY));
          if (not fScrollPastEOL) and (CX > (Length(LineText) + 1)) then
          begin
            if CaretY < Lines.Count then
            begin
              CaretX := 1;
              CaretY := CaretY + 1;
            end;
          end else
            CaretX := CX;
        end;
      ecUp,
      ecSelUp:
        begin
          CaretY := CaretY - 1;
          if (Command = ecSelUp) then
          begin
            if not SelAvail then
              SetBlockBegin(Point(CaretX, CaretY + 1));
            SetBlockEnd(CaretXY);
          end else
            SetBlockBegin(CaretXY);
        end;
      ecDown,
      ecSelDown:
        begin
          CaretY := CaretY + 1;
          if Command = ecSelDown then
          begin
            if not SelAvail then
              SetBlockBegin(Point(CaretX, CaretY - 1));
            SetBlockEnd(CaretXY);
          end else
            SetBlockBegin(CaretXY);
        end;
      ecWordLeft,
      ecSelWordLeft:
        begin
          if (Command = ecSelWordLeft) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretXY := LastWordPos;
          if Command = ecSelWordLeft then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecWordRight,
      ecSelWordRight:
        begin
          if (Command = ecSelWordRight) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretXY := NextWordPos;
          if Command = ecSelWordRight then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecLineStart,
      ecSelLineStart:
        begin
          if (Command = ecSelLineStart) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretX := 1;
          if Command = ecSelLineStart then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecLineEnd,
      ecSelLineEnd:
       begin
          if (Command = ecSelLineEnd) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CX := Length(LineText) + 1;
          CaretX := CX;
          if Command = ecSelLineEnd then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecPageUp,
      ecSelPageUp:
        begin
          if (Command = ecSelPageUp) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          TopLine := TopLine - LinesInWindow;
          CaretY := CaretY - LinesInWindow;
          if Command = ecSelPageUp then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecPageDown,
      ecSelPageDown:
        begin
          if (Command = ecSelPageDown) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          TopLine := TopLine + LinesInWindow;
          CaretY := CaretY + LinesInWindow;
          if (Command = ecSelPageDown) then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecPageLeft,
      ecSelPageLeft:
        begin
          if (Command = ecSelPageLeft) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          LeftChar := LeftChar - CharsInWindow;
          CaretX := CaretX - CharsInWindow;
          if (Command = ecSelPageLeft) then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecPageRight,
      ecSelPageRight:
        begin
          if (Command = ecSelPageRight) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          LeftChar := LeftChar + CharsInWindow;
          CaretX := CaretX + CharsInWindow;
          if (Command = ecSelPageRight) then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecPageTop,
      ecSelPageTop:
        begin
          if (Command = ecSelPageTop) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretY := TopLine;
          if Command = ecSelPageTop then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecPageBottom,
      ecSelPageBottom:
        begin
          if (Command = ecSelPageBottom) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretY := TopLine + (LinesInWindow - 1);
          if (Command = ecSelPageBottom) then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecEditorTop,
      ecSelEditorTop:
        begin
          if (Command = ecSelEditorTop) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretX := 1;
          CaretY := 1;
          if Command = ecSelEditorTop then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecEditorBottom,
      ecSelEditorBottom:
        begin
          if (Command = ecSelEditorBottom) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretY := Lines.Count;
          TopLine := Lines.Count - fLinesInWindow + 1;
          CaretX := Length(Lines[Lines.Count-1]) + 1;
          LeftChar := 1;
          if Command = ecSelEditorBottom then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecGotoXY,
      ecSelGotoXY:
        begin
          if Data = NIL then exit;
          if (Command = ecSelGotoXY) and (not SelAvail) then
            SetBlockBegin(CaretXY);
          CaretXY := PPoint(Data)^;
          if Command = ecSelGotoXY then SetBlockEnd(CaretXY)
          else SetBlockBegin(CaretXY);
        end;
      ecSelectAll:
        begin
          SelectAll;
        end;
      ecDeleteLastChar:
        begin
          if ReadOnly then exit;
          if SelAvail then
          begin
            FUndoList.AddChange(mwcrDelete, fBlockBegin, fBlockEnd, PChar(SelText));
            SetSelText('');
            exit;
          end;
          Temp := LineText;
          Len := Length(Temp);
          SpaceCount2 := 0;
          if Len + 1 >= CaretX then
          begin
            if CaretX > 1 then
            begin
              SpaceCount1 := LeftSpaces(Temp);
              if (Temp[CaretX - 1] <= #32) and (SpaceCount1 = fCaretX) then
              begin
                if SpaceCount1 > 0 then
                begin
                  BackCounter := fCaretY - 1;
                  if BackCounter < 0 then SpaceCount2 := 0;
                  while BackCounter > 0 do
                  begin
                    SpaceCount2 := LeftSpaces(Lines[BackCounter]);
                    if SpaceCount2 < SpaceCount1 then break;
                    Dec(BackCounter);
                  end;
                end;
                Helper := Copy(Temp, 1, SpaceCount1 - SpaceCount2);
                Delete(Temp, 1, SpaceCount1 - SpaceCount2);
                LineText := TrimRight(Temp);
                CaretX := CaretX - (SpaceCount1 - SpaceCount2);
                FUndoList.AddChange(mwcrDelete, CaretXY, CaretXY, PChar(Helper));
              end else begin
                CaretX := CaretX - 1;
                Helper := Temp[CaretX];
                Caret := CaretXY;                                               //gp 12/27/1998
//              FUndoList.AddChange(mwcrDelete, Caret, Caret, PChar(Helper));   //gp 12/27/1998
                Delete(Temp, CaretX, 1);
                LineText := TrimRight(Temp);
                FUndoList.AddChange(mwcrDelete, Caret, Caret, PChar(Helper));   //gp 12/27/1998
              end;
            end else begin
              if fCaretY > 0 then
              begin
                Lines.Delete(fCaretY);
                {begin}                                                         //CdeB 12/16/1998
                for i := 0 to Marks.count-1 do
                {$IFDEF MWE_FPC}
                  if Marks.Items[i].Line >= CaretY then
                    Marks.ITems[i].Line := Marks.Items[i].Line - 1;
                {$ELSE}
                  if Marks[i].Line >= CaretY then
                    Marks[i].Line := Marks[i].Line - 1;
                {$ENDIF}
                {end}                                                           //CdeB 12/16/1998
                CaretY := CaretY - 1;
                CaretX := Length(LineText) + 1;
                Caret := CaretXY;                                               //gp 12/27/1998
//              FUndoList.AddChange(mwcrDelete, CaretXY, CaretXY, PChar(#13#10));
                                                                                //gp 12/27/1998
                LineText := LineText + TrimRight(Temp);
                FUndoList.AddChange(mwcrDelete, CaretXY, CaretXY, PChar(#13#10));
                                                                                //gp 12/27/1998 
              end;
            end;
          end else begin
            FUndoList.AddChange(mwcrDelete, CaretXY, CaretXY, '');
            CaretX := CaretX - 1;
          end;
        end;
      ecDeleteChar:
        begin
          if ReadOnly then exit;
          if SelAvail then
          begin
            FUndoList.AddChange(mwcrDelete, fBlockBegin, fBlockEnd, PChar(SelText));
            SetSelText('');
            exit;
          end;
          Temp := LineText;
          Len := Length(Temp);
          if Len >= CaretX then
          begin
            Helper := Temp[CaretX];
            FUndoList.AddChange(mwcrDelete, CaretXY, CaretXY, PChar(Helper));
            Delete(Temp, CaretX, 1);
            LineText := TrimRight(Temp);
            {begin}                                                             //CdeB 12/16/1998
            for i:= 0 to Marks.count-1 do
              {$IFDEF MWE_FPC}
              if Marks.Items[i].Line > CaretY then
                Marks.ITems[i].Line := Marks.Items[i].Line - 1;
              {$ELSE}
              if Marks[i].Line > CaretY then
                Marks[i].Line := Marks[i].Line - 1;
              {$ENDIF}
            {end}                                                               //CdeB 12/16/1998
          end else begin
            if fCaretY < Lines.Count - 1 then
            begin
              Helper := StringOfChar(' ', fCaretX - Len);
              LineText := TrimRight(Temp + Helper + Lines[CaretY]);
              FUndoList.AddChange(mwcrDelete, CaretXY, CaretXY, #13#10);
              Lines.Delete(CaretY);
              if Lines.Count = 0 then
                Lines.Add('');
            end;
          end;
        end;
      ecDeleteWord:
        begin
          if ReadOnly then exit;
          WP := NextWordPos;
          if (CaretX = WP.x) and (CaretY = WP.y) then exit; // nothing to do
          SetBlockBegin(CaretXY);
          SetBlockEnd(WP);
          FUndoList.AddChange(mwcrDeleteAfterCursor, CaretXY, CaretXY,
             PChar(SelText));
          SetSelText('');
          CaretXY := CaretXY; // Fix up cursor position
        end;
      ecDeleteLastWord:
        begin
          if ReadOnly then exit;
          WP := LastWordPos;
          if (CaretX = WP.x) and (CaretY = WP.y) then exit; // nothing to do
          SetBlockBegin(WP);
          SetBlockEnd(CaretXY);
          FUndoList.AddChange(mwcrDeleteAfterCursor, WP, CaretXY,
             PChar(SelText));
          SetSelText('');
          CaretXY := CaretXY; // Fix up cursor position
        end;
      ecDeleteBOL:
        begin
          FUndoList.AddChange(mwcrDelete, Point(0, CaretY), Point(0, CaretY),
             PChar(Copy(LineText, 1, CaretX-1)));
          SetBlockBegin(Point(1, CaretY));
          SetBlockEnd(CaretXY);
          SetSelText('');
          CaretXY := Point(1, CaretY); // Fix up cursor position
        end;
      ecDeleteEOL:
        begin
          FUndoList.AddChange(mwcrDeleteAfterCursor, CaretXY, CaretXY,
             PChar(Copy(LineText, CaretX, Length(LineText)-CaretX+1)));
          SetBlockBegin(CaretXY);
          SetBlockEnd(Point(Length(LineText)+1, CaretY));
          SetSelText('');
          CaretXY := CaretXY; // Fix up cursor position
        end;
      ecDeleteLine:
        begin
          if SelAvail then
            SetBlockBegin(CaretXY);
          FUndoList.AddChange(mwcrDeleteAfterCursor, Point(1, CaretY), CaretXY,
             PChar(LineText+#13#10));
          Lines.Delete(CaretY-1);
          if Lines.Count = 0 then
            Lines.Add('');
          CaretXY := CaretXY; // Fix up cursor position
        end;
      ecClearAll:
        begin
          if not ReadOnly then ClearAll;
        end;
      ecInsertLine,
      ecLineBreak:
        begin
          if ReadOnly then exit;
          SpaceCount2 := 0;
          Temp := LineText;
          Temp2 := LineText;
          Len := Length(Temp);
          if Len > 0 then
          begin
            if Len >= CaretX then
            begin
              if CaretX > 1 then
              begin
                SpaceCount1 := LeftSpaces(Temp);
                if fCaretX <= SpaceCount1 then
                begin
                  Lines[fCaretY] := '';
                  Lines.Insert(CaretY, Temp);
//                FUndoList.AddChange(mwcrLineBreak, CaretXY, CaretXY, PChar(Temp));
                                                                                //tskurz 12/21/1998
                  FUndoList.AddChange(mwcrLineBreak,Point(CaretX,CaretY-1),     //tskurz 12/21/1998
                    CaretXY,PChar(Temp));
                  if Command = ecLineBreak then
                    CaretY := CaretY + 1;
                end else begin
                  Temp := Copy(LineText, 1, fCaretX);
                  LineText := TrimRight(Temp);
                  Helper := StringOfChar(' ', SpaceCount1);
                  Delete(Temp2, 1, fCaretX);
                  FUndoList.AddChange(mwcrLineBreak, CaretXY, CaretXY,
                     PChar(Temp2));
                  Temp2 := Helper + Temp2;
                  Lines.Insert(CaretY, Temp2);
                  if Command = ecLineBreak then
                    CaretXY := Point(SpaceCount1 + 1, CaretY + 1);
                end;
              end else begin
                BackCounter := fCaretY - 1;
                while BackCounter > 0 do
                begin
                  if Length(Lines[BackCounter]) > 0 then break;
                  dec(BackCounter);
                end;
                Lines[fCaretY] := '';
                Lines.Insert(CaretY, Temp);
                FUndoList.AddChange(mwcrLineBreak, CaretXY, CaretXY,
                   PChar(Temp2));
                if Command = ecLineBreak then
                  CaretY := CaretY + 1;
              end;
            end else begin
              BackCounter := fCaretY;
              while BackCounter > 0 do
              begin
                SpaceCount2 := LeftSpaces(Lines[BackCounter]);
                if Length(Lines[BackCounter]) > 0 then break;
                dec(BackCounter);
              end;
              FUndoList.AddChange(mwcrLineBreak, CaretXY, CaretXY, NIL);
              if Command = ecLineBreak then
                CaretX := SpaceCount2 + 1;
              Lines.Insert(CaretY, '');
              if Command = ecLineBreak then
                CaretY := CaretY + 1;
            end;
          end else begin
            BackCounter := fCaretY;
            while BackCounter >= 0 do
            begin
              SpaceCount2 := LeftSpaces(Lines[BackCounter]);
              if Length(Lines[BackCounter]) > 0 then break;
              dec(BackCounter);
            end;
            FUndoList.AddChange(mwcrLineBreak, CaretXY, CaretXY, NIL);
            if Command = ecLineBreak then
              CaretX := SpaceCount2 + 1;
            Lines.Insert(fCaretY, '');
            if Command = ecLineBreak then
              CaretY := CaretY + 1;
          end;
          {begin}                                                               //CdeB 12/16/1998
          {$IFDEF MWE_FPC}
          for i:= 0 to Marks.count-1 do
            if Marks.Items[i].Line >= CaretY then
              Marks.Items[i].Line := Marks.Items[i].Line + 1;
          {$ELSE}
          for i:= 0 to Marks.count-1 do
            if Marks[i].Line >= CaretY then
              Marks[i].Line := Marks[i].Line + 1;
          {$ENDIF}
          {end}                                                                 //CdeB 12/16/1998
        end;
      ecChar:
        begin
          // #127 is Ctrl+Backspace.
          if (AChar < #32) or (AChar = #127) or ReadOnly then exit; // This will cause problems when we put tabs in.
          if SelAvail then
          begin
            FUndoList.AddChange(mwcrDelete,fBlockBegin,fBlockEnd,PChar(SelText));
            StartOfBlock := fBlockBegin;
            SetSelText(AChar);
            FUndoList.AddChange(mwcrInsert,StartOfBlock,fBlockEnd,PChar(SelText));
            exit;
          end;
          Temp := LineText;
          Len := Length(Temp);
          if Len < CaretX then
          begin
            Helper := StringOfChar(' ', CaretX - Len);
            Temp := Temp + Helper;
            Helper := '';
          end;
          if fInserting then
          begin
            oldScroll := fScrollPastEOL;
            fScrollPastEOL := true;
            try
              StartOfBlock := CaretXY;
              {$IFDEF MWE_FPC}
              System.Insert(AChar, Temp, CaretX);
              {$ELSE}
              Insert(AChar, Temp, CaretX);
              {$ENDIF}
              CaretX := CaretX + 1;
              LineText := TrimRight(Temp); ;
              FUndoList.AddChange(mwcrInsert, StartOfBlock, CaretXY, NIL);
              Temp := '';
              if CaretX >= LeftChar + fCharsInWindow then
                LeftChar := LeftChar + min(25, fCharsInWindow - 1);
            finally
              fScrollPastEOL := oldScroll;
            end;
          end else begin
            oldScroll := fScrollPastEOL;
            fScrollPastEOL := true;
            try
              StartOfBlock := CaretXY;
              Helper := Temp[CaretX];
// This is extremely inefficient.  Need some sort of direct access to the string here.
              Temp[CaretX] := AChar;
              CaretX := CaretX + 1;
              LineText := TrimRight(Temp);
//              CaretX := CaretX + 1;                                           //bds 12/28/1998, removed
              FUndoList.AddChange(mwcrInsert, StartOfBlock, CaretXY, PChar(Helper));
              Temp := '';
              if CaretX >= LeftChar + fCharsInWindow then
                LeftChar := LeftChar + min(25, fCharsInWindow - 1);
            finally
              fScrollPastEOL := oldScroll;
            end;
          end;
        end;
      ecUndo:
        begin
          Undo;
        end;
      ecRedo:
        begin
          //raise Exception.Create('Redo not yet supported by mwEdit');
          Redo;                                                                 //tskurz 1/4/1999
        end;
      ecGotoMarker0..ecGotoMarker9:
        begin
          if BookMarkOptions.EnableKeys then begin                              //gp 12/30/1998
            GotoBookMark(Command - ecGotoMarker0);                              //CdeB 12/16/1998 - removed, //gp 12/31/1998 - reinstated
          end;

        end;
      ecSetMarker0..ecSetMarker9:
        begin
          if BookMarkOptions.EnableKeys then begin                              //gp 12/30/1998
            CX := Command - ecSetMarker0;
                                                                                //CdeB 12/16/1998 - removed, //gp 12/31/1998 - reinstated
            if assigned(fBookMarks[CX]) then                                    //gp 12/31/1998
            begin
              moveBkm := (fBookMarks[CX].Line <> CaretY);
              ClearBookMark(CX);
              if moveBkm then
                SetBookMark(CX, CaretX, CaretY);
            end else
              SetBookMark(CX, CaretX, CaretY);

{begin}                                                                         //bds 12/23/1998
(* This commented out stuff will become ecSetMarkerAt in the next version
{end}                                                                           //bds 12/23/1998
            if Data <> NIL then
            begin
              with PPoint(Data)^ do
              begin
                CX := Command - ecSetMarker0;
                if fBookMarks[CX].IsSet then
                begin
                  moveBkm := (fBookMarks[CX].Y <> Y);
                  ClearBookMark(CX);
                  if moveBkm then
                    SetBookMark(CX, X, Y);
                end else
                  SetBookMark(CX, X, Y);
              end;
            end;
*)                                                                              //bds 12/23/1998
          end; // if BookMarkOptions.EnableKeys
        end;
      ecCut:
        begin
          if (not ReadOnly) and SelAvail then
            CutToClipboard;
        end;
      ecCopy:
        begin
          CopyToClipboard;
        end;
      ecPaste:
        begin
          if not ReadOnly then PasteFromClipboard;
        end;
      ecScrollUp:
        begin
          TopLine := TopLine - 1;
          if CaretY > TopLine + LinesInWindow - 1 then
            CaretY := TopLine + LinesInWindow - 1;
        end;
      ecScrollDown:
        begin
          TopLine := TopLine + 1;
          if CaretY < TopLine then
            CaretY := TopLine;
        end;
      ecScrollLeft:
        begin
          LeftChar := LeftChar - 1;
          if CaretX > LeftChar + CharsInWindow then
            CaretX := LeftChar + CharsInWindow;
        end;
      ecScrollRight:
        begin
          LeftChar := LeftChar + 1;
          if CaretX < LeftChar then
            CaretX := LeftChar;
        end;
      ecInsertMode:
        begin
          InsertMode := TRUE;
        end;
      ecOverwriteMode:
        begin
          InsertMode := FALSE;
        end;
      ecToggleMode:
        begin
          InsertMode := not InsertMode;
        end;
      ecBlockIndent:
        begin
        end;
      ecBlockUnindent:
        begin
        end;
    end;
  finally
    DecPaintLock;
  end;

  // Add to macro recorder stuff here?  Can't without rewriting several "exit"
  // statements above, or putting it in the finally section.
end;

procedure TmwCustomEdit.ProcessCommand(var Command: TmwEditorCommand;           //bds 12/23/1998
   var AChar: char; Data: pointer);
begin
  if Command < ecUserFirst then
  begin
    if assigned(FOnProcessCommand) then
      FOnProcessCommand(Command, AChar, Data);
  end else begin
    if assigned(FOnProcessUserCommand) then
      FOnProcessUserCommand(Command, AChar, Data);
  end;
end;

procedure TmwCustomEdit.ClearAll;
begin
  IncPaintLock;
  try
    if SelAvail then
      SetBlockBegin(Point(1,1));
    Lines.Clear;
    Lines.Add('');
  finally
    DecPaintLock;
  end;
end;

// NOTE:  These functions (NextWordPos & LastWordPos) are not very efficient or
//        pretty.  I just wrote them as "placeholders" until we work out
//        exactly how next/prev word should really be implemented.
//mt 12/22/1998 fixed to support IdentChars. Still could use some cleaning up.
function TmwCustomEdit.NextWordPos: TPoint;
var
  CX, CY: Integer;
  X: integer;
  FoundWhiteSpace: boolean;
  Len: Integer;
  Temp: string;
  IdChars: TIdentChars;                                                         //mt 12/22/1998
begin
  Result := CaretXY;
  if CaretX < Length(LineText) then
  begin
    CX := CaretX;
    CY := CaretY;
  end else begin
    if CaretY >= Lines.Count then exit;
    CY := CaretY + 1;
    if CY > Lines.Count then exit;
    if Trim(Lines[CY-1]) = '' then
    begin
      Result := Point(1, CY);
      exit;
    end;
    CX := 1;
  end;

  {begin}                                                                       //bds 12/22/1998
  if fHighlighter <> NIL then
    IdChars := fHighlighter.IdentChars
  else
    IDchars := [#33..#255];
  {end}                                                                         //bds 12/22/1998

  Temp := Lines[CY-1];
  Len := Length(Temp);
  FoundWhiteSpace := (not (Temp[CX] in IdChars)) or (CY <> CaretY);             //mt 12/22/1998
  if not FoundWhiteSpace then
    for X := CX to Len do
      if (not (Temp[X] in IdChars)) then                                        //mt 12/22/1998
      begin
        CX := X;
        FoundWhiteSpace := TRUE;
        break;
      end;

  if FoundWhiteSpace then
  begin
    FoundWhiteSpace := FALSE;                                                   //mt 12/22/1998
    for X := CX to Len do
      if (Temp[X] in IdChars) then                                              //mt 12/22/1998
      begin
        Result := Point(X, CY);
        FoundWhiteSpace := TRUE;
        break;
      end;
  end;

  if not FoundWhiteSpace then
    Result := Point(Len+1, CY);

end;

//mt 12/22/1998 fixed to support IdentChars
function TmwCustomEdit.LastWordPos: TPoint;
var
  CX, CY: Integer;
  X: integer;
  FoundNonWhiteSpace: boolean;
  Len: Integer;
  Temp: string;
  IdChars: TIdentChars;                                                         //mt 12/22/1998
begin
  Result := CaretXY;
  CX := CaretX;
  Len := Length(LineText);
  if CX > Len then
    CX := Len+1;
  if (CX > 1) then
  begin
    CY := CaretY;
  end else begin
    if CaretY < 2 then
    begin
      Result := Point(1,1);
      exit;
    end;
    CY := CaretY - 1;
    CX := Length(Lines[CY-1]);
    begin
      Result := Point(CX+1, CY);
      exit;
    end;
  end;

  {begin}                                                                       //bds 12/22/1998
  if fHighlighter <> NIL then
    IdChars := fHighlighter.IdentChars
  else
    IDchars := [#33..#255];
  {end}                                                                         //bds 12/22/1998

  Temp := Lines[CY-1];
  FoundNonWhiteSpace := (Temp[CX-1] in IdChars);                                //mt 12/22/1998
  if not FoundNonWhiteSpace then
    for X := CX downto 2 do
      if Temp[X-1] in IdChars then                                              //mt 12/22/1998
      begin
        CX := X-1;
        FoundNonWhiteSpace := TRUE;
        break;
      end;

  if FoundNonWhiteSpace then
  begin
    FoundNonWhiteSpace := FALSE;
    for X := CX downto 2 do
      if (not(Temp[X-1] in IdChars)) then                                       //mt 12/22/1998
      begin
        Result := Point(X, CY);
        FoundNonWhiteSpace := TRUE;
        break;
      end;
    if not FoundNonWhiteSpace then
      Result := Point(1, CY);
  end else begin
    dec(CY);
    if CY = 0 then Result := Point(0,0)                                         //gp 01/19/1999
              else 
              Result := Point(Length(Lines[CY-1])+1, CY);
  end;

end;

(*
  These were all removed when I ripped out the hard-wired key stuff and put in  //bds 12/20/1998
  the editor commands instead.  I have left them here for reference in case I
  screwed something up in the transfer over.

  Removed them completely. I have all the backups if the need occurs.           //gp 12/24/1998
*)
{end}                                                                           //bds 12/20/1998

{begin}                                                                         //jdj 12/27/1998
procedure TmwCustomEdit.BeginUpdate;
begin
  IncPaintLock;
end;

procedure TmwCustomEdit.EndUpdate;
begin
  DecPaintLock;
end;
{end}                                                                           //jdj 12/27/1998

{ TUndoList }
{begin}                                                                         //tskurz 12/10/1998
procedure TUndoList.AddChange(ChangeReason: TChangeReason; ChangeStartPos,
              ChangeEndPos: TPoint; ChangeStr: PChar);
var
  chngptr : TChangePtr;
begin
   if FList.Count >= FMaxUndo then RemoveChange(0);
   try
      GetMem(chngptr,SizeOf(TChange));
      try
        chngptr^.ChangeStr := StrNew(ChangeStr);
      except
        FreeMem(chngptr);
        exit;
      end;
   except
     exit;
   end;
   chngptr^.ChangeReason := ChangeReason;
   {begin}                                                                      //tskurz 12/14/1998
   //chngptr^.ChangeStartPos := ChangeStartPos;
   //chngptr^.ChangeEndPos := ChangeEndPos;
   chngptr^.ChangeStartPos := minPoint(ChangeStartPos,ChangeEndPos);
   chngptr^.ChangeEndPos := maxPoint(ChangeStartPos,ChangeEndPos);
   {end}                                                                        //tskurz 12/14/1998
   FList.Add(chngptr);
   fOwner.Modified := true;                                                     //CdeB 12/16/1998
   if Assigned(FOwner.OnChange) then FOwner.OnChange(FOwner);                   //tskurz 12/15/1998
end;

constructor TUndoList.Create(AOwner: TmwCustomEdit);                            //tskurz 12/15/1998
begin
   inherited Create;
   FOwner := AOwner;                                                            //tskurz 12/15/1998
   FList := TList.Create;
   FMaxUndo := 10;
end;

destructor TUndoList.Destroy;
var
   i : Integer;
begin
// for i := 0 to FList.Count - 1 do                                             //tskurz 12/21/1998
   for i := FList.Count - 1 downto 0 do                                         //tskurz 12/21/1998
     RemoveChange(i);
   FList.Free;
   inherited Destroy;
end;

function TUndoList.GetChange(var ChangeStartPos, ChangeEndPos: TPoint;
  var ChangeStr: PChar): TChangeReason;
begin
   if FList.Count = 0 then
   begin
     result := mwcrNone;
     exit;
   end;
   ChangeStartPos := TChangePtr(FList.Items[FList.Count-1])^.ChangeStartPos;
   ChangeEndPos := TChangePtr(FList.Items[FList.Count-1])^.ChangeEndPos;
   ChangeStr := StrNew(TChangePtr(FList.Items[FList.Count-1])^.ChangeStr);
   result := TChangePtr(FList.Items[FList.Count-1])^.ChangeReason;
   RemoveChange(FList.Count-1);
end;

function TUndoList.GetCanUndo : Integer;
begin
  FCanUndo := FList.Count;
  result := FCanUndo;
end;

procedure TUndoList.SetMaxUndo(const Value: Integer);
var
  i : Integer;
begin
  if Value < FList.Count - 1 then
     for i := 0 to FList.Count - Value - 1 do
        RemoveChange(0);
  FMaxUndo := Value;
end;

procedure TUndoList.RemoveChange(index: Integer);
var
  chngptr : TChangePtr;
begin
  if (index > -1) and (index < FList.Count) then
  begin
    chngptr := FList.Items[Index];
    try
      StrDispose(chngptr^.ChangeStr);
      try
        FreeMem(chngptr);
      except
        exit;
      end;
    except
      exit;
    end;
    Flist.Delete(index);
  end;
end;

function TUndoList.GetChangeReason: TChangeReason;                              //tskurz 12/16/1998
begin
   if FList.Count = 0 then
     result := mwcrNone
   else
     result := TChangePtr(FList.Items[FList.Count-1])^.ChangeReason;
end;

procedure TUndoList.ClearList;
var
  i: Integer;
begin
  for i := FList.Count - 1 downto 0 do                                          //tskurz 1/4/1999
    RemoveChange(i);
end;

{ TmwCustomEdit }

{begin}                                                                         //CdeB 12/16/1998
procedure TmwCustomEdit.SetBookmarkImages(const Value: TImageList);             //gp 1/2/1999 - renamed
begin
  fBookmarkImages := Value;
  if fBookmarkImages <> nil then
    fBookmarkImages.FreeNotification(Self);
  Invalidate;                                                                   //gp 1/2/1999
end;

{ Called by FMarkList if change }
procedure TmwCustomEdit.MarkListChange(Sender: TObject);
//var
//  i: integer;                                                                 //gp 1/9/1999
begin
  {begin}                                                                       //gp 1/9/1999
//it is now impossible to create a TMark without an owner
//for i:= 0 to Marks.count-1 do
//  Marks[i].Edit:= Self;
  {end}                                                                         //gp 1/9/1999
  Invalidate;                                                                   //gp 1/2/1999
end;
{end}                                                                           //CdeB 12/16/1998

procedure TmwCustomEdit.SetLinesInWindow(const Value: integer);                 //gp 1/10/1999
begin
  fLinesInWindow := Value;
  ReallocMem(fGutterOffsets,fLinesInWindow*SizeOf(integer));

end;

{begin}                                                                         //tb 1/7/1999
function TmwCustomEdit.GetSelStart: integer;
  function llen(data: string): integer;
  begin
    result:=length(Data)+2;
  end;
var
  loop: integer;
  x, y: integer;
begin
  x:=Caretx;
  y:=CaretY;

  result:=0;
  loop:=0;
  while loop<(Y-1)do begin
    Result:=result+llen(lines.strings[loop]);
    inc(loop);
  end;

  result:=result+Min(X, llen(lines.strings[loop]));
end;

procedure TmwCustomEdit.SetSelStart(const Value: integer);
  function llen(data: string): integer;
  begin
    result:=length(Data)+2;
  end;
var
  loop: integer;
  count: integer;
begin
  loop:=0;
  count:=0;
  while(count+llen(lines.strings[loop])<value)do begin
    count:=count+llen(lines.strings[loop]);
    inc(loop);
  end;
  CaretX:=value-count;
  CaretY:=loop+1;

end;

function TmwCustomEdit.GetSelEnd: integer;
  function llen(data: string): integer;
  begin
    result:=length(Data)+2;
  end;
var
  loop: integer;
  x, y: integer;
begin
  x:=BlockEnd.x;
  y:=BlockEnd.y;

  result:=0;
  loop:=0;
  while loop<(y-1)do begin
    Result:=result+llen(lines.strings[loop]);
    inc(loop);
  end;
  result:=result+x;
end;

procedure TmwCustomEdit.SetSelEnd(const Value: integer);
  function llen(data: string): integer;
  begin
    result:=length(Data)+2;
  end;
var
  p: tpoint;
  loop: integer;
  count: integer;
begin
  loop:=0;
  count:=0;
  while(count+llen(lines.strings[loop])<value)do begin
    count:=count+llen(lines.strings[loop]);
    inc(loop);
  end;
  p.X:=value-count; p.y:=loop+1;
  Blockend:=p;
end;
{end}                                                                           //tb 1/7/1999

procedure TmwCustomEdit.SetExtraLineSpacing(const Value: integer);
begin
  fExtraLineSpacing := Value;
  FontChanged(self);
end;

{begin}                                                                         //gp 1/10/1999
function TmwCustomEdit.GetBookMark(BookMark: integer; var X, Y: integer): boolean;
var
  i: integer;
begin
  Result := false;
  if assigned(Marks) then
    for i := 0 to Marks.Count-1 do
      {$IFDEF MWE_FPC}
      if Marks.Items[i].IsBookmark and (Marks.Items[i].BookmarkNumber = BookMark) then begin
        X := Marks.Items[i].Column;
      {$ELSE}
      if Marks[i].IsBookmark and (Marks[i].BookmarkNumber = BookMark) then begin
        X := Marks[i].Column;
      {$ENDIF}

        {$IFDEF MWE_FPC}
        Y := Marks.Items[i].Line;
        {$ELSE}
        Y := Marks[i].Line;
        {$ENDIF}
        Result := true;
        Exit;
      end;
end;

function TmwCustomEdit.IsBookmark(BookMark: integer): boolean;
var
  x,y: integer;
begin
  Result := GetBookMark(BookMark,x,y);
end; 
{end}                                                                           //gp 1/10/1999

procedure TmwCustomEdit.ClearUndo;
begin
  fUndoList.ClearList;
  fRedoList.ClearList;
end;

procedure TmwCustomEdit.SetSelTextExternal(const Value: String);                //tskurz 1/7/1999
var
   StartOfBlock: TPoint;
begin
  if SelAvail then
    FUndoList.AddChange(mwcrDelete,fBlockBegin,fBlockEnd,PChar(GetSelText));
  StartOfBlock := fBlockBegin;
  SetSelText(Value);
  FUndoList.AddChange(mwcrInsert,StartOfBlock,fBlockBegin,PChar(Value));
end;

{ TMark }

function TMark.GetEdit: TmwCustomEdit;        
begin
  if FEdit<>nil then
    try
      if FEdit.Marks.IndexOf(self)=-1 then
        FEdit:= nil;
    except
      FEdit:= nil;
    end;
  Result := FEdit;
end;

function TMark.GetIsBookmark: boolean;
begin
  Result := (fBookmarkNum >= 0);
end;

procedure TMark.SetIsBookmark(const Value: boolean);
begin
  if Value then fBookmarkNum := 0
           else fBookmarkNum := -1;
end;

procedure TMark.SetColumn(const Value: Integer);
begin
  FColumn:= Value;
end;

procedure TMark.SetImage(const Value: Integer);
begin
  FImage:= Value;
  if assigned(fEdit) then fEdit.Invalidate;                                     //gp 1/2/1999
end;

procedure TMark.SetInternalImage(const Value: boolean);                         //gp 1/2/1999
begin
  fInternalImage := Value;
  if assigned(fEdit) then fEdit.Invalidate;
end;

procedure TMark.SetLine(const Value: Integer);
begin
  fLine:= Value;
  if assigned(fEdit) then fEdit.Invalidate;                                     //gp 1/2/1999
end;

procedure TMark.SetVisible(const Value: boolean);                               //gp 1/2/1999
begin
  fVisible := Value;
  if assigned(fEdit) then fEdit.Invalidate;
end;

constructor TMark.Create(owner: TmwCustomEdit);                                 //gp 1/9/1999
begin
  inherited Create;
  fEdit := owner;
end;

{ TMarkList }

function TMarkList.Add(Item: TMark): Integer;
begin
  Result:= Inherited Add(Item);
  DoChange;
end;

procedure TMarkList.ClearLine(line: integer);                                   //gp 1/9/1999
var
  i: integer;
begin
  i := 0;
  while (i < Count) do begin
    if (not Items[i].IsBookmark) and (Items[i].Line = line)
      then Delete(i)
      else Inc(i);
  end;
end;

constructor TMarkList.Create(owner: TmwCustomEdit);                             //gp 1/9/1999
begin
  inherited Create;
  fEdit := owner;
end;

procedure TMarkList.Delete(Index: Integer);
begin
  inherited Delete(Index);
  DoChange;
end;

procedure TMarkList.DoChange;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

function TMarkList.First: TMark;
begin
  {$IFDEF MWE_FPC}
  result:= TMark(inherited First);
  {$ELSE}
  result:= inherited First;
  {$ENDIF}
end;

function TMarkList.Get(Index: Integer): TMark;
begin
  {$IFDEF MWE_FPC}
  result:= TMark(inherited Get(Index));
  {$ELSE}
  result:= inherited Get(Index);
  {$ENDIF}
end;

//Returns up to maxMarks book/gutter marks for a chosen line.
procedure TMarkList.GetMarksForLine(line: integer; var marks: TMarks);
var
  cnt: integer;
  i  : integer;
begin
  FillChar(marks,SizeOf(marks),0);
  cnt := 0;
  for i := 0 to Count-1 do begin
    if Items[i].Line = line then begin
      Inc(cnt);
      marks[cnt] := Items[i];
      if cnt = maxMarks then break;
    end;
  end;
end;

procedure TMarkList.Insert(Index: Integer; Item: TMark);
begin
  Inherited Insert(Index, Item);
  DoChange;
end;

function TMarkList.Last: TMark;
begin
  {$IFDEF MWE_FPC}
  result := TMark(Inherited Last);
  {$ELSE}
  result:= Inherited Last;
  {$ENDIF}
end;

procedure TMarkList.Place(mark: TMark);
begin
  if assigned(fEdit) then
    if assigned(fEdit.OnPlaceBookmark) then fEdit.OnPlaceBookmark(mark);        //gp 1/9/1999
  if assigned(mark) then                                                        //gp 1/9/1999
    Add(mark);
  DoChange;
end;

procedure TMarkList.Put(Index: Integer; Item: TMark);
begin
  Inherited Put(Index, Item);
  DoChange;
end;

function TMarkList.Remove(Item: TMark): Integer;
begin
  result:= Inherited Remove(Item);
  DoChange;
end;

end.



