{+-----------------------------------------------------------------------------+
 | Class:       TmwCustomHighlighter
 | Created:     07.98 - 10.98
 | Last change: 10.1.99
 | Author:      Martin Waldenburg
 | Description: A very fast SyntaxScanner for Pascal.
 | Version:     0.64 (version history follows licence agreement)
 | Copyright (c) 1998 Martin Waldenburg
 | All rights reserved.
 |
 | Thanks to: Primoz Gabrijelcic, Michael Trier, James Jacobson,
 |            Cyrille de Brebisson
 |
 | LICENCE CONDITIONS
 |
 | USE OF THE ENCLOSED SOFTWARE
 | INDICATES YOUR ASSENT TO THE
 | FOLLOWING LICENCE CONDITIONS.
 |
 |
 |
 | These Licence Conditions are exlusively
 | governed by the Law and Rules of the
 | Federal Republic of Germany.
 |
 | Redistribution and use in source and binary form, with or without
 | modification, are permitted provided that the following conditions
 | are met:
 |
 | 1. Redistributions of source code must retain the above copyright
 |    notice, this list of conditions and the following disclaimer.
 |    If the source is modified, the complete original and unmodified
 |    source code has to distributed with the modified version.
 |
 | 2. Redistributions in binary form must reproduce the above
 |    copyright notice, these licence conditions and the disclaimer
 |    found at the end of this licence agreement in the documentation
 |    and/or other materials provided with the distribution.
 |
 | 3. Software using this code must contain a visible line of credit.
 |
 | 4. If my code is used in a "for profit" product, you have to donate
 |    to a registered charity in an amount that you feel is fair.
 |    You may use it in as many of your products as you like.
 |    Proof of this donation must be provided to the author of
 |    this software.
 |
 | 5. If you for some reasons don't want to give public credit to the
 |    author, you have to donate three times the price of your software
 |    product, or any other product including this component in any way,
 |    but no more than $500 US and not less than $200 US, or the
 |    equivalent thereof in other currency, to a registered charity.
 |    You have to do this for every of your products, which uses this
 |    code separately.
 |    Proof of this donations must be provided to the author of
 |    this software.
 |
 |
 | DISCLAIMER:
 |
 | THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS'.
 |
 | ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 | PARTICULAR PURPOSE ARE DISCLAIMED.
 |
 | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 | (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 | OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 | INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 | WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 | NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 |
 |  Martin.Waldenburg@T-Online.de
 |
 | Version history:
 |   0.6: (first version)
 |     - Primoz Gabrijelcic
 |       - moved TmwCustomHighlighter from mwPasSyn.pas into this unit
 |       - moved registry access code frm mwPasSyn.pas into this unit
 |       - general $IFDEF cleanup
 |   0.61:
 |     - James Jacobson
 |       - Some changes that will update the Editor when any changes are made to
 |         the Highlighter. That is when any color or font style is changed and
 |         when the Highlighter is added or removed the Editor will be
 |         invalidated.
 |     - Primoz Gabrijelcic
 |       - changed UseUserSettings and EnumUserSettings from abstract to
 |         do-nothing so that highlighters don't have to implement them
 |   0.62:
 |     - Michael Trier
 |       - Added IdentChars implementation to let the editor know which
 |         characters are non "whitespace".
 |     - Primoz Gabrijelcic
 |       - Added property Name to class TmwHighLightAttributes.
 |       - Updated TmwHighLightAttributes.Create, TmwHighLightAttributes.Assign.
 |       - Extended TmwCustomHighlighter with properties Name, AttrCount, and
 |         Attribute; added corresponding Get functions (all pure virtual).
 |   0.63:
 |     - Primoz Gabrijelcic
 |       - Renamed LoadFromRegistry to LoadFromBorlandRegistry.
 |       - Added Capability property.
 |       - Added code to save/load TmwCustomHighlighter to/from registry.
 |     - Cyrille de Brebisson
 |       - Added code to save/load TmwHighLightAttributes to/from registry.
 |       - Added property IntegerStyle.
 |   0.64:
 |     - Michael Trier
 |       - Renamed Name property to LanguageName.
 |       - Added DefaultFilter property.
 |
 +----------------------------------------------------------------------------+}

{$I MWEDIT.INC}                                                                 //mt 12/16/1998
//mt 12/16/1998 removed {$ObjExportAll On} since it's defined in the include

{$IFDEF MWE_FPC}
  {$mode objfpc}
  {$H+}
{$ENDIF}

unit mwHighlighter;

interface

uses
  {$IFDEF MWE_FPC}
    {$IFDEF LINUX}
      LCLLinux, 
    {$ELSE}
      Windows,
    {$ENDIF}
  {$ELSE}
  Windows,
  {$ENDIF}
  SysUtils, Classes, Graphics, Registry, Controls;                     //jdj 12/18/1998 - added Controls


{$DEFINE _Gp_MustEnhanceRegistry}
{$IFDEF MWE_COMPILER_4_UP}                                                      //gp 12/16/1998
  {$UNDEF _Gp_MustEnhanceRegistry}
{$ENDIF}
type
  TBetterRegistry = class(TRegistry)
  {$IFDEF _Gp_MustEnhanceRegistry}
    function OpenKeyReadOnly(const Key: string): Boolean;
  {$ENDIF}
  end;

  TmwEditList = class(TList)                                                    //jdj 12/18/1998
    procedure AddInstance(Control: TWinControl);
    procedure RemoveInstance(Control: TWinControl);
    procedure Invalidate;
  end;

  TmwHighLightAttributes = Class(TPersistent)
  private
    fBackground: TColor;                                                        //mt 12/14/1998 - renamed
    fForeground: TColor;                                                        //mt 12/14/1998 - renamed
    fStyle: TFontStyles;
    fOnChange: TNotifyEvent;                                                    //jdj 12/18/1998
    fName: string;                                                              //gp 12/24/1998
    procedure SetBackground(Value: TColor);                                     //jdj 12/18/1998
    procedure SetForeground(Value: TColor);                                     //jdj 12/18/1998
    procedure SetStyle(Value: TFontStyles);                                     //jdj 12/18/1998
    function GetStyleFromInt: integer;                                          //CdeB 12/16/1998
    procedure SetStyleFromInt(const Value: integer);                            //CdeB 12/16/1998
  protected
  public
    procedure Assign(Source: TPersistent); override;
    property Name: string read fName;                                           //gp 12/24/1998
    property OnChange: TNotifyEvent read fOnChange write fOnChange;             //jdj 12/18/1998
    function LoadFromBorlandRegistry(rootKey: HKEY;                             //gp 12/16/1998, //gp 12/28/1998 - made protected
               attrKey, attrName: string; oldStyle: boolean): boolean; virtual;
    function LoadFromRegistry(Reg: TBetterRegistry): boolean;                   //CdeB 12/16/1998, //gp 12/28/1998 - adapted to TBetterRegistry, changed into function
    function SaveToRegistry(Reg: TBetterRegistry): boolean;                     //CdeB 12/16/1998, //gp 12/28/1998 - adapted to TBetterRegistry, changed into function
  published
    constructor Create(attribName: string);                                     //gp 12/24/1998
    property Background: TColor read fBackground write SetBackground;           //mt 12/14/1998 - renamed, //jdj 12/18/1998 - added Set
    property Foreground: TColor read fForeground write SetForeground;           //mt 12/14/1998 - renamed, //jdj 12/18/1998 - added Set
    property Style: TFontStyles read fStyle write SetStyle;                     //jdj 12/18/1998 - added Set
    property IntegerStyle: integer read GetStyleFromInt write SetStyleFromInt;  //CdeB 12/16/1998
  end;

  TIdentChars = set of char;                                                    //mt 12/22/1998

  THighlighterCapabilities = (                                                  //gp 12/28/1998
    hcUserSettings, // supports Enum/UseUserSettings
    hcRegistry      // supports LoadFrom/SaveToRegistry
  );

  THighlighterCapability = set of THighlighterCapabilities;

  TmwCustomHighLighter = Class(TComponent)
  private
//  fmwEdit: TWinControl;                                                       //jdj 12/18/1998
    fmwEditList: TmwEditList;                                                   //jdj 12/18/1998
  protected
    fDefaultFilter: string;                                                     //gp 1/10/1999
    function GetIdentChars: TIdentChars; virtual;                               //mt 12/22/1998
    function GetLanguageName: string; virtual; abstract;                        //gp 12/24/1998
                                                                                //mt 1/2/1999 change to GetLangaugeName
    function GetAttribCount: integer; virtual; abstract;                        //gp 12/24/1998
    function GetAttribute(idx: integer): TmwHighLightAttributes; virtual; abstract;
    function GetCapability: THighlighterCapability; virtual;                    //gp 12/28/1998
    function GetDefaultFilter: string; virtual;                                 //mt 1/2/1999
    procedure SetDefaultFilter(Value: string); virtual;                         //mt 1/2/1999
  public
    constructor Create(AOwner: TComponent); override;                           //jdj 12/18/1998
    destructor Destroy; override;                                               //jdj 12/18/1998
    function GetEol: Boolean; virtual; abstract;
    function GetRange: Pointer; virtual; abstract;
    function GetToken: String; virtual; abstract;
    function GetTokenPos: Integer; virtual; abstract;
    procedure Next; virtual; abstract;
    procedure SetCanvas(Value: TCanvas); virtual; abstract;
    procedure SetLine(NewValue: String); virtual; abstract;
    procedure SetRange(Value: Pointer); virtual; abstract;
    procedure ReSetRange; virtual; abstract;
    function UseUserSettings(settingIndex: integer): boolean; virtual;          //mt 12/14/1998
                                                                                //gp 12/18/1998 - removed "abstract"
    procedure EnumUserSettings(settings: TStrings); virtual;                    //mt 12/14/1998
                                                                                //gp 12/18/1998 - removed "abstract"
    // idx = 0..GetAttribCount-1                                                //gp 12/24/1998
    function LoadFromRegistry(RootKey: HKEY; Key: string): boolean; virtual;    //gp 12/28/1998
    function SaveToRegistry(RootKey: HKEY; Key: string): boolean; virtual;      //gp 12/28/1998
//  property mwEdit: TWinControl read fmwEdit write fmwEdit;                    //jdj 12/18/1998
    property mwEditList: TmwEditList read FmwEditList;                          //jdj 12/18/1998
    property IdentChars: TIdentChars read GetIdentChars;                        //mt 12/22/1998
    property LanguageName: string read GetLanguageName;                         //gp 12/24/1998
                                                                                //mt 1/2/1999 change to GetLanguageName
    property AttrCount: integer read GetAttribCount;                            //gp 12/24/1998
    property Attribute[idx: integer]: TmwHighLightAttributes read GetAttribute; //gp 12/24/1998
    property Capability: THighlighterCapability read GetCapability;             //gp 12/28/1998
  published
    property DefaultFilter: string read GetDefaultFilter write SetDefaultFilter;//mt 1/2/1999
  end;

implementation

{$IFDEF _Gp_MustEnhanceRegistry}
  function IsRelative(const Value: string): Boolean;
  begin
    Result := not ((Value <> '') and (Value[1] = '\'));
  end;

  function TBetterRegistry.OpenKeyReadOnly(const Key: string): Boolean;
  var
    TempKey: HKey;
    S: string;
    Relative: Boolean;
    {$IFDEF MWE_FPC}
    RegObj : TRegistry;
    {$ENDIF}

  begin
    S := Key;
    Relative := IsRelative(S);

    if not Relative then Delete(S, 1, 1);
    TempKey := 0;
    {$IFDEF MWE_FPC}
    { TODO: Write Registry or Ini file code to replace the RegOpenKeyEx
      function below }
      {$IFDEF WIN32}
        Result := RegOpenKeyEx(GetBaseKey(Relative), PChar(S), 0,
            KEY_READ, TempKey) = ERROR_SUCCESS;
      {$ELSE}
        RegObj := TRegistry.Create;

        try 
          Result := RegObj.OpenKey(S,True);
       {$ENDIF}
    {$ELSE}
    Result := RegOpenKeyEx(GetBaseKey(Relative), PChar(S), 0,
        KEY_READ, TempKey) = ERROR_SUCCESS;
    {$ENDIF}

    if Result then
    begin
      if (CurrentKey <> 0) and Relative then S := CurrentPath + '\' + S;
      ChangeKey(TempKey, S);
    end;

    {$IFDEF MWE_FPC}
      {$IFNDEF WIN32}
        finally
          RegObj.Free;
        end;
      {$ENDIF}
    {$ENDIF}

  end; { TBetterRegistry.OpenKeyReadOnly }
{$ENDIF _Gp_MustEnhanceRegistry}

{ TmwHighLightAttributes }

procedure TmwHighLightAttributes.Assign(Source: TPersistent);                   //gp 12/09/1998
begin
  if Source is TmwHighLightAttributes then begin
    fBackground    := (Source as TmwHighLightAttributes).fBackground;
    fForeground    := (Source as TmwHighLightAttributes).fForeground;
    fStyle         := (Source as TmwHighLightAttributes).fStyle;
    fName          := (Source as TmwHighLightAttributes).fName;                 //gp 12/24/1998
  end
  else inherited Assign(Source);
end;

constructor TmwHighLightAttributes.Create(attribName: string);                  //gp 12/10/1998, //gp 12/24/1998
begin
  inherited Create;
  Background := clWindow;
  Foreground := clWindowText;
  fName := attribName;                                                          //gp 12/24/1998
end;

function TmwHighLightAttributes.LoadFromBorlandRegistry(rootKey: HKEY;          //gp 12/16/1998, //gp 12/28/1998 - renamed
  attrKey, attrName: string; oldStyle: boolean): boolean;
  // How the highlighting information is stored:
  // Delphi 1.0:
  //   I don't know and I don't care.
  // Delphi 2.0 & 3.0:
  //   In the registry branch HKCU\Software\Borland\Delphi\x.0\Highlight
  //   where x=2 or x=3.
  //   Each entry is one string value, encoded as
  //     <foreground RGB>,<background RGB>,<font style>,<default fg>,<default Background>,<fg index>,<Background index>
  //   Example:
  //     0,16777215,BI,0,1,0,15
  //     foreground color (RGB): 0
  //     background color (RGB): 16777215 ($FFFFFF)
  //     font style: BI (bold italic), possible flags: B(old), I(talic), U(nderline)
  //     default foreground: no, specified color will be used (black (0) is used when this flag is 1)
  //     default background: yes, white ($FFFFFF, 15) will be used for background
  //     foreground index: 0 (foreground index (Pal16), corresponds to foreground RGB color)
  //     background index: 15 (background index (Pal16), corresponds to background RGB color)
  // Delphi 4.0:
  //   In the registry branch HKCU\Software\Borland\Delphi\4.0\Editor\Highlight.
  //   Each entry is subkey containing several values:
  //     Foreground Color: foreground index (Pal16), 0..15 (dword)
  //     Background Color: background index (Pal16), 0..15 (dword)
  //     Bold: fsBold yes/no, 0/True (string)
  //     Italic: fsItalic yes/no, 0/True (string)
  //     Underline: fsUnderline yes/no, 0/True (string)
  //     Default Foreground: use default foreground (clBlack) yes/no, False/-1 (string)
  //     Default Background: use default backround (clWhite) yes/no, False/-1 (string)
const
  Pal16: array [0..15] of TColor = (clBlack, clMaroon, clGreen, clOlive,
          clNavy, clPurple, clTeal, clDkGray, clLtGray, clRed, clLime,
          clYellow, clBlue, clFuchsia, clAqua, clWhite);

  function LoadOldStyle(rootKey: HKEY; attrKey, attrName: string): boolean;
  var
    descript : string;
    fgColRGB : string;
    bgColRGB : string;
    fontStyle: string;
    fgDefault: string;
    bgDefault: string;
    fgIndex16: string;
    bgIndex16: string;
    reg      : TBetterRegistry;

    function Get(var name: string): string;
    var
      p: integer;
    begin
      p := Pos(',',name);
      if p = 0 then p := Length(name)+1;
      Result := Copy(name,1,p-1);
      name := Copy(name,p+1,Length(name)-p);
    end; { Get }

  begin { LoadOldStyle }
    Result := false;
    try
      reg := TBetterRegistry.Create;
      reg.RootKey := rootKey;
      try
        with reg do begin
          if OpenKeyReadOnly(attrKey) then begin
            try
              if ValueExists(attrName) then begin
                descript := ReadString(attrName);
                fgColRGB  := Get(descript);
                bgColRGB  := Get(descript);
                fontStyle := Get(descript);
                fgDefault := Get(descript);
                bgDefault := Get(descript);
                fgIndex16 := Get(descript);
                bgIndex16 := Get(descript);
                if bgDefault = '1'
                  then Background := clWindow                                   //gp 12/10/1998
                  else Background := Pal16[StrToInt(bgIndex16)];
                if fgDefault = '1'
                  then Foreground := clWindowText                               //gp 12/10/1998
                  else Foreground := Pal16[StrToInt(fgIndex16)];
                Style := [];
                if Pos('B',fontStyle) > 0 then Style := Style + [fsBold];
                if Pos('I',fontStyle) > 0 then Style := Style + [fsItalic];
                if Pos('U',fontStyle) > 0 then Style := Style + [fsUnderline];
                Result := true;
              end;
            finally CloseKey; end;
          end; // if
        end; // with
      finally reg.Free; end;
    except end;
  end; { LoadOldStyle }

  function LoadNewStyle(rootKey: HKEY; attrKey, attrName: string): boolean;
  var
    fgIndex16    : DWORD;
    bgIndex16    : DWORD;
    fontBold     : string;
    fontItalic   : string;
    fontUnderline: string;
    fgDefault    : string;
    bgDefault    : string;
    reg          : TBetterRegistry;

    function IsTrue(value: string): boolean;
    begin
      Result := not ((UpperCase(value) = 'FALSE') or (value = '0')); 
    end; { IsTrue }

  begin
    Result := false;
    try
      reg := TBetterRegistry.Create;
      reg.RootKey := rootKey;
      try
        with reg do begin
          if OpenKeyReadOnly(attrKey+'\'+attrName) then begin
            try
              if ValueExists('Foreground Color')
                then fgIndex16 := ReadInteger('Foreground Color')
                else Exit;
              if ValueExists('Background Color')
                then bgIndex16 := ReadInteger('Background Color')
                else Exit;
              if ValueExists('Bold')
                then fontBold := ReadString('Bold')
                else Exit;
              if ValueExists('Italic')
                then fontItalic := ReadString('Italic')
                else Exit;
              if ValueExists('Underline')
                then fontUnderline := ReadString('Underline')
                else Exit;
              if ValueExists('Default Foreground')
                then fgDefault := ReadString('Default Foreground')
                else Exit;
              if ValueExists('Default Background')
                then bgDefault := ReadString('Default Background')
                else Exit;
              if IsTrue(bgDefault)
                then Background := clWindow                                     //gp 12/10/1998
                else Background := Pal16[bgIndex16];
              if IsTrue(fgDefault)
                then Foreground := clWindowText                                 //gp 12/10/1998
                else Foreground := Pal16[fgIndex16];
              Style := [];
              if IsTrue(fontBold) then Style := Style + [fsBold];
              if IsTrue(fontItalic) then Style := Style + [fsItalic];
              if IsTrue(fontUnderline) then Style := Style + [fsUnderline];
              Result := true;
            finally CloseKey; end;
          end; // if
        end; // with
      finally reg.Free; end;
    except end;
  end; { LoadNewStyle }

begin
  if oldStyle then Result := LoadOldStyle(rootKey, attrKey, attrName)
              else Result := LoadNewStyle(rootKey, attrKey, attrName);
end; { TmwHighLightAttributes.LoadFromBorlandRegistry }

{begin}                                                                         //jdj 12/18/1998
procedure TmwHighLightAttributes.SetBackground(Value: TColor);
begin
  if fBackGround <> Value then
    begin
      fBackGround := Value;
      if Assigned(fOnChange) then
        fOnChange(Self);
    end;
end;

procedure TmwHighLightAttributes.SetForeground(Value: TColor);
begin 
  if fForeGround <> Value then
    begin
      fForeGround := Value;
      if Assigned(fOnChange) then
        fOnChange(Self);
    end;
end;

procedure TmwHighLightAttributes.SetStyle(Value: TFontStyles);
begin 
  if fStyle <> Value then
    begin
      fStyle := Value;
      if Assigned(fOnChange) then
        fOnChange(Self);
    end;
end;
{end}                                                                           //jdj 12/18/1998

{begin}                                                                         //CdeB 12/16/1998, //gp 12/28/1998 - rewritten for TRegistry
function TmwHighLightAttributes.LoadFromRegistry(Reg: TBetterRegistry): boolean;
var
  key: string;
begin
  key := Reg.CurrentPath;
  if Reg.OpenKeyReadOnly(Name) then begin
    if Reg.ValueExists('Background') then Background := Reg.ReadInteger('Background');
    if Reg.ValueExists('Foreground') then Foreground := Reg.ReadInteger('Foreground');
    if Reg.ValueExists('Style') then IntegerStyle := Reg.ReadInteger('Style');
    reg.OpenKeyReadOnly('\'+key);
    Result := true;
  end
  else Result := false;
end;

function TmwHighLightAttributes.SaveToRegistry(Reg: TBetterRegistry): boolean;
var
  key: string;
begin
  key := Reg.CurrentPath;
  if Reg.OpenKey(Name,true) then begin
    Reg.WriteInteger('Background', Background);
    Reg.WriteInteger('Foreground', Foreground);
    Reg.WriteInteger('Style', IntegerStyle);
    reg.OpenKey('\'+key,false);
    Result := true;
  end
  else Result := false;
end;

function TmwHighLightAttributes.GetStyleFromInt: integer;
begin
  if fsBold in Style then Result:= 1 else Result:= 0;
  if fsItalic in Style then Result:= Result+2;
  if fsUnderline in Style then Result:= Result+4;
  if fsStrikeout in Style then Result:= Result+8;
end;

procedure TmwHighLightAttributes.SetStyleFromInt(const Value: integer);
begin
  if Value and $1 = 0 then  Style:= [] else Style:= [fsBold];
  if Value and $2 <> 0 then Style:= Style+[fsItalic];
  if Value and $4 <> 0 then Style:= Style+[fsUnderline];
  if Value and $8 <> 0 then Style:= Style+[fsStrikeout];
end;
{end}                                                                           //CdeB 12/16/1998

{ TmwCustomHighLighter }

{begin}                                                                         //jdj 12/18/1998
constructor TmwCustomHighLighter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fmwEditList := TmwEditList.Create;
  fDefaultFilter := '';
end;

destructor TmwCustomHighLighter.Destroy;
begin
  fmwEditList.Free;
  inherited Destroy;
end;
{end}                                                                           //jdj 12/18/1998

{begin}                                                                         //gp 12/18/1998
procedure TmwCustomHighLighter.EnumUserSettings(settings: TStrings);
begin
  settings.Clear;
end;

function TmwCustomHighLighter.UseUserSettings(
  settingIndex: integer): boolean;
begin
  Result := false;
end;

function TmwCustomHighLighter.GetIdentChars: TIdentChars;
begin
  Result := [#33..#255];
end;

{end}                                                                           //gp 12/18/1998

{TmwEditList}

{begin}                                                                         //jdj 12/18/1998
procedure TmwEditList.AddInstance(Control: TWinControl);
begin
  if (IndexOf(Control) = -1) then
    Add(Control);
end;

procedure TmwEditList.RemoveInstance(Control: TWinControl);
var
  A: Integer;
begin
  A := IndexOf(Control);
  if (A <> -1) then
    Delete(A);
end;

procedure TmwEditList.Invalidate;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
    TWinControl(Items[i]).Invalidate;
end;
{end}                                                                           //jdj 12/18/1998

{begin}                                                                         //gp 12/28/1998
function TmwCustomHighLighter.LoadFromRegistry(RootKey: HKEY; Key: string): boolean;
var
  r: TBetterRegistry;
  i: integer;
begin
  r := TBetterRegistry.Create;
  try
    r.RootKey := RootKey;
    if r.OpenKeyReadOnly(Key) then begin
      Result := true;
      for i := 0 to AttrCount-1 do
        Result := Result and Attribute[i].LoadFromRegistry(r);
    end
    else Result := false;
  finally r.Free; end;
end;

function TmwCustomHighLighter.SaveToRegistry(RootKey: HKEY; Key: string): boolean;
var
  r: TBetterRegistry;
  i: integer;
begin
  r := TBetterRegistry.Create;
  try
    r.RootKey := RootKey;
    if r.OpenKey(Key,true) then begin
      Result := true;
      for i := 0 to AttrCount-1 do
        Result := Result and Attribute[i].SaveToRegistry(r);
    end
    else Result := false;
  finally r.Free; end;
end;
{end}                                                                           //gp 12/28/1998

function TmwCustomHighLighter.GetCapability: THighlighterCapability;
begin
  Result := [hcRegistry]; //registry save/load supported by default
end;

function TmwCustomHighLighter.GetDefaultFilter: string;                         //mt 1/2/1999
begin
  Result := fDefaultFilter;                                                     //gp 1/10/1999
end;

procedure TmwCustomHighLighter.SetDefaultFilter(Value: string);                 //mt 1/2/1999
begin
  if fDefaultFilter <> Value then fDefaultFilter := Value;                      //gp 1/10/1999
end;

end.
