{==============================================================================
  Project:  cmntins(comment insertion expert for Delphi IDE)
 ==============================================================================
  Content:  TICommentInsertionExpert class
 ==============================================================================
  The contents of this file are subject to the Mozilla Public License Ver. 1.0
  (the "License"); you may not use this file except in compliance with the
  License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
  the specific language governing rights and limitations under the License.
 ==============================================================================
  The Original Code is cmntins.
 ==============================================================================
  The Initial Developer of the Original Code is HANAI Tohru (Japan)
  Portions created by HANAI Tohru are Copyright (C) 1999.
  All Rights Reserved.
 ==============================================================================
  Contributor(s):   HANAI Tohru
 ==============================================================================
  History:  25/09/1999  HANAI Tohru
                        Initial Version
 ==============================================================================}

unit uMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Forms, Dialogs,
  Menus,
  ExptIntf, ToolIntf, EditIntf;

  procedure Register;

implementation

type
  { TICommentInsertionExpert }

  TICommentInsertionExpert = class(TIExpert)
  private
    FMenus: TList;
    procedure CreateMenus;
    procedure DeleteMenus;
    procedure CommentBeginClick(Sender: TIMenuItemIntf);
    procedure CommentEndClick(Sender: TIMenuItemIntf);
    procedure CommentTailClick(Sender: TIMenuItemIntf);
    function GetEditInterfaces(out EditorIntf: TIEditorInterface;
      out EditView: TIEditView): Boolean;
    function CreateComment(const s: string): string;
  protected
    procedure MenuTest(TopMenuItem: TIMenuItemIntf);
  public
    constructor Create;
    destructor Destroy; override;
    function GetName: string; override;
    function GetAuthor: string; override;
    function GetComment: string; override;
    function GetPage: string; override;
    function GetGlyph: HICON; override;
    function GetStyle: TExpertStyle; override;
    function GetState: TExpertState; override;
    function GetIDString: string; override;
    function GetMenuText: string; override;
    procedure Execute; override;
  end;

procedure Register;
begin
  RegisterLibraryExpert(TICommentInsertionExpert.Create);
end;

function TICommentInsertionExpert.GetEditInterfaces(
  out EditorIntf: TIEditorInterface; out EditView: TIEditView): Boolean;
var
  ModIntf: TIModuleInterface;
begin
  Result := False;
  ModIntf := nil;
  try
    with ToolServices do
      ModIntf := GetModuleInterface(GetCurrentFile);
    if Assigned(ModIntf) then
    begin
      EditorIntf := ModIntf.GetEditorInterface;
      if EditorIntf.GetViewCount = 1 then
      begin
        EditView := EditorIntf.GetView(0);
        Result := True;
      end;
    end;
  finally
    ModIntf.Free;
  end;
end;

type
  TMenuItemsData = record
    Caption: string;
    Name: string;
    Hint: string;
    Key: Word;
    Shift: TShiftState;
    Code: Pointer;
  end;

//=== [variables] ==============================================================
const
  sLineBreak: string    = #$0D#$0A;
  sCommentHead: string  = '//th ';
  sDateFmt: string      = 'yyyy-mm-dd';
  sCommentTail: string  = '';

  MenuItemsData: array[0..2] of TMenuItemsData = (
    (Caption: '{begin}';
     Name: 'ciemnBegin';
     Hint: '';
     Key: Ord('J');
     Shift: [ssAlt, ssCtrl];
     Code: @TICommentInsertionExpert.CommentBeginClick),
    (Caption: '{end}';
     Name: 'ciemnEnd';
     Hint: '';
     Key: Ord('K');
     Shift: [ssAlt, ssCtrl];
     Code: @TICommentInsertionExpert.CommentEndClick),
    (Caption: 'tail';
     Name: 'ciemnTail';
     Hint: '';
     Key: Ord('L');
     Shift: [ssAlt, ssCtrl];
     Code: @TICommentInsertionExpert.CommentTailClick)
    );
//==============================================================================

{ TICommentInsertionExpert }

constructor TICommentInsertionExpert.Create;
begin
  inherited;

  FMenus := TList.Create;
  CreateMenus;
end;

destructor TICommentInsertionExpert.Destroy;
begin
  DeleteMenus;

  inherited;
end;

procedure TICommentInsertionExpert.Execute;
begin
end;

function TICommentInsertionExpert.GetAuthor: string;
begin
  Result := 'HANAI Tohru';
end;

function TICommentInsertionExpert.GetComment: string;
begin
  Result := '';
end;

function TICommentInsertionExpert.GetGlyph: HICON;
begin
  Result := 0;
end;

function TICommentInsertionExpert.GetIDString: string;
begin
  Result := 'HANAI Tohru.CommentInsertionExpert';
end;

function TICommentInsertionExpert.GetMenuText: string;
begin
  Result := '';
end;

function TICommentInsertionExpert.GetName: string;
begin
  Result := 'CommentInsertionExpert';
end;

function TICommentInsertionExpert.GetPage: string;
begin
  Result := '';
end;

function TICommentInsertionExpert.GetState: TExpertState;
begin
  Result := [];
end;

function TICommentInsertionExpert.GetStyle: TExpertStyle;
begin
  Result := esAddIn;
end;

procedure TICommentInsertionExpert.CreateMenus;
var
  TopItem: TIMenuItemIntf;
  DstMenu: TIMenuItemIntf;
  NewItem: TIMenuItemIntf;
  i: Integer;
  Method: TMethod;
begin
  TopItem := nil;
  DstMenu := nil;
  try
    TopItem := ToolServices.GetMainMenu.GetMenuItems;
//    MenuTest(TopItem);
    DstMenu := ToolServices.GetMainMenu.FindMenuItem('EditMenu');
    if Assigned(DstMenu) then
    begin
      FMenus.Add(DstMenu.InsertItem(DstMenu.GetItemCount, '-', 'ciemnBreak',
        '', 0, 0, 0, [mfVisible], nil));
      NewItem := DstMenu.InsertItem(DstMenu.GetItemCount, 'Comment',
        'ciemnSubTop', '', 0, 0, 0, [mfEnabled, mfVisible], nil);
      FMenus.Add(NewItem);
      Method.Data := Self;
      for i := Low(MenuItemsData) to High(MenuItemsData) do
        with MenuItemsData[i] do
        begin
          Method.Code := Code;
          FMenus.Add(NewItem.InsertItem(NewItem.GetItemCount,
            Caption, Name, Hint, ShortCut(Key, Shift), 0, 0,
            [mfEnabled, mfVisible], TIMenuClickEvent(Method)));
        end;
    end;
  finally
    TopItem.Free;
    DstMenu.Free;
  end;
end;

procedure TICommentInsertionExpert.DeleteMenus;
begin
  while FMenus.Count > 0 do
  begin
    TIMenuItemIntf(FMenus.Last).Free;
    FMenus.Delete(FMenus.Count - 1);
  end;
end;

procedure TICommentInsertionExpert.MenuTest(TopMenuItem: TIMenuItemIntf);

  procedure BuildMenu(AOwner: TComponent;
    SrcMenu: TIMenuItemIntf; DstMenu: TMenuItem);
  var
    i: Integer;
    AnItem: TMenuItem;
    AnIdeItem: TIMenuItemIntf;
  begin
    AnItem := TMenuItem.Create(AOwner);
    with AnItem do
    begin
      Caption := Format('%s(%s)', [SrcMenu.GetCaption, SrcMenu.GetName]);
      ShortCut := SrcMenu.GetShortCut;
    end;
    DstMenu.Add(AnItem);
    for i := 0 to SrcMenu.GetItemCount - 1 do
    begin
      AnIdeItem := SrcMenu.GetItem(i);
      try
        BuildMenu(AOwner, AnIdeItem, AnItem);
      finally
        AnIdeItem.Free;
      end;
    end;
  end;


var
  AForm: TForm;
  AMainMenu: TMainMenu;
  AnIdeItem: TIMenuItemIntf;
  i: Integer;
begin
  AForm := nil;
  try
    AForm := TForm.Create(nil);
    AMainMenu := TMainMenu.Create(AForm);
    for i := 0 to TopMenuItem.GetItemCount - 1 do
    begin
      AnIdeItem := TopMenuItem.GetItem(i);
      try
        BuildMenu(AForm, AnIdeItem, AMainMenu.Items);
      finally
        AnIdeItem.Free;
      end;
    end;

    AForm.ShowModal;
  finally
    AForm.Free;
  end;
end;

procedure TICommentInsertionExpert.CommentBeginClick(Sender: TIMenuItemIntf);
const
  cmnt = '{begin}';
var
  EditorIntf: TIEditorInterface;
  EditView: TIEditView;
  Writer: TIEditWriter;
  EditPos: TEditPos;
  CharPos: TCharPos;
begin
  Writer := nil;
  if GetEditInterfaces(EditorIntf, EditView) then
    try
      EditPos := EditView.CursorPos;
      EditPos.Col := 1;
      EditView.ConvertPos(True, EditPos, CharPos);
      Writer := EditorIntf.CreateUndoableWriter;
      Writer.CopyTo(EditView.CharPosToPos(CharPos));
      Writer.Insert(PChar(cmnt + CreateComment(cmnt) + sLineBreak));
    finally
      EditorIntf.Free;
      EditView.Free;
      Writer.Free;
    end;
end;

procedure TICommentInsertionExpert.CommentEndClick(Sender: TIMenuItemIntf);
const
  cmnt = '{end}';
var
  EditorIntf: TIEditorInterface;
  EditView: TIEditView;
  Writer: TIEditWriter;
  EditPos: TEditPos;
  CharPos: TCharPos;
begin
  Writer := nil;
  if GetEditInterfaces(EditorIntf, EditView) then
    try
      EditPos := EditView.CursorPos;
      EditPos.Col := 1;
      EditView.ConvertPos(True, EditPos, CharPos);
      Writer := EditorIntf.CreateUndoableWriter;
      Writer.CopyTo(EditView.CharPosToPos(CharPos));
      Writer.Insert(PChar(cmnt + CreateComment(cmnt) + sLineBreak));
    finally
      EditorIntf.Free;
      EditView.Free;
      Writer.Free;
    end;
end;

procedure TICommentInsertionExpert.CommentTailClick(Sender: TIMenuItemIntf);
var
  EditorIntf: TIEditorInterface;
  EditView: TIEditView;
  Reader: TIEditReader;
  Writer: TIEditWriter;
  EditPos: TEditPos;
  CharPos: TCharPos;
  s: array[0..950] of Char;
  nPos: LongInt;
  p: PChar;
begin
  Writer := nil;
  if GetEditInterfaces(EditorIntf, EditView) then
    try
      EditPos := EditView.CursorPos;
      EditPos.Col := 1;
      EditView.ConvertPos(True, EditPos, CharPos);
      nPos := EditView.CharPosToPos(CharPos);
      Reader := EditorIntf.CreateReader;
      try
        s[Reader.GetText(nPos, s, SizeOf(s) - 1)] := #0;
      finally
        Reader.Free;
      end;
      Writer := EditorIntf.CreateUndoableWriter;
      p := AnsiStrScan(s, #$0D);
      if Assigned(p) then
        p^ := #0;
      Inc(nPos, (LongInt(p) - LongInt(@s[Low(s)])));
      if Writer.CopyTo(nPos) then
        Writer.Insert(PChar(CreateComment(s)));
    finally
      EditorIntf.Free;
      EditView.Free;
      Writer.Free;
    end;
end;

function TICommentInsertionExpert.CreateComment(const s: string): string;
var
  Len: Integer; 
  sDate: string;
begin
  sDate := FormatDateTime(sDateFmt, Now);
  Len := Length(s);
  if Len < 80 then
    Result := Format('%s%s%s%s',
      [StringOfChar(#$20, 80 - Len), sCommentHead, sDate, sCommentTail])
  else
    Result := Format('%s%s%s',
      [sCommentHead, sDate, sCommentTail])
end;

end.
